/*
 * MeasurementDetailsPanel.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Ikayaki.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package ikayaki.gui;

import ikayaki.*;

import javax.swing.*;
import javax.swing.table.AbstractTableModel;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.text.DecimalFormat;
import java.text.NumberFormat;

/**
 * Shows the details of the active measurement step.
 *
 * @author Esko Luontola
 */
public class MeasurementDetailsPanel extends ProjectComponent {

    private JTable detailsTable;
    private DetailsTableModel detailsTableModel;

    private JTable errorsTable;
    private ErrorsTableModel errorsTableModel;

    private JCheckBox stopCheckBox;

    /**
     * The measurement step whose details are being shown or null to show a blank table.
     */
    private MeasurementStep step;

    public MeasurementDetailsPanel() {

        // build the tables
        detailsTableModel = new DetailsTableModel();
        detailsTable = new JTable(detailsTableModel);
        detailsTable.setFocusable(false);
        detailsTable.setEnabled(false);
        detailsTable.setDefaultRenderer(StyledWrapper.class, new StyledTableCellRenderer());

        errorsTableModel = new ErrorsTableModel();
        errorsTable = new JTable(errorsTableModel);
        errorsTable.setFocusable(false);
        errorsTable.setEnabled(false);
        errorsTable.setDefaultRenderer(StyledWrapper.class, new StyledTableCellRenderer());

        detailsTable.getColumnModel().getColumn(0).setPreferredWidth(50);
        detailsTable.getColumnModel().getColumn(1).setPreferredWidth(100);
        detailsTable.getColumnModel().getColumn(2).setPreferredWidth(100);
        detailsTable.getColumnModel().getColumn(3).setPreferredWidth(100);
        detailsTable.getTableHeader().setReorderingAllowed(false);
        detailsTable.getTableHeader().setResizingAllowed(false);

        errorsTable.getColumnModel().getColumn(0).setPreferredWidth(40);
        errorsTable.getColumnModel().getColumn(1).setPreferredWidth(75);
        errorsTable.getColumnModel().getColumn(2).setPreferredWidth(75);
        errorsTable.getColumnModel().getColumn(3).setPreferredWidth(75);
        errorsTable.getColumnModel().getColumn(4).setPreferredWidth(95);
        errorsTable.getColumnModel().getColumn(5).setPreferredWidth(40);
        errorsTable.getTableHeader().setReorderingAllowed(false);
        errorsTable.getTableHeader().setResizingAllowed(false);

        // emulate the looks of a JScrollPane
        JPanel detailsTablePanel = new JPanel(new BorderLayout());
        detailsTablePanel.add(detailsTable.getTableHeader(), BorderLayout.NORTH);
        detailsTablePanel.add(detailsTable, BorderLayout.CENTER);
        detailsTablePanel.setBorder(new JScrollPane().getBorder());

        JPanel errorsTablePanel = new JPanel(new BorderLayout());
        errorsTablePanel.add(errorsTable.getTableHeader(), BorderLayout.NORTH);
        errorsTablePanel.add(errorsTable, BorderLayout.CENTER);
        errorsTablePanel.setBorder(new JScrollPane().getBorder());

        stopCheckBox = new JCheckBox("Stop on warning of poor measurement");
        JPanel checkBoxPanel = new JPanel(new BorderLayout());
        checkBoxPanel.add(stopCheckBox, BorderLayout.WEST);

        // lay out the components
        JPanel tablePanel = new JPanel(new BorderLayout(10, 10));
        tablePanel.add(detailsTablePanel, "North");
        tablePanel.add(errorsTablePanel, "Center");
        tablePanel.add(checkBoxPanel, "South");

        JPanel contentPane = new JPanel(new FlowLayout(FlowLayout.CENTER));
        contentPane.add(tablePanel);

        JScrollPane scrollPane = new JScrollPane(contentPane,
                JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
                JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        scrollPane.setBorder(null);
        this.setLayout(new BorderLayout());
        this.add(scrollPane);
        this.setMinimumSize(new Dimension(0, 245));
        this.setPreferredSize(new Dimension(435, 245));

        stopCheckBox.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                saveCheckBoxProperty();                
            }
        });
    }

    @Override public void setProject(Project project) {
        if (getProject() != null) {
            boolean stopProp = getProject().getProperty(Project.STOP_ON_WARNING_PROPERTY, "1").equals("1");
            // The check box state is saved only if it has been changed
            if (stopProp != stopCheckBox.isSelected()) {
                // Save the check box state of the old project
                saveCheckBoxProperty();
            }
        }
        super.setProject(project);
        setStep(null);
        // Check the check box state of the new project
        if (project != null && (project.getType() == Project.Type.AF || project.getType() == Project.Type.GAF)) {
            boolean stop = true;
            String stopStr = project.getProperty(Project.STOP_ON_WARNING_PROPERTY);
            if (stopStr != null && stopStr.equals("0")) {
                stop = false;
            }
            stopCheckBox.setEnabled(true);
            stopCheckBox.setSelected(stop);
        } else {
            // Disable the check box in non-AF projects
            stopCheckBox.setEnabled(false);
        }
    }

    public MeasurementStep getStep() {
        return step;
    }

    public void setStep(MeasurementStep step) {
        this.step = step;
        this.detailsTableModel.setStep(step);
        this.errorsTableModel.setStep(step);
    }

    @Override public void measurementUpdated(MeasurementEvent event) {
        if (event.getStep() == step) {
            detailsTableModel.fireTableDataChanged();
            errorsTableModel.fireTableDataChanged();
        }
    }

    private void saveCheckBoxProperty() {
        Project project = getProject();
        if (project != null) {
            String str = stopCheckBox.isSelected()? "1": "0";
            project.setProperty(Project.STOP_ON_WARNING_PROPERTY, str);
        }
    }

    /**
     * Table model for the details table.
     *
     * @author Esko Luontola
     */
    private class DetailsTableModel extends AbstractTableModel {

        private MeasurementStep step;

        private final String[] COLUMNS = new String[]{" ", "X", "Y", "Z"};
        private final int HEADER_COLUMN = 0;
        private final int X_COLUMN = 1;
        private final int Y_COLUMN = 2;
        private final int Z_COLUMN = 3;

        private NumberFormat numberFormat = new DecimalFormat("0.000000E0");

        private StyledWrapper defaultWrapper = new StyledWrapper();
        private StyledWrapper headerWrapper = new StyledWrapper();

        public DetailsTableModel() {
            defaultWrapper.horizontalAlignment = SwingConstants.TRAILING;
            headerWrapper.horizontalAlignment = SwingConstants.TRAILING;
            headerWrapper.font = new JLabel("").getFont().deriveFont(Font.BOLD);
        }

        public MeasurementStep getStep() {
            return step;
        }

        public void setStep(MeasurementStep step) {
            this.step = step;
            fireTableDataChanged();
        }

        public int getRowCount() {

            // calculate the expected number of results based on measurement rotations and project type
            int nonSample = 3;
            if (Settings.getHolderCalibration() == null
                    || (step != null && step.getProject().isHolderCalibration())) {
                nonSample--;
            }
            int expected = Math.max(1, 4 * Settings.getMeasurementRotations()) + nonSample;

            Project project = getProject();
            if ( project!=null ) {
                Project.Type type = project.getType();
                if ( Project.Type.GAF == type && Settings.getMeasurementRotations()>=1 ) {
                        expected += 4; // 4 more background measurements (as 6 in total)
                }
            }

            if (step == null) {
                return expected;

            } else if (step.getState() == MeasurementStep.State.DONE
                    || step.getState() == MeasurementStep.State.DONE_RECENTLY) {
                return step.getResults();

            } else {
                // try to estimate the number of steps
                int count = step.getResults();

                // if the last step is not BG, there are more steps coming
                if (count >= nonSample && step.getResult(count - 1).getType() != MeasurementResult.Type.NOISE) {
                    count++;
                }
                return Math.max(expected, count);
            }
        }

        public int getColumnCount() {
            return COLUMNS.length;
        }

        @Override public String getColumnName(int column) {
            return COLUMNS[column];
        }

        @Override public Class<?> getColumnClass(int columnIndex) {
            return StyledWrapper.class;
        }

        public Object getValueAt(int rowIndex, int columnIndex) {
            Object value = "N/A";
            Project project = getProject();

            if (step != null && rowIndex < step.getResults()) {

                // get the values from the step
                switch (columnIndex) {
                case HEADER_COLUMN:
                    switch (step.getResult(rowIndex).getType()) {
                    case HOLDER:
                        value = "Holder";
                        break;
                    case NOISE:
                        value = "BG";
                        break;
                    case SAMPLE:
                        value = Integer.toString(step.getResult(rowIndex).getRotation());
                        break;
                    default:
                        assert false;
                        value = null;
                        break;
                    }
                    break;
                case X_COLUMN:
                    value = numberFormat.format(step.getResult(rowIndex).getRawX());
                    break;
                case Y_COLUMN:
                    value = numberFormat.format(step.getResult(rowIndex).getRawY());
                    break;
                case Z_COLUMN:
                    value = numberFormat.format(step.getResult(rowIndex).getRawZ());
                    break;
                default:
                    assert false;
                    value = null;
                    break;
                }

            } else {

                // try to guess the values
                if (columnIndex == HEADER_COLUMN) {
                    int i = (Settings.getHolderCalibration() == null
                            || (step != null && step.getProject().isHolderCalibration()))
                            ? rowIndex + 1
                            : rowIndex;

                    Project.Type type = Project.Type.AF;
                    if ( project!=null ) {
                        type = project.getType();
                    }

                    if (i == 0) {
                        value = "Holder";
                    } else if ((project==null || type != Project.Type.GAF)
								&& (i == 1 || rowIndex == getRowCount() - 1)) {
                        value = "BG";
                    } else if (type == Project.Type.GAF &&
                               Settings.getMeasurementRotations() >= 1 &&
                               (i <= 3 || rowIndex >= getRowCount() - 3)) {
                        value = "BG";
                    } else if (type == Project.Type.GAF &&
                               Settings.getMeasurementRotations() == 0 &&
                               (i == 1 || rowIndex >= getRowCount() - 1)) {
                        value = "BG";                               
                    } else {
                        int switchIdx;
                        if ( type == Project.Type.GAF ) {
                            if ( Settings.getMeasurementRotations() >= 1 ) {
                                switchIdx = (i-4) % 4;
                            } else {
                                switchIdx = 0;
                             }
                        } else {
                            switchIdx = (i-2) % 4;
                        }
                        switch (switchIdx) {
                        case 0:
                            value = "0";
                            break;
                        case 1:
                            value = "90";
                            break;
                        case 2:
                            value = "180";
                            break;
                        case 3:
                            value = "270";
                            break;
                        default:
                            assert false;
                            value = null;
                            break;
                        }
                    }
                } else {
                    value = null;
                }
            }
            return wrap(value, rowIndex, columnIndex);
        }

        public StyledWrapper wrap(Object value, int rowIndex, int columnIndex) {
            StyledWrapper wrapper;

            // choose the style according to the column
            if (columnIndex == HEADER_COLUMN) {
                wrapper = headerWrapper;
            } else {
                wrapper = defaultWrapper;
            }

            // wrap the cell's value and return it
            wrapper.value = value;
            return wrapper;
        }
    }

    /**
     * Table model for the error table.
     *
     * @author Esko Luontola
     */
    private class ErrorsTableModel extends AbstractTableModel {

        private MeasurementStep step;

        private final String[] COLUMNS = new String[]{" ", "Signal/Noise", "Signal/Drift", "Signal/Holder", "Signal/Background", "\u03b863"};
        private final int HEADER_COLUMN = 0;
        private final int SIGNAL_NOISE_COLUMN = 1;
        private final int SIGNAL_DRIFT_COLUMN = 2;
        private final int SIGNAL_HOLDER_COLUMN = 3;
        private final int SIGNAL_BACKGROUND_COLUMN = 4;
        private final int THETA63_COLUMN = 5;

        private StyledWrapper defaultWrapper = new StyledWrapper();
        private StyledWrapper warningWrapper = Settings.getWarningWrapperInstance();
        private StyledWrapper headerWrapper = new StyledWrapper();
        private DecimalFormat numberFormat = new DecimalFormat();

        public ErrorsTableModel() {
            defaultWrapper.horizontalAlignment = SwingConstants.TRAILING;
            headerWrapper.horizontalAlignment = SwingConstants.TRAILING;
            warningWrapper.horizontalAlignment = SwingConstants.TRAILING;
            headerWrapper.font = new JLabel("").getFont().deriveFont(Font.BOLD);
            numberFormat.setMinimumFractionDigits(3);
            numberFormat.setMaximumFractionDigits(3);
        }

        public MeasurementStep getStep() {
            return step;
        }

        public void setStep(MeasurementStep step) {
            this.step = step;
            fireTableDataChanged();
        }

        public int getRowCount() {
            return 1;
        }

        public int getColumnCount() {
            return COLUMNS.length;
        }

        @Override public String getColumnName(int column) {
            return COLUMNS[column];
        }

        @Override public Class<?> getColumnClass(int columnIndex) {
            return StyledWrapper.class;
        }

        public Object getValueAt(int rowIndex, int columnIndex) {
            Object value;
            switch (columnIndex) {
            case HEADER_COLUMN:
                value = "Error";
                break;
            case SIGNAL_NOISE_COLUMN:
                value = MeasurementValue.SIGNAL_TO_NOISE.getValue(step);
                break;
            case SIGNAL_DRIFT_COLUMN:
                value = MeasurementValue.SIGNAL_TO_DRIFT.getValue(step);
                break;
            case SIGNAL_HOLDER_COLUMN:
                value = MeasurementValue.SIGNAL_TO_HOLDER.getValue(step);
                break;
            case SIGNAL_BACKGROUND_COLUMN:
                value = MeasurementValue.SIGNAL_TO_BACKGROUND.getValue(step);
                break;
            case THETA63_COLUMN:
                value = MeasurementValue.THETA63.getValue(step);
                break;
            default:
                value = null;
                break;
            }

            if (value != null && value instanceof Number) {
                value = numberFormat.format(((Number) value).doubleValue());
            }
            return wrap(value, rowIndex, columnIndex);
        }

        public StyledWrapper wrap(Object value, int rowIndex, int columnIndex) {
            StyledWrapper wrapper;

            // choose the style according to the column and check for warnings
            switch (columnIndex) {
                case HEADER_COLUMN:
                    wrapper = step != null && step.getWarnings()? warningWrapper: defaultWrapper;
                    break;
                case SIGNAL_NOISE_COLUMN:
                    wrapper = step != null && step.getSignalToNoiseWarning()? warningWrapper: defaultWrapper;
                    break;
                case SIGNAL_DRIFT_COLUMN:
                    wrapper = step != null && step.getSignalToDriftWarning()? warningWrapper: defaultWrapper;
                    break;
                case SIGNAL_HOLDER_COLUMN:
                    wrapper = step != null && step.getSignalToHolderWarning()? warningWrapper: defaultWrapper;
                    break;
                case SIGNAL_BACKGROUND_COLUMN:
                    wrapper = step != null && step.getSignalToBackgroundWarning()? warningWrapper: defaultWrapper;
                    break;
                case THETA63_COLUMN:
                    wrapper = step != null && step.getTheta63Warning()? warningWrapper: defaultWrapper;
                    break;
                default:
                    assert false;
                    wrapper = defaultWrapper;
                    break;
            }

            // wrap the cell's value and return it
            wrapper.value = value;
            return wrapper;
        }
    }
}
