/*
 * ProgramSettingsPanel.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Ikayaki.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package ikayaki.gui;

import com.intellij.uiDesigner.core.GridConstraints;
import com.intellij.uiDesigner.core.GridLayoutManager;
import com.intellij.uiDesigner.core.Spacer;
import ikayaki.Ikayaki;
import ikayaki.MeasurementSequence;
import ikayaki.Settings;
import ikayaki.ProjectEvent;
import static ikayaki.ProjectEvent.Type.DATA_CHANGED;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.text.DefaultFormatterFactory;
import javax.swing.text.NumberFormatter;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.List;

/**
 * Controls for editing the program settings.
 *
 * @author Esko Luontola
 */
public class ProgramSettingsPanel extends JPanel {

    private JDialog creator;

    private JFormattedTextField measurementRotationsField;
    private JFormattedTextField sigToDriftField;
    private JFormattedTextField sigToHandlerField;
    private JFormattedTextField sigToNoiseField;
    private JFormattedTextField sigToBackgroundField;
    private JFormattedTextField degaussTimeLimit;
    private JFormattedTextField theta63Field;
    private JComboBox holderCalibrationCombo;
    private JTable sequencesTable;
    private JButton sequencesDeleteButton;
    private JPanel defaultColumnsPane;
    private JButton closeButton;
    private JPanel contentPane;
    private JComboBox csvSeparatorCombo;
    private JComboBox numformatTableCombo;
    private JComboBox numformatPrintoutCombo;
    private JLabel ThetaLabel;

    public static final String[] NUMFORMATS = { "0.000", "0.000E0" };
    public static final String[] CSV_SEPARATORS = { ";", "\t", ",", "#" };
    private static final String[] CSV_SEPARATOR_NAMES = {
            "[;] semicolon", "[ ] tab", "[,] comma", "[#] hash"
    };

    public ProgramSettingsPanel(JDialog dialog) {
        this.creator = dialog;

        $$$setupUI$$$();
        ThetaLabel.setText("\u03b863");
        setLayout(new BorderLayout());
        add(contentPane, BorderLayout.CENTER);

        sequencesDeleteButton.setMnemonic(KeyEvent.VK_D);
        closeButton.setMnemonic(KeyEvent.VK_C);

        /* Measurement Rotations */

        NumberFormatter formatRot = new NumberFormatter();
        formatRot.setMinimum(new Integer(0));
        formatRot.setMaximum(new Integer(999));
        measurementRotationsField.setFormatterFactory(new DefaultFormatterFactory(formatRot));
        measurementRotationsField.setValue(new Integer(Settings.getMeasurementRotations()));

        // set formatting of some other textfields
        NumberFormatter formatDegT = new NumberFormatter();
        formatDegT.setMinimum(new Integer(1));
        formatDegT.setMaximum(new Integer(15));
        degaussTimeLimit.setFormatterFactory(new DefaultFormatterFactory(formatDegT));

        // autosaving
        measurementRotationsField.addPropertyChangeListener("value", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                int value = ((Number) measurementRotationsField.getValue()).intValue();
                Settings.setMeasurementRotations(value);
            }
        });

        // automatically select all text to make the entering of a new value easy
        measurementRotationsField.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        measurementRotationsField.setSelectionStart(0);
                        measurementRotationsField.setSelectionEnd(measurementRotationsField.getText().length());
                    }
                });
            }

            public void focusLost(FocusEvent e) {
                // DO NOTHING
            }
        });

        /* Sample Holder Calibration */

        final File[] calibrationFiles = Settings.getCalibrationProjectFiles();
        File holderCalibrationFile = Settings.getHolderCalibrationFile();
        holderCalibrationCombo.addItem("");     // option for selecting no file

        // add all calibration projects to the list
        for (int i = 0; i < calibrationFiles.length; i++) {
            File file = calibrationFiles[i];
            String name = file.getName().substring(0, file.getName().lastIndexOf(Ikayaki.FILE_TYPE));
            holderCalibrationCombo.addItem(name);

            if (file.equals(holderCalibrationFile)) {
                holderCalibrationCombo.setSelectedIndex(holderCalibrationCombo.getItemCount() - 1);
            }
        }

        // autosaving
        holderCalibrationCombo.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                int index = holderCalibrationCombo.getSelectedIndex();
                index--;
                if (index >= 0 && index < calibrationFiles.length) {
                    Settings.setHolderCalibrationFile(calibrationFiles[index]);
                } else {
                    Settings.setHolderCalibrationFile(null);
                }
            }
        });

        /* Saved Sequences */

        final EditSequencesTableModel tableModel = new EditSequencesTableModel();
        sequencesTable.setModel(tableModel);
        sequencesTable.setTableHeader(null);
        sequencesTable.getParent().setBackground(sequencesTable.getBackground());
        sequencesTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        // deleting of sequences
        sequencesDeleteButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                int index = sequencesTable.getSelectedRow();
                if (sequencesTable.isEditing()) {
                    // cancel the editing of the selected cell to avoid renaming the next sequence in the list
                    sequencesTable.editCellAt(-1, -1);
                }
                if (index >= 0) {
                    tableModel.deleteSequence(index);
                }
            }
        });

        // a sequence must be selected to be able to delete it
        sequencesDeleteButton.setEnabled(false);
        sequencesTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e) {
                sequencesDeleteButton.setEnabled(sequencesTable.getSelectedRow() >= 0);
            }
        });

        /* Default Columns */

        SequenceColumn[] allColumns = SequenceColumn.getAllColumns();
        List<SequenceColumn> defaultColumns = Settings.getDefaultColumns();

        defaultColumnsPane.setLayout(new GridBagLayout());
        GridBagConstraints gc = new GridBagConstraints();
        gc.fill = GridBagConstraints.BOTH;
        gc.ipadx = 10;
        gc.gridy = 0;

        // add all columns to the list as checkboxes
        for (final SequenceColumn column : allColumns) {
            final JCheckBox checkBox = new JCheckBox(column.getColumnName(null));
            JLabel description = new JLabel(column.getToolTipText(null));

            // autosaving
            checkBox.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    Settings.setDefaultColumn(column, checkBox.isSelected());
                }
            });
            checkBox.setSelected(defaultColumns.contains(column));

            gc.gridx = 0;
            defaultColumnsPane.add(checkBox, gc);
            gc.gridx = 1;
            defaultColumnsPane.add(description, gc);
            gc.gridy++;
        }


        /* Formatting options */

        int csv_selected_idx = 0;
        String csv_selected_sep = Settings.getCSVfileSeparator();
        for(int i=0; i<CSV_SEPARATORS.length; i++) {
            csvSeparatorCombo.addItem(CSV_SEPARATOR_NAMES[i]);
            if(CSV_SEPARATORS[i] == csv_selected_sep) {
                csv_selected_idx = i;
            }
        }
        csvSeparatorCombo.setSelectedIndex(csv_selected_idx);
        // autosave
        csvSeparatorCombo.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                int index = csvSeparatorCombo.getSelectedIndex();
                if (index >= 0 && index<CSV_SEPARATORS.length) {
                    Settings.setCSVfileSeparator(CSV_SEPARATORS[index]);
                }
            }
        });

        numformatTableCombo.addItem("fractional  (12.345)");
        numformatTableCombo.addItem("engineering (1.234E1)");
        numformatPrintoutCombo.addItem("fractional  (12.345)");
        numformatPrintoutCombo.addItem("engineering (1.234E1)");

        String tableFormat = Settings.getTableNumberFormat();
        String printoutFormat = Settings.getPrintoutNumberFormat();

        if(tableFormat.indexOf("E") > 0) numformatTableCombo.setSelectedIndex(1);
        else numformatTableCombo.setSelectedIndex(0);
        if(printoutFormat.indexOf("E") > 0) numformatPrintoutCombo.setSelectedIndex(1);
        else numformatPrintoutCombo.setSelectedIndex(0);

        numformatTableCombo.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                int index = numformatTableCombo.getSelectedIndex();
                if (index >= 0 && index <= NUMFORMATS.length) {
                    Settings.setTableNumberFormat(NUMFORMATS[index]);
                }
            }
        });

        numformatPrintoutCombo.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                int index = numformatPrintoutCombo.getSelectedIndex();
                if (index >= 0 && index <= NUMFORMATS.length) {
                    Settings.setPrintoutNumberFormat(NUMFORMATS[index]);
                }
            }
        });

        /* Warning threshold values for S/D, S/H, S/N and THETA63 */

        sigToDriftField.setValue(Settings.getSDthreshold());
        sigToHandlerField.setValue(Settings.getSHthreshold());
        sigToNoiseField.setValue(Settings.getSNthreshold());
        sigToBackgroundField.setValue(Settings.getSBthreshold());
        theta63Field.setValue(Settings.getTheta63threshold());
        degaussTimeLimit.setValue(Settings.getDegaussTimeLimit());

        // autosaving

        PropertyChangeListener thresholdsPropertyChangeListener = new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                double value;
                value = ((Number) sigToDriftField.getValue()).doubleValue();
                Settings.setSDthreshold(value);
                value = ((Number) sigToHandlerField.getValue()).doubleValue();
                Settings.setSHthreshold(value);
                value = ((Number) sigToNoiseField.getValue()).doubleValue();
                Settings.setSNthreshold(value);
                value = ((Number) sigToBackgroundField.getValue()).doubleValue();
                Settings.setSBthreshold(value);
                value = ((Number) theta63Field.getValue()).doubleValue();
                Settings.setTheta63threshold(value);
                int time;
                time = ((Number)degaussTimeLimit.getValue()).intValue();
                Settings.setDegaussTimeLimit(time);
            }
        };

        sigToDriftField.addPropertyChangeListener("value", thresholdsPropertyChangeListener);
        sigToHandlerField.addPropertyChangeListener("value", thresholdsPropertyChangeListener);
        sigToBackgroundField.addPropertyChangeListener("value", thresholdsPropertyChangeListener);
        sigToNoiseField.addPropertyChangeListener("value", thresholdsPropertyChangeListener);
        theta63Field.addPropertyChangeListener("value", thresholdsPropertyChangeListener);
        degaussTimeLimit.addPropertyChangeListener("value", thresholdsPropertyChangeListener);

        // automatically select all text to make the entering of a new value easy
        sigToDriftField.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        sigToDriftField.setSelectionStart(0);
                        sigToDriftField.setSelectionEnd(sigToDriftField.getText().length());
                    }
                });
            }

            public void focusLost(FocusEvent e) {
                // DO NOTHING
            }
        });

        sigToHandlerField.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        sigToHandlerField.setSelectionStart(0);
                        sigToHandlerField.setSelectionEnd(sigToHandlerField.getText().length());
                    }
                });
            }

            public void focusLost(FocusEvent e) {
                // DO NOTHING
            }
        });

        sigToNoiseField.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        sigToNoiseField.setSelectionStart(0);
                        sigToNoiseField.setSelectionEnd(sigToNoiseField.getText().length());
                    }
                });
            }

            public void focusLost(FocusEvent e) {
                // DO NOTHING
            }
        });

        sigToBackgroundField.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        sigToBackgroundField.setSelectionStart(0);
                        sigToBackgroundField.setSelectionEnd(sigToBackgroundField.getText().length());
                    }
                });
            }

            public void focusLost(FocusEvent e) {
                // DO NOTHING
            }
        });

        theta63Field.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        theta63Field.setSelectionStart(0);
                        theta63Field.setSelectionEnd(theta63Field.getText().length());
                    }
                });
            }

            public void focusLost(FocusEvent e) {
                // DO NOTHING
            }
        });

        degaussTimeLimit.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        degaussTimeLimit.setSelectionStart(0);
                        degaussTimeLimit.setSelectionEnd(degaussTimeLimit.getText().length());
                    }
                });
            }

            public void focusLost(FocusEvent e) {
                // DO NOTHING
            }
        });

        /* Close */

        closeButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                creator.setVisible(false);
            }
        });

        // small hack to set the default button
        this.addHierarchyListener(new HierarchyListener() {
            public void hierarchyChanged(HierarchyEvent e) {
                if (getRootPane().getDefaultButton() != closeButton) {
                    getRootPane().setDefaultButton(closeButton);
                }
            }
        });

        // avoid the need to press enter twise in the measurementRotationsField to close the window
        measurementRotationsField.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                closeButton.doClick();
            }
        });
    }

    {
// GUI initializer generated by IntelliJ IDEA GUI Designer
// >>> IMPORTANT!! <<<
// DO NOT EDIT OR ADD ANY CODE HERE!
        $$$setupUI$$$();
    }

    /**
     * Method generated by IntelliJ IDEA GUI Designer
     * >>> IMPORTANT!! <<<
     * DO NOT edit this method OR call it in your code!
     */
    private void $$$setupUI$$$() {
        contentPane = new JPanel();
        contentPane.setLayout(new GridLayoutManager(5, 2, new Insets(11, 11, 11, 11), -1, -1));
        final JPanel panel1 = new JPanel();
        panel1.setLayout(new GridLayoutManager(2, 1, new Insets(0, 4, 4, 4), -1, -1));
        contentPane.add(panel1, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Saved Sequences"));
        final JScrollPane scrollPane1 = new JScrollPane();
        scrollPane1.setVerticalScrollBarPolicy(22);
        panel1.add(scrollPane1, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, null, null, null));
        sequencesTable = new JTable();
        sequencesTable.setPreferredScrollableViewportSize(new Dimension(200, 200));
        scrollPane1.setViewportView(sequencesTable);
        final JPanel panel2 = new JPanel();
        panel2.setLayout(new GridLayoutManager(1, 2, new Insets(0, 0, 0, 0), -1, -1));
        panel1.add(panel2, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        final Spacer spacer1 = new Spacer();
        panel2.add(spacer1, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_WANT_GROW, 1, null, null, null));
        sequencesDeleteButton = new JButton();
        sequencesDeleteButton.setText("Delete Sequence");
        panel2.add(sequencesDeleteButton, new GridConstraints(0, 1, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JPanel panel3 = new JPanel();
        panel3.setLayout(new GridLayoutManager(2, 1, new Insets(0, 4, 4, 4), -1, -1));
        contentPane.add(panel3, new GridConstraints(0, 1, 3, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        panel3.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Default Columns"));
        defaultColumnsPane = new JPanel();
        panel3.add(defaultColumnsPane, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        final Spacer spacer2 = new Spacer();
        panel3.add(spacer2, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_VERTICAL, 1, GridConstraints.SIZEPOLICY_WANT_GROW, null, null, null));
        final JPanel panel4 = new JPanel();
        panel4.setLayout(new GridLayoutManager(1, 1, new Insets(0, 4, 4, 4), -1, -1));
        contentPane.add(panel4, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        panel4.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Measurements"));
        final JPanel panel5 = new JPanel();
        panel5.setLayout(new GridLayoutManager(2, 3, new Insets(0, 0, 0, 0), -1, -1));
        panel4.add(panel5, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        holderCalibrationCombo = new JComboBox();
        panel5.add(holderCalibrationCombo, new GridConstraints(1, 1, 1, 2, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label1 = new JLabel();
        label1.setText("Sample Holder Calibration");
        panel5.add(label1, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label2 = new JLabel();
        label2.setText("Measurement Rotations");
        panel5.add(label2, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        measurementRotationsField = new JFormattedTextField();
        panel5.add(measurementRotationsField, new GridConstraints(0, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(50, -1), null));
        final Spacer spacer3 = new Spacer();
        panel5.add(spacer3, new GridConstraints(0, 2, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_WANT_GROW, 1, null, null, null));
        final JPanel panel6 = new JPanel();
        panel6.setLayout(new GridLayoutManager(1, 2, new Insets(0, 0, 0, 0), -1, -1));
        contentPane.add(panel6, new GridConstraints(4, 0, 1, 2, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        closeButton = new JButton();
        closeButton.setText("Close");
        panel6.add(closeButton, new GridConstraints(0, 1, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final Spacer spacer4 = new Spacer();
        panel6.add(spacer4, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_WANT_GROW, 1, null, null, null));
        final JPanel panel7 = new JPanel();
        panel7.setLayout(new GridLayoutManager(1, 1, new Insets(0, 4, 4, 4), -1, -1));
        contentPane.add(panel7, new GridConstraints(2, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        panel7.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Formatting"));
        final JPanel panel8 = new JPanel();
        panel8.setLayout(new GridLayoutManager(3, 2, new Insets(0, 0, 0, 0), -1, -1));
        panel7.add(panel8, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        final JLabel label3 = new JLabel();
        label3.setText("CSV separator");
        panel8.add(label3, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        csvSeparatorCombo = new JComboBox();
        panel8.add(csvSeparatorCombo, new GridConstraints(0, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label4 = new JLabel();
        label4.setText("Meas. Table Numbers");
        panel8.add(label4, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label5 = new JLabel();
        label5.setText("Meas. Printout Numbers");
        panel8.add(label5, new GridConstraints(2, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        numformatTableCombo = new JComboBox();
        panel8.add(numformatTableCombo, new GridConstraints(1, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        numformatPrintoutCombo = new JComboBox();
        panel8.add(numformatPrintoutCombo, new GridConstraints(2, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JPanel panel9 = new JPanel();
        panel9.setLayout(new GridLayoutManager(1, 1, new Insets(0, 4, 4, 4), -1, -1));
        contentPane.add(panel9, new GridConstraints(3, 0, 1, 2, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        panel9.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Warning Thresholds"));
        final JPanel panel10 = new JPanel();
        panel10.setLayout(new GridLayoutManager(2, 8, new Insets(0, 0, 0, 0), -1, -1));
        panel9.add(panel10, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        final JLabel label6 = new JLabel();
        label6.setText("Signal To Drift (S/D)");
        panel10.add(label6, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label7 = new JLabel();
        label7.setText("Signal To Handler (S/H)");
        panel10.add(label7, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        sigToDriftField = new JFormattedTextField();
        panel10.add(sigToDriftField, new GridConstraints(0, 1, 1, 1, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(50, -1), null));
        sigToHandlerField = new JFormattedTextField();
        panel10.add(sigToHandlerField, new GridConstraints(1, 1, 1, 1, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(50, -1), null));
        final JLabel label8 = new JLabel();
        label8.setText("Signal To Noise (S/N)");
        panel10.add(label8, new GridConstraints(0, 3, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        ThetaLabel = new JLabel();
        ThetaLabel.setText("Theta 63");
        panel10.add(ThetaLabel, new GridConstraints(1, 3, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        sigToNoiseField = new JFormattedTextField();
        panel10.add(sigToNoiseField, new GridConstraints(0, 4, 1, 1, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(50, -1), null));
        theta63Field = new JFormattedTextField();
        panel10.add(theta63Field, new GridConstraints(1, 4, 1, 1, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(50, -1), null));
        final JLabel label9 = new JLabel();
        label9.setText("   ");
        panel10.add(label9, new GridConstraints(0, 2, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label10 = new JLabel();
        label10.setText("Degauss alert limit (mins)");
        panel10.add(label10, new GridConstraints(1, 6, 1, 1, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        degaussTimeLimit = new JFormattedTextField();
        degaussTimeLimit.setText("");
        panel10.add(degaussTimeLimit, new GridConstraints(1, 7, 1, 1, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(50, -1), null));
        final JLabel label11 = new JLabel();
        label11.setText("Signal To Background (S/B)");
        panel10.add(label11, new GridConstraints(0, 6, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        sigToBackgroundField = new JFormattedTextField();
        panel10.add(sigToBackgroundField, new GridConstraints(0, 7, 1, 1, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(50, -1), null));
    }

    private class EditSequencesTableModel extends AbstractTableModel {

        private MeasurementSequence[] sequences = new MeasurementSequence[0];

        public EditSequencesTableModel() {
            updateSequences();
        }

        /**
         * Reads the saved sequences from the settings and fills the table with them. Restores the selected
         * MeasurementSequence if the table contents has changed.
         */
        private void updateSequences() {

            // save selection
            MeasurementSequence selected = null;
            int selectedRow = sequencesTable.getSelectedRow();
            if (selectedRow >= 0) {
                selected = sequences[selectedRow];
            }

            // update table data
            sequences = Settings.getSequences();
            sequencesTable.getSelectionModel().clearSelection();
            fireTableDataChanged();

            // restore selection
            if (selected != null) {
                for (int i = 0; i < sequences.length; i++) {
                    if (sequences[i] == selected) {
                        sequencesTable.getSelectionModel().setSelectionInterval(i, i);
                        break;
                    }
                }
            }
        }

        public int getRowCount() {
            return sequences.length;
        }

        public int getColumnCount() {
            return 1;
        }

        @Override public boolean isCellEditable(int rowIndex, int columnIndex) {
            return true;
        }

        public Object getValueAt(int rowIndex, int columnIndex) {
            return sequences[rowIndex].getName();
        }

        /**
         * Change the name of a saved sequence.
         *
         * @param aValue      a new name for the sequence.
         * @param rowIndex
         * @param columnIndex
         */
        @Override public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
            String name = aValue.toString();
            MeasurementSequence target = sequences[rowIndex];

            if (target.getName().equals(name)) {
                return;     // no change
            }

            // check for a duplicate name
            for (MeasurementSequence sequence : sequences) {
                if (sequence.getName().equals(name)) {
                    return;
                }
            }

            // set the new name
            target.setName(name);
            Settings.fireSequencesModified();
            updateSequences();
        }

        /**
         * Deletes the saved sequence at the speficied row index.
         */
        public void deleteSequence(int rowIndex) {
            Settings.removeSequence(sequences[rowIndex]);
            updateSequences();
        }
    }
}
