/*
 * StereoPlot.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Ikayaki.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package ikayaki.gui;

import ikayaki.MeasurementStep;
import ikayaki.MeasurementValue;
import ikayaki.Project;

import java.awt.*;
import java.awt.geom.Point2D;
import java.util.Vector;


/**
 * Implements stereographic plot
 *
 * @author Aki Sysmlinen
 *
 * Fixed plot to use algorithm client provided.
 *
 * @author Kimmo Holm
 */
public class StereoPlot extends AbstractPlot {

    /**
     * Contains all the data that is shown in this graph.
     */
    private Vector<Point2D> points = new Vector<Point2D>();

    /**
     * Contains information if inclination was positive or negative positive = true, negative = false; 0 is positive
     */
    private Vector<Boolean> incSign = new Vector<Boolean>();

    private Vector<Integer> incSignInt = new Vector<Integer>();

    private Project project = null;

    private Double decValueFoo = null;
    private int incValueFoo = 0;

    /**
     * Adds one measurement step to this graph and converts the data to x- and y- coordinates
     */
    public void add(MeasurementStep step) {
        if (step.getProject() != null) {
            project = step.getProject();
        }
        Double incValue = MeasurementValue.INCLINATION.getValue(step);
        Double decValue = MeasurementValue.DECLINATION.getValue(step);

        if (incValue != null && decValue != null) {
            if (incValue.doubleValue() >= 0) {

                // checks whether equator is crossed. If so, will add another point to the equator.
                if(this.incValueFoo<0){
                    points.add(toXY(calcMiddleDec(decValue, this.decValueFoo), 0.0));
                    incSign.add(new Boolean(false));
                    incSignInt.add(1);
                }
                points.add(toXY(decValue, incValue));
                incSign.add(new Boolean(true));
                incSignInt.add(0);
                this.incValueFoo=1;
            } else {
                if(this.incValueFoo>0){
                    points.add(toXY(calcMiddleDec(decValue, this.decValueFoo), 0.0));
                    incSign.add(new Boolean(true));
                    incSignInt.add(1);
                }
                points.add(toXY(decValue, incValue));
                incSign.add(new Boolean(false));
                incSignInt.add(0);
                this.incValueFoo=-1;
            }

            this.decValueFoo = decValue;
        }
    }

    private Double calcMiddleDec(Double a, Double b){
        if(Math.abs(a-b) > 180)
            return (a+b)/2.0 + 180;
        else
            return (a+b)/2.0;
    }

    /**
     * Does the stereoplot projection for given declination and inclination values.
     *
     * @param decValue declination
     * @param incValue inclination
     * @return a stereoplot projected point in XY-coordinates
     *
     * 3.2006 Fixed plot to use different algorithm -- Kimmo Holm
     */

    private Point2D.Double toXY(Double decValue, Double incValue){
        double SS = 1.0; // radius, needs to be 1.0 to get proper values.
        double CE = SS * Math.tan(Math.PI / 4.0 - Math.abs(incValue)*Math.PI / 360.0);
        double XI = (SS*SS - CE*CE) / (2.0 * CE);
        double RI = Math.sqrt(SS*SS + XI*XI);
        double X2 = XI - RI;
        double XL = X2 * Math.cos((decValue+90.0) * Math.PI / 180.0);
        double YL = -X2 * Math.sin((decValue+90.0) * Math.PI / 180.0);

        return new Point2D.Double(XL, YL);
    }

    /**
     * Resets and repaints this plot.
     */
    public void reset() {
        points.clear();
        incSign.clear();
        incSignInt.clear();
        this.incValueFoo=0;
        repaint();
    }

    /**
     * Returns the number of points in this graph.
     *
     * @return the number of points in this graph
     */
    public int getNumMeasurements() {
        return points.size();
    }

    /**
     * Draws the contents of the plot
     *
     * @param w  Width of the drawable area
     * @param h  Height of the drawable area
     * @param g2 Graphics context
     */
    public void render(int w, int h, Graphics2D g2) {
        // margin
        int m = 5;
        // area for texts on edges
        int txtArea = 10;
        // minimum dimension of the plot
        int dim = Math.min(w, h);
        // area for points in x and y direction = width = height of the actual plot
        int area = dim - (2 * (m + txtArea));
        // font for texts
        g2.setFont(new Font("Arial", Font.PLAIN, 8 + (area / 60)));
        FontMetrics metrics = g2.getFontMetrics();
        // text width for letter W
        int txtW = 0;
        // text height
        int txtH = metrics.getHeight();
        // placement of Origo in textarea. Do not change this.
        int origo = (m + txtArea + area/2);
        // draw circle
        g2.drawOval(m + txtArea, m + txtArea, area, area);
        // draw ticks
        int tLength = 2; // circle ticks length in degrees, originally 5 degrees.
        int atLength = (area / 40) + 1; // horizontal and vertical ticks length

        for (int i = 0; i < 360; i = i + 10) {
            int x1 = new Double(toXY(new Double(i), new Double(0)).getX() * area/2.0).intValue();
            int y1 = new Double(toXY(new Double(i), new Double(0)).getY() * area/2.0).intValue();
            int x2, y2;
            if (i % 90 == 0) {
                x2 = new Double(toXY(new Double(i), new Double(tLength * 2)).getX() * area/2.0).intValue();
                y2 = new Double(toXY(new Double(i), new Double(tLength * 2)).getY() * area/2.0).intValue();
            } else {
                x2 = new Double(toXY(new Double(i), new Double(tLength)).getX() * area/2.0).intValue();
                y2 = new Double(toXY(new Double(i), new Double(tLength)).getY() * area/2.0).intValue();
            }
            g2.drawLine(origo + x1, origo + y1, origo + x2, origo + y2);
        }

        for (int i = 10; i <= 90; i = i + 10) {
            int x1, y1, x2, y2;
            x1 = (new Double(toXY(new Double(0), new Double(i)).getX() * area/2.0).intValue()) - (atLength / 2);
            y1 = new Double(toXY(new Double(0), new Double(i)).getY() * area/2.0).intValue();
            x2 = x1 + atLength;
            y2 = y1;
            g2.drawLine(origo + x1, origo + y1, origo + x2, origo + y2);
        }
        for (int i = 10; i <= 90; i = i + 10) {
            int x1, y1, x2, y2;
            x1 = new Double(toXY(new Double(90), new Double(i)).getX() * area/2.0).intValue();
            y1 = new Double(toXY(new Double(90), new Double(i)).getY() * area/2.0).intValue() - (atLength / 2);
            x2 = x1;
            y2 = y1 + atLength;
            g2.drawLine(origo + x1, origo + y1, origo + x2, origo + y2);
        }
        for (int i = 10; i <= 90; i = i + 10) {
            int x1, y1, x2, y2;
            x1 = (new Double(toXY(new Double(180), new Double(i)).getX() * area/2.0).intValue()) - (atLength / 2);
            y1 = new Double(toXY(new Double(180), new Double(i)).getY() * area/2.0).intValue();
            x2 = x1 + atLength;
            y2 = y1;
            g2.drawLine(origo + x1, origo + y1, origo + x2, origo + y2);
        }
        for (int i = 10; i <= 90; i = i + 10) {
            int x1, y1, x2, y2;
            x1 = new Double(toXY(new Double(270), new Double(i)).getX() * area/2.0).intValue();
            y1 = (new Double(toXY(new Double(270), new Double(i)).getY() * area/2.0).intValue() - (atLength / 2));
            x2 = x1;
            y2 = y1 + atLength;
            g2.drawLine(origo + x1, origo + y1, origo + x2, origo + y2);
        }


        // draw symbols
        txtW = metrics.stringWidth("N");
        g2.drawString("N", m + txtArea + (area / 2) - (txtW / 2), m + (txtArea / 2));
        txtW = metrics.stringWidth("W");
        g2.drawString("W", m + (txtArea / 2) - txtW, m + txtArea + (area / 2) + (txtH / 2));
        txtW = metrics.stringWidth("E");
        g2.drawString("E", m + txtArea + area + (txtArea / 2), m + txtArea + (area / 2) + (txtH / 2));
        txtW = metrics.stringWidth("S");
        g2.drawString("S", m + txtArea + (area / 2) - (txtW / 2), m + txtArea + area + (txtArea / 2) + (txtH / 2));

        // draw points
        int ps = (area / 60) + 4; // points size
        for (int i = 0; i < points.size(); i++) {
            int addX = new Double(points.elementAt(i).getX() * area/2.0).intValue();
            int addY = new Double(points.elementAt(i).getY() * area/2.0).intValue();
            int x = origo + addX;
            int y = origo - addY;
            if (i == 0)
                g2.drawString("NRM", x + (area / 50), y - (area / 50));

            // This first if() is simply to check when we cross equator and skip point drawing for it.
            if(incSignInt.elementAt(i).intValue() != 1){
                if (incSign.elementAt(i).booleanValue()) { // positive inclination
                    g2.fillOval(x - (ps / 2), y - (ps / 2), ps, ps);
                } else { // negative inclination
                    g2.drawOval(x - (ps / 2), y - (ps / 2), ps, ps);
                }
            }
        }

        // draw lines
        if (points.size() >= 2) {
            for (int i = 1; i < points.size(); i++) {

                int x1 = new Double((points.elementAt(i - 1).getX() * area/2.0)).intValue();
                int y1 = new Double((points.elementAt(i - 1).getY() * area/2.0)).intValue();
                int x2 = new Double((points.elementAt(i).getX() * area/2.0)).intValue();
                int y2 = new Double((points.elementAt(i).getY() * area/2.0)).intValue();
                g2.drawLine(origo + x1, origo - y1, origo + x2,origo - y2);
            }
        }
    }
}
