/*
 * Degausser.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Ikayaki.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package ikayaki.squid;

import ikayaki.Settings;

import javax.swing.*;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.Collection;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.ArrayBlockingQueue;
import java.awt.*;
import java.io.StreamTokenizer;
import java.io.Reader;

/**
 * Offers an interface for controlling the degausser (demagnetizer). Because the data link is implemented in the
 * degausser by a single board computer running a small basic program, the response time of the degausser to commands is
 * slow. This class will make sure that commands are not sent faster than the device can handle.
 *
 * @author Aki Korpua
 */
public class Degausser implements SerialIOListener {

    /**
     * buffer for incoming messages, readed when needed. (NEVER USED)
     */
    // private Stack<String> messageBuffer;

    /**
     * Synchronous queue for waiting result message from degausser
     */

    // private SynchronousQueue<String> queue;
    private ArrayBlockingQueue<String> queue;

    // Response wait timeouts used for receiving e.g. command echo (Note: DERC response has other,
    // user-specified timeout)
    private int pollTimeout = 30;
    private TimeUnit pollTimeUnit = TimeUnit.SECONDS;

    /**
     * COM port for communication.
     */
    protected SerialIO serialIO;
    protected String   inputData;

    /**
     * 1-9 seconds default delay 1 second
     */
    private int degausserDelay;

    /**
     * (3, 5, 7, 9) default 3
     */
    private int degausserRamp;

    private double minimumField;
    private double maximumField;

    private volatile boolean demagnetizing = false;

    /**
     * Creates a new degausser interface. Opens connection to degausser COM port (if not open yet) and reads settings
     * from the Setting class.
     */
    public Degausser() throws SerialIOException {
        inputData = "";
        this.serialIO = SerialIO.openPort(new SerialParameters(Settings.getDegausserPort()));
        serialIO.addSerialIOListener(this);
        // messageBuffer = new Stack<String>(); // (NEVER USED)
        // queue = new SynchronousQueue<String>();
        queue = new ArrayBlockingQueue<String>(50, true); // capacity of 50 commands, fair=true
        this.degausserDelay = Settings.getDegausserDelay();
        this.degausserRamp = 3; // Just to be sure that theres always atleast some value. See "writeCorrectRamp" function.
        this.minimumField = Settings.getDegausserMinimumField();
        this.maximumField = Settings.getDegausserMaximumField();

        //needs to call new functions setDelay() and setRamp(). TODO
        //TODO: do we need to check values? (original does)
        try {
            blockingWrite("DCD" + this.degausserDelay);
        } catch (SerialIOException ex1) {
            System.err.println("Error using port in degausser:" + ex1);
        }
    }
    /**
     * Writes ramp rate to degausser according to amplitude.
     */
    public void writeCorrectRamp(double amp) {

        if(amp <= Settings.getDegausserRampLimit(1)) this.degausserRamp = Settings.getDegausserRamp(1);
        else if(amp <= Settings.getDegausserRampLimit(2)) this.degausserRamp = Settings.getDegausserRamp(2);
        else if(amp <= Settings.getDegausserRampLimit(3)) this.degausserRamp = Settings.getDegausserRamp(3);
        else this.degausserRamp = Settings.getDegausserRamp(4);

        try {
            blockingWrite("DCR" + this.degausserRamp);
        } catch (SerialIOException ex1) {
            System.err.println("Error using port in degausser:" + ex1);
        }
    }

    /**
     * Checks which settings have changed and updates the degausser interface. This method will be called by the Squid
     * class.
     */
    public void updateSettings() {
        // No check, only two options. Doesnt matter.
        this.degausserDelay = Settings.getDegausserDelay();
        this.minimumField = Settings.getDegausserMinimumField();
        this.maximumField = Settings.getDegausserMaximumField();
        try {
            blockingWrite("DCD" + this.degausserDelay);
        } catch (SerialIOException ex1) {
            System.err.println("Error using port in degausser:" + ex1);
        }
    }


    /**
     * Sends new ramp rate and delay as received directly in parameters,
     * and not from Settings.
     *
     *
     */
    public boolean updateSettings(int ramprate, int delay) {
        // set ramp
        if (3==ramprate || 5==ramprate || 7==ramprate || 9==ramprate) {
            try {
                blockingWrite("DCD" + ramprate);
            } catch (SerialIOException ex1) {
                System.err.println("Error using port in degausser:" + ex1);
                return false;
            }
        } else {
            return false;
        }
        // set delay
        if (1<=delay && delay<=9) {
            try {
                blockingWrite("DCR" + delay);
            } catch (SerialIOException ex1) {
                System.err.println("Error using port in degausser:" + ex1);
                return false;
            }
        } else {
            return false;
        }
        return true;
    }

    /**
     * Sets coil X,Y,Z.
     *
     * @param coil coil to set on.
     */
    protected void setCoil(char coil) {
        if (coil == 'X' || coil == 'Y' || coil == 'Z') {
            try {
                blockingWrite("DCC" + coil);
            } catch (SerialIOException e) {
                e.printStackTrace();
            }
        } else {
            throw new IllegalArgumentException("coil = " + coil);
        }
    }

    /**
     * Sets amplitude to ramp, range 1.0 to maximumField. A value of 1.0 will actually be rounded to 1.1 which is the
     * actual minimum amplitude of the degausser.
     *
     * @param amplitude amplitude to demag.
     * @throws IllegalArgumentException if the amplitude is not in the allowed range.
     */
    protected void setAmplitude(double amplitude) {
        if (amplitude >= minimumField && amplitude <= maximumField) {
            try {
                String amps = Integer.toString((int) Math.round(amplitude * 10.0));
                while (amps.length() < 4) {
                    amps = "0" + amps;
                }
                blockingWrite("DCA" + amps);
                Thread.sleep(1500);     // needs to wait for the degausser to process the command

            } catch (SerialIOException e) {
                e.printStackTrace();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        } else {
            throw new IllegalArgumentException("amplitude = " + amplitude);
        }
    }

    /**
     * Performs Ramp up. If this is used, make sure you Ramp down in less than 10 seconds because it can damage coil
     *
     * @returns true if command echo was ok
     */
    protected boolean executeRampUp() {
        try {
            return blockingWrite("DERU");
        } catch (SerialIOException e) {
            e.printStackTrace();
            return false;
        }
    }

    /**
     * Brings Ramp down.
     *
     * @returns true if command echo was ok
     */
    protected boolean executeRampDown() {
        try {
            return blockingWrite("DERD");
        } catch (SerialIOException e) {
            e.printStackTrace();
            return false;
        }
    }

    /**
     * Performs Ramp up and down.
     *
     * @returns true if command echo was ok
     */
    protected boolean executeRampCycle() {
        try {
            return blockingWrite("DERC");
        } catch (SerialIOException e) {
            e.printStackTrace();
            return false;
        }
    }


    /**
     * Writes a command to the serial port, waits until command echo is received.
     *
     * @returns true if command echo was ok
     */
    protected boolean blockingWrite(String command) throws SerialIOException {
        // Purge input buffers
        clearCommands();

        // Send new command
        serialIO.writeMessage(command + "\r");
        String answer = "";

        try {
            answer = (String) queue.poll(pollTimeout, pollTimeUnit);
        } catch(InterruptedException ie) {
            System.err.println("Degausser.blockingWrite() : no timely echo response from the degausser!");
            ie.printStackTrace(); // (maybe a bit exaggerated... ;-)
            return false;
        }

        if (!command.equals(answer)) {
//                for (int i = 0; i < answer.length(); i++) {
//                    System.out.println((int) answer.charAt(i));
//                }
//                throw new IllegalArgumentException("sent: " + command + " recieved: " + answer);
                System.err.println("Degausser.blockingWrite() wrong echo, sent: " + command + " received: " + answer);
            return false;
        }
        return true;
    }


    /**
     * Initiates a ramp cycle and waits for it to complete, checks
     * degausser return code. Ramps down and repeats the cycle once
     * if the degausser responds with blank lines (known bug).
     * Semi-blocking up to 2 * Settings.getDegaussTimeLimit() minutes.
     */
    protected boolean demagnetizeCycle() {
        String answer;
        int attempts = 0;
        int timeoutMinutes = Settings.getDegaussTimeLimit();
        //TODO: might want to use configurable max attempts limit
        int maxAttempts = Settings.getDegausserMaxReattempts(); // default: 2
        boolean timeoutExceeded;

        demagnetizing = true;

        while(attempts<maxAttempts) {

            attempts++;

            // start the degauss cycle, check also that command echo was ok
            timeoutExceeded = !executeRampCycle();

            // wait for success/fail response (DONE or TRACK ERROR)
            try {
                answer = null;
                if(!timeoutExceeded) {
                    answer = queue.poll(timeoutMinutes*60, TimeUnit.SECONDS);
                }
            } catch (InterruptedException e) {
                System.err.println("Degausser.demagnetizeCycle() : user-specified timeout exceeded for degausser DONE/TRACK ERROR response");
                e.printStackTrace();
                timeoutExceeded = true;
                answer = null;
            }

            if (!timeoutExceeded && answer!=null && answer.length()>=("DONE".length())) {

                // proper length degausser response, return True if DONE,
                // False if TRACK ERROR or other long string
                demagnetizing = false;
                return (answer.equals("DONE"));

             } else {

                // Response is somehow incorrect (e.g., blank) or there was no response at all
                // Solution II to degausser blank responses

                // readDegausserStatus(); // (for debugging, to get DSS into the serial.log)

                // commission ramp down, then initiate the same cycle again
                timeoutExceeded = !executeRampDown();

                // valid answers: Z if success, ZERO ERROR if fail
                // wait max 1 minute for an answer
                try {
                    if (!timeoutExceeded) {
                        answer = queue.poll(60, TimeUnit.SECONDS);
                    } else {
                        System.err.println("Degausser.executeRampDown() : no response! ");
                    }
                } catch (InterruptedException e) {
                    System.err.println("Degausser.demagnetizeCycle() : no response! ");
                    e.printStackTrace();
                    timeoutExceeded = true;
                }
                if (null==answer || timeoutExceeded) {
                    // no response at all from degausser, exit while() and alert now
                    break;
                } else {
                    // there was "some" response, retry with one new ramp cycle
                }
            }
        }

        // Reattempt count exceeded or no emergency ramp-down response from degausser
        // Show warning message box and start alerting
        demagnetizing = false;
        class Annoyer extends Thread {
             public volatile boolean doExit;
             public void run() {
                 doExit = false;
                 while(!doExit) {
                     try { sleep(800); }
                     catch (InterruptedException e) { return; }
                     Toolkit.getDefaultToolkit().beep();
                 }
             }
         }
         Annoyer annoyer = new Annoyer();
         annoyer.start();
         JOptionPane.showMessageDialog(null,
             "Degausser did not respond properly within the " + timeoutMinutes + " minute timeout.\r\n" +
                     "Number of degauss attempts was: " + attempts,
             "Degausser may be overheating", JOptionPane.ERROR_MESSAGE
         );
        annoyer.doExit = true;

        return false;
    }

    /**
     * Performs full sequence to demagnetize Z coil with the given amplitude. Blocking method.
     *
     * @param amp amplitude to demag.
     * @return true if process was sended succesfully, otherwise false.
     */
    public boolean demagnetizeZ(double amp) {
        setAmplitude(amp);
        writeCorrectRamp(amp);
        setCoil('Z');
        return demagnetizeCycle();
    }

    /**
     * Performs full sequence to demagnetize Y coil with the given amplitude. Blocking method.
     *
     * @param amp amplitude to demag.
     * @return true if process was sended succesfully, otherwise false.
     */
    public boolean demagnetizeY(double amp) {
        setAmplitude(amp);
        writeCorrectRamp(amp);
        setCoil('Y');
        return demagnetizeCycle();
    }


    /**
     * Performs full sequence to demagnetize X coil with the given amplitude. Blocking method.
     *
     * @param amp amplitude to demag.
     * @return true if process was sended succesfully, otherwise false.
     */
    public boolean demagnetizeX(double amp) {
        setAmplitude(amp);
        writeCorrectRamp(amp);
        setCoil('X');
        return demagnetizeCycle();
    }


    /**
     * Returns current internal state about demagnetization process
     *
     * @return true if demagnetizing process is ongoing
     */
    public boolean isDemagnetizing() {
        return demagnetizing;
    }

    /**
     * Sends status query to degausser and returns answer. Blocking with longish timeout.
     */
    protected String readDegausserStatus() {
        String answer = null;
        boolean timeout = false;
        try {
            timeout = !blockingWrite("DSS");
        } catch (SerialIOException e) {
            e.printStackTrace();
            timeout = true;
        }
        try {
            if(!timeout) {
                answer = (String) queue.poll(pollTimeout, pollTimeUnit);
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return answer;
    }

    /**
     * Sends status query to degausser and returns answer. Blocking.
     *
     * @return Z=Zero, T=Tracking, ?=Unknown
     */
    public char getRampStatus() {
        String answer = readDegausserStatus();
        if (null==answer) {
            return '?';
        }
        return answer.charAt(1);
    }

    /**
     * Sends ramp query to degausser and returns answer. Blocking.
     *
     * @return 3, 5, 7 or 9
     */
    public int getRamp() {
        String answer = readDegausserStatus();
        if (null==answer) {
            return 3;
        }
        return (int) answer.charAt(4);
    }

    /**
     * Sends delay query to degausser and returns answer. Blocking.
     *
     * @return 1 to 9 as seconds
     */
    public int getDelay() {
        String answer = readDegausserStatus();
        if (null==answer) {
            return 1;
        }
        return (int) answer.charAt(7);
    }

    /**
     * Sends coil query to degausser and returns answer. Blocking.
     *
     * @return X=X Axis, Y=Y Axis, Z=Z Axis, ?=Unknown
     */
    public char getCoil() {
        String answer = readDegausserStatus();
        if (null==answer) {
            return '?';
        }
        return answer.charAt(10);
    }

    /**
     * Sends amplitude query to degausser and returns answer. Blocking.
     *
     * @return 0 to 3000
     */
    public int getAmplitude() {
        String answer = readDegausserStatus();
        if (null==answer) {
            return 0;
        }
        return Integer.parseInt(answer.substring(13, 17));
    }

    /**
     * Checks if connection is ok.
     *
     * @return true if ok.
     */
    public boolean isOK() {
        return (this.serialIO != null);
    }



    /**
     * Clears the internal buffer of received commands
     *
     */
    private void clearCommands() {
        int count = queue.size();
        if ( count != 0 ) {
            System.err.println("Degausser.clearCommands() - removing " + count + " unprocessed commands from queue.");
            for(int i=0; i<count; i++) {
                try {
                    System.err.println("Command " + i + " was : " + queue.poll(1, TimeUnit.MICROSECONDS));
                } catch (InterruptedException ie) {
                    // do nothing
                }
            }
        }
        queue.clear();
    }


    /**
     * Accepts incoming data from a SerialIO class
     *
     */
    public void serialIOEvent(SerialIOEvent event) {
        String str = "";

        // The SerialIO class already encapsulates the port's InputStream so
        // something as simple as using StreamTokenizer won't work :|

        // The degausser seems to use 0x0D 0x0A = CR LF as the separator in replys
        // so we use stringtokenizer on that

        inputData = inputData.concat(event.getMessage());

        if (inputData.contains("\r\n")) { // any complete command there yet?

            // yeah very efficient...

            boolean clear_after_dispatch = inputData.endsWith("\r\n");

            StringTokenizer strTok = new StringTokenizer(inputData, "\r\n", false); // don't return delimiters

            while(strTok.hasMoreTokens()) {
                str = strTok.nextToken();
                try {
                    queue.put(str);
                } catch (InterruptedException e) {
                    System.err.println("Interrupted Degausser message event");
                } catch (NullPointerException e) {
                    System.err.println("Null from SerialEvent in Degausser");
                }
            }

            if ( !clear_after_dispatch ) {
                inputData = str; // keep the incomplete command that was at the end
            } else {
                inputData = "";  // string ended with CRLF, nothing to keep
            }

        }
    }

    public void OLDserialIOEvent(SerialIOEvent event) {
        //TODO: problem when Degausser and Magnetometer uses same port :/
        String message = event.getCleanMessage();
        /*
        // -- solution I to degausser blank responses
        //    (not used yet as no info from 2G Enterprises yet whether keepalives
        //     are correct response behaviour or if firmware just hangs up)
        //
        if (null == message) {
            return;
        }
        // detect degausser's keepalive CR+LF blank line responses,
        // add only actual responses to the message queue
        int untrimmedLength = message.length();
        message = ((message.replace('\r', ' ')).replace('\n', ' ')).trim();
        if (0 != message.length()) {
            if (waitingForMessage) {
                try {
                    queue.put(message);
                } catch (InterruptedException e) {
                    System.err.println("Interrupted Degausser message event");
                } catch (NullPointerException e) {
                    System.err.println("Null from SerialEvent in Degausser");
                }
            }
            messageBuffer.add(message);
        } else {
            if (untrimmedLength>0) { // blank line or keepalive
                // queue.refreshTimeout() -- ?
            }
        }
    }
    */
    if (message != null) {
            //if (waitingForMessage) {
                try {
                    queue.put(message);
                } catch (InterruptedException e) {
                    System.err.println("Interrupted Degausser message event");
                } catch (NullPointerException e) {
                    System.err.println("Null from SerialEvent in Degausser");
                }
            //} else {
            //    System.err.println("Discarded unawaited degausser serial message:\t" + message);
            //}
            // messageBuffer.add(message); // (NEVER USED)
            // unawaited answers will be lost
        }
    }
}
