/*
 * SquidEmulator.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Ikayaki.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package ikayaki.squid;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.concurrent.SynchronousQueue;

/**
 * This class tries to emulate behavior of real squid-system. It starts 3 threads (handler,magnetometer,degausser),
 * opens COM-ports for them and adds SerialIO Listeners. Threads generates random data values or loaded values as
 * results and generates random error situations to see that program using real squid system does survive those. Uses
 * 2-3 COM ports. Usage SquidEmulator x z.. filename where x is 0 or 1 and indicates if Magnetometer and Demagnetizer
 * are on same COM port. z... values are COM ports (Handler,Magnetometer,Degausser). filename is name of log file we are
 * using or it is existing log file, which is used to generate same sequence used to verify that old and new program
 * behaves same way.
 *
 * @author Aki Korpua
 * @author Rami Jrvinen
 */
public class SquidEmulator {
    public SquidEmulator() {
        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /*
    Event A: On New IO Message - reads message and puts it in Buffer
    */

    /**
     * indicates if system have been started
     */
    //private static boolean online;

    /**
     * log file we are using read or write
     */
    private static File logFile;

    /**
     * starts Threads which reads messages from selected COM port. Own listener for each. Offers write commads to port
     * too.
     */
    private static SerialIO handlerPort;
    private static SerialIO magnetometerPort;
    private static SerialIO degausserPort;

    private static FileWriter logWriter = null;

    private static HandlerEmu handler;
    private static MagnetometerEmu magnetometer;
    private static DegausserEmu degausser;
    private static boolean running;

    /**
     * send message to SerialIO to be sented.
     *
     * @param message any message reply we are sending back
     * @param port    port number to be sent
     */
    public static void writeMessage(String message, SerialIO port) {
        //writes log if indicated to do so
        try {
            logWriter.write("SEND:" + message + "\n");
            logWriter.flush();
        } catch (IOException e) {
            System.err.println("Error on writing log file");
        }
        try {
            port.writeMessage(message);
        } catch (SerialIOException ex) {
            System.err.println(ex);
        }
        return;
    }

    /**
     * First creates or loads log file and sets settings. Runs sequence where read data from buffer and run cheduled
     * actions (move, rotate, demag, measure) and send feedback to COM ports.
     */
    public static void main(String[] args) {
        System.out.println("Starting...");
        try {
            handlerPort = SerialIO.openPort(new SerialParameters("COM"+args[0]));
            magnetometerPort = SerialIO.openPort(new SerialParameters("COM"+args[1]));
            degausserPort = SerialIO.openPort(new SerialParameters("COM"+args[2]));

            logFile = new File(args[3]);
            logWriter = new FileWriter(logFile);
        } catch (SerialIOException e) {
            System.out.println(e);
            return;
        } catch (Exception e) {
            System.out.println(
                    "Usage \"java SquidEmulator HandlerPort MagnetometerPort DegausserPort Log_filename\"");
            return;
        }

        running = true;

        handler = new HandlerEmu();
        handler.start();
        magnetometer = new MagnetometerEmu();
        magnetometer.start();
        degausser = new DegausserEmu();
        degausser.start();

        System.out.println("System running...");
        try {
            logWriter.write("SESSION STARTED:\n");
            logWriter.flush();
        } catch (IOException ex1) {
            System.out.println("Writing error");
        }

        try {
            //wait for signal to quit 8)
            System.in.read();
        } catch (IOException ex) {
        }

        SerialIO.closeAllPorts();
        try {
            logWriter.close();
        } catch (IOException ex2) {
        }

        running = false;

        return;
    }

    private void jbInit() throws Exception {
    }

    /**
     * Runs handler emulation process. Process incoming messages and sends data back. When message comes, process it
     * (wait if needed for a while), updates own status and sends result back.
     */
    private static class HandlerEmu extends Thread implements SerialIOListener {

        /* axis selection: 0 = translation, 1 = rotation */
        private int handlerMode;

        /* maximum speed of motor */
        private int velocity;

        /* motor acceleration and deceleration parameters (not used in actual simulation process) */
        private int acceleration;
        private int deceleration;

        /* programmed motor translation or rotation distance (used when counting execution delays) */
        private int commandedDistance;

        /* true = translation into rock magnetometer or clockwise sample rotation
           false = translation out of rock magnetometer or counterclockwise sample rotation */
        private boolean motorDirection;

        /* no need to be up to date with current handler position or rotation
        private int currentRotation;
        private int currentPosition;
        */

        /* true = move until reaching hard limit stop, false = move to programmed distance */
        private boolean continuousMovement;

        /* simulated delay of handler movement */
        private int sleeptime;

        /* queue for received commands */
        private SynchronousQueue<String> answerQueue;

        public HandlerEmu() {
            handlerPort.addSerialIOListener(this);
            answerQueue = new SynchronousQueue<String>();
        }

        public void run() {

            String command = "";

            // set default values
            handlerMode = 0;
            acceleration = 5;
            deceleration = 10;
            velocity = 50;

            while (running) {

              try {
                command = (String) answerQueue.take();
              } catch (InterruptedException e) {
                  e.printStackTrace();
              }

               if (command.startsWith("@0", 0)) {
                 System.out.println("ikayaki said @0");
               }

               else if (command.startsWith("O", 0)) {
                   // axis selection (translation / rotation)
                   if (command.startsWith("1,0", 1)) {
                       // translation mode (DEFAULT)
                       System.out.println("set translation mode");
                       handlerMode = 0;
                   }
                   else if (command.startsWith("1,1", 1)) {
                       System.out.println("set rotation mode");
                       // rotation mode
                       handlerMode = 1;
                   }
               }

               else if (command.startsWith("M", 0)) {
                   // set velocity (50-12000)
                   velocity = Integer.parseInt(command.substring(1, command.length()-1));
                   System.out.println("velocity set to: " + velocity);
               }

               else if (command.startsWith("A", 0)) {
                   // set acceleration (0-127) (default 5) (usually 20-50)
                   acceleration = Integer.parseInt(command.substring(1, command.length()-1));
                   System.out.println("accel set to: "+ acceleration);
               }

               else if (command.startsWith("D", 0)) {
                   // set deceleration (default 10)
                   deceleration = Integer.parseInt(command.substring(1, command.length()-1));
                   System.out.println("deceleration set to: " + deceleration);
               }

               else if (command.startsWith("N", 0)) {
                   // set translation or rotation distance to be traveled
                   // 1 - 16777215 (numbers divisible by 10)
                   // full scale travel 42000 counts (half stepping)
                   // sample cycle or rotation 2000 counts (half stepping)
                   commandedDistance = Integer.parseInt(command.substring(1));

                   // debug..
                   if(commandedDistance < 1) {
                       System.out.println("commanded distance below 1! setting it to 1.");
                       commandedDistance = 1;
                   }

                   System.out.println(" set distance: " + commandedDistance);
               }

               else if (command.charAt(0) == '+') {
                   // motor direction positive
                   motorDirection = true;
               }
               else if (command.charAt(0) == '-') {
                   // motor direction negative
                   motorDirection = false;
               }

               else if (command.startsWith("S", 0)) {
                   // start continuous motion using acceleration and velocity parameter
                   continuousMovement = true;

                   if(handlerMode == 0) {
                       try {
                           System.out.println("executing S command");
                           sleep(5500);
                           System.out.println(" done");
                       } catch(InterruptedException ex1) {
                       }
                   }
               }

               else if (command.startsWith("Q", 0)) {
                   // stop motion using deceleration parameter

                   // should be able to receive this and react when motor is running!!
               }

               else if (command.startsWith("G", 0)) {  // +G -G
                   continuousMovement = false;
                   // programmed movement for a given counts spesified by the N parameter

                   // ....count and execute delay
                   if(handlerMode == 0) {
                       try {
                           //System.out.println("commandedDistance: " + commandedDistance + " currentPosition: " + currentPosition);
                           sleeptime = commandedDistance / velocity * 1000;
                           //System.out.println("sleeping" + sleeptime + "secs");
                           System.out.println("translating");
                           sleep(sleeptime);
                           System.out.println("translation complete");

                           /*
                           if(motorDirection)
                               currentPosition += commandedDistance;
                           else
                               currentPosition -= commandedDistance;
                           */

                       } catch(InterruptedException ex1) {
                       }
                   }
                   else {
                       try {
                          sleeptime = commandedDistance / velocity * 51;  // 51 is just a good guess
                          System.out.println("rotating");
                          sleep(sleeptime);
                          System.out.println("rotate complete");
                       } catch(InterruptedException ex1) {
                       }
                   }

               }

               else if (command.startsWith("H1", 0)) {
                   continuousMovement = false;
                   if(handlerMode == 0) {
                       try {
                           System.out.println("handler translating/rotating to home position");
                           sleep(6000);
                           System.out.println(" done");
                       } catch(InterruptedException ex1) { }
                   }
               }

               // verification of parameters (never used in Ikayaki?)
               else if (command.startsWith("V", 0)) {
                   if (command.charAt(1) == 'P') {
                       // tell position (internal counter)
                       // emulator does not calculate exact internal counter changes
                       //writeMessage(Integer.toString(currentPosition), handlerPort);
                   }
                   else if (command.charAt(1) == 'N') {
                       // tell programmed move distance
                       //writeMessage(Integer.toString(commandedDistance), handlerPort);
                   }
                   else if (command.charAt(1) == 'A') {
                       // tell acceleration parameter
                       //writeMessage(Integer.toString(acceleration), handlerPort);
                   }
                   else if (command.charAt(1) == 'M') {
                       // tell velocity parameter
                       //writeMessage(Integer.toString(velocity), handlerPort);
                   }
               }

               else if (command.startsWith("F%", 0)) {
                   // F - wait for motor to be idle and the wait timer to be
                   //     zero before accepting any further inputs
                   // % - poll device for waiting messages (is never used as single command in Ikayaki)

                   // assuming that we receive only motor movement related queries

                   if (continuousMovement)
                       writeMessage("7", handlerPort); // hard limit stop
                   else
                       writeMessage("5", handlerPort); // end of move notice
               }

            }
            try {
                Thread.sleep(100);
            } catch (InterruptedException ex) {
            }

        }

        protected void waitForMessage() throws SerialIOException {
            try {
                answerQueue.take();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

        public void serialIOEvent(SerialIOEvent event) {
            int i;
            try {
                logWriter.write("HANDLER_RECEIVE:" + event.getCleanMessage() + "\n");
                logWriter.flush();
            } catch (IOException ex) {
            }

            try {
                answerQueue.put(event.getCleanMessage());
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }


    /**
     * Runs magnetometer emulation process. Process incoming messages and send data back
     * when appropriate. Uses random values for sample measurements.
     */
    private static class MagnetometerEmu extends Thread implements SerialIOListener {

        // queue for received commands
        private SynchronousQueue<String> answerQueue;
        private String command = "";

        public MagnetometerEmu() {
            magnetometerPort.addSerialIOListener(this);
            answerQueue = new SynchronousQueue<String>();
        }

        public void run() {
            while (running) {

               try {
                  command = (String) answerQueue.take();
               } catch (InterruptedException e) {
                     e.printStackTrace();
               }

               // send back the analog data
               if (command.startsWith("SD", 1)) {
                   // sending plain double values
                   // -> defined otherwise in protocol, but ikayaki accepts
                   Double response = Math.random() * 0.0001;
                   if ((Math.random() *10) < 3)
                       response = -response;
                   writeMessage(Double.toString(response), magnetometerPort);
                   System.out.println("magnetometer: sent values");
               }

               // send back the counter value
               else if (command.startsWith("SC", 1)) {
                   Double response = Math.random() * 0.000001;
                   writeMessage(Double.toString(response), magnetometerPort);
                   System.out.println("magnetometer: sent values");
               }

               else {
                   // other commands don't need responses or are not used in Ikayaki
               }

            }
            try {
                Thread.sleep(100);
            } catch (InterruptedException ex) {
            }
            //}

        }

        protected void waitForMessage() throws SerialIOException {
            try {
                answerQueue.take();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

        public void serialIOEvent(SerialIOEvent event) {
            try {
                logWriter.write("MAGNETOMETER_RECEIVE:" + event.getCleanMessage() + "\n");
                logWriter.flush();
            } catch (IOException ex) {
            }

            try {
                String message = event.getCleanMessage();
                String[] commands = message.split("/r");
                //we only accept first "part". And only if it starts with A,X,Y,Z
                if (commands[0].charAt(0) == 'A' || commands[0].charAt(0) == 'X' || commands[0].charAt(0) == 'Y' || commands[0].charAt(
                       0) == 'Z') {
                answerQueue.put(commands[0]);
                }
            } catch(InterruptedException e) {
                e.printStackTrace();
            }
        }
    }


    /**
     * Runs degausser emulation process. Process incoming messages and sends data back. When message comes, process it
     * (wait if needed for a while), updates own status and sends result back.
     */
    private static class DegausserEmu
            extends Thread implements SerialIOListener {

        // queue for received commands
        private SynchronousQueue<String> answerQueue;
        String command = "";

        // sleeping delay for a ramp cycle
        private int cycleDelay;

        // degaussing amplitude (used for computing cycle delay
        private int degausserAmplitude;

        // active coil
        private int degausserCoil;

        // following variables have no operational relevance in emulation
        private int degausserDelay;
        private int degausserRamp;
        private char degausserStatus;

        public DegausserEmu() {
            degausserPort.addSerialIOListener(this);

            answerQueue = new SynchronousQueue<String>();
            degausserRamp = 3;
            degausserAmplitude = 0;
            degausserDelay = 1;
            degausserCoil = 2;
        }

        public void run() {
            while (running) {

                try {
                    command = (String) answerQueue.take();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

                // delay (what does this do?)
                if (command.startsWith("DCD", 0)) {
                    writeMessage(command, degausserPort);
                    degausserDelay = Integer.parseInt(command.substring(3));
                    System.out.println("degausser: delay set to "+degausserDelay);
                }

                // set ramp rate (3,5,7,9) default 3
                else if (command.startsWith("DCR", 0)) {
                    writeMessage(command, degausserPort);
                    degausserRamp = Integer.parseInt(command.substring(3));
                    System.out.println("degausser: ramp rate set to "+degausserRamp);
                }

                // set amplitude 0000-3000
                else if (command.startsWith("DCA", 0)) {
                    writeMessage(command, degausserPort);
                    degausserAmplitude = Integer.parseInt(command.substring(3));
                    System.out.println("degausser: amplitude set to "+degausserAmplitude);
                }

                // set coil axis DCCX, DCCY, DCCZ
                else if (command.startsWith("DCC", 0)) {
                    writeMessage(command, degausserPort);
                    if(command.charAt(3) == 'X') degausserCoil = 0;
                    else if(command.charAt(3) == 'Y') degausserCoil = 1;
                    else if(command.charAt(3) == 'Z') degausserCoil = 2;
                    System.out.println("degausser: "+ command.charAt(3) + " axis selected");
                }

                // execute ramp cycle
                else if (command.startsWith("DERC", 0)) {
                    degausserStatus = 'T';
                    writeMessage(command, degausserPort);
                    System.out.println("degausser: executing ramp cycle");
                    try {
                        if(degausserAmplitude < 300) {
                            cycleDelay = degausserAmplitude * 29;
                        }
                        else {
                            cycleDelay = degausserAmplitude * 38;
                        }
                        Thread.sleep(cycleDelay); // ramp cycle delay
                    } catch (InterruptedException ex) {
                    }
                    degausserStatus = 'Z';
                    writeMessage("DONE", degausserPort);
                    //writeMessage("TRACK ERROR", degausserPort); if cycle is unsuccessful
                    System.out.println(" done");
                }

                // execute Ramp Up (never used in ikayaki)
                else if (command.startsWith("DERU", 0)) {
                    degausserStatus = 'T';
                    writeMessage("T", degausserPort);
                    // writeMessage("TRACK ERROR", degausserPort); if tracking does not occur
                }

                // execute Ramp Down (never used in ikayaki)
                else if (command.startsWith("DERD", 0)) {
                    degausserStatus = 'Z';
                    writeMessage("Z", degausserPort); //
                    // writeMessage("ZERO ERROR", degausserPort); if ramp down unsuccessful
                }

                // Send Status (never used?)
                else if (command.startsWith("DSS", 0)) {
                }

            }

            try {
                Thread.sleep(100);
            } catch (InterruptedException ex) {
            }

        }

        protected void waitForMessage() throws SerialIOException {
            try {
                answerQueue.take();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

        public void serialIOEvent(SerialIOEvent event) {
            try {
                logWriter.write("MAGNETOMETER_RECEIVE:" + event.getCleanMessage() + "\n");
                logWriter.flush();
            } catch (IOException ex) {
            }

            try {
                String message = event.getCleanMessage();
                String[] commands = message.split("/r");
                //we only accept first "part". And only if it starts with D
                if (commands[0].charAt(0) == 'D') {
                  answerQueue.put(commands[0]);
                }
            } catch(InterruptedException e) {
                e.printStackTrace();
            }
        }

    }

}
