/*
 * ComponentCalculation.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis;

import javax.vecmath.Vector3d;
import javax.vecmath.Vector2d;
import java.util.Vector;
import java.util.Random;

import Jama.*;

/**
 * This class is used to calculate the properties of magnetic components.
 */

public final class ComponentCalculation {

    private ComponentCalculation() {
    }

    /**
     * Returns maximum measurement amplitude of assigned steps.
     *
     * @param steps component steps
     * @return maximum amplitude
     */
    public static double calculateMaxIntensity(Vector steps) {

        if (null == steps || steps.size() <= 0) {
            return 9999;
        }

        int count = steps.size();
        double stepvalue = 0.0;

        MeasurementStep step;
        for (int i = 0; i < count; i++) {
            step = (MeasurementStep) steps.get(i);
            if (step.getStepValue() > stepvalue) {
                stepvalue = step.getStepValue();
            }
        }

        return stepvalue;
    }

    /**
     * Returns minimum measurement amplitude of assigned steps.
     *
     * @param steps component steps
     * @return minimum amplitude
     */
    public static double calculateMinIntensity(Vector steps) {

        if (null == steps || steps.size() <= 0) {
            return -9999;
        }

        int count = steps.size();
        double stepvalue = ((MeasurementStep) steps.get(0)).getStepValue();

        MeasurementStep step;
        for (int i = 0; i < count; i++) {
            step = (MeasurementStep) steps.get(i);
            if (step.getStepValue() < stepvalue) {
                stepvalue = step.getStepValue();
            }
        }

        return stepvalue;
    }

    /**
     * Returns average intensity of steps included in the component.
     *
     * @param steps component steps
     * @return average intensity
     */

    public static double calculateAvgIntensity(Vector steps) {

        if (null == steps || steps.size() <= 0) {
            return 0.0;
        }

        int count = steps.size();
        double sumintensity = 0.0;
        double avgintensity;

        MeasurementStep step;

        for (int i = 0; i < count; i++) {
            step = (MeasurementStep) steps.get(i);
            sumintensity += MeasurementValue.MAGNETIZATION.getValue(step);
        }

        avgintensity = sumintensity / count;

        return avgintensity;
    }

    /**
     * Private method for calculating the center point (average) of
     * the steps' magnetic vectors.
     * <p/>
     * Returns a Vector3d pointing to the center of mass.
     *
     * @param osteps component steps
     * @return vector to center of mass, or null on errors
     */
    private static Vector3d calculateCenterOfMass(Vector osteps) {

        //Vector steps = getStepsWithoutOrigo(osteps);
        Vector steps = osteps;
        MeasurementStep step;

        int totalsteps = steps.size();
        double xsum, ysum, zsum;

        xsum = 0.0;
        ysum = 0.0;
        zsum = 0.0;

        if (!containsOrigo(steps)) {
            for (int i = 0; i < totalsteps; i++) {
                step = (MeasurementStep) steps.get(i);

                if (null == step) {
                    return null;
                }

                xsum += MeasurementValue.ZIDJERVELD_X_NORMALIZED.getValue(step);
                ysum += MeasurementValue.ZIDJERVELD_Y_NORMALIZED.getValue(step);
                zsum += MeasurementValue.ZIDJERVELD_Z_NORMALIZED.getValue(step);
            }
        }
        return new Vector3d(xsum / totalsteps, ysum / totalsteps, zsum / totalsteps);
    }

    /**
     * Private method for calculating values needed in declination and inclination calculations.
     * <p/>
     * Returns array of l,m,n values and eigenvalue variances.
     *
     * @param osteps component steps
     * @return l, m, n values and eigenvalue variances in double array
     */
    private static double[] calculate(Vector osteps) {

        //Vector steps = getStepsWithoutOrigo(osteps);
        Vector steps = osteps;
        MeasurementStep step;

        int totalsteps = steps.size();

        double xvalue, yvalue, zvalue;

        double[][] biased_steps = new double[totalsteps][3];

        for (int i = 0; i < totalsteps; i++) {
            step = (MeasurementStep) steps.get(i);

            if (null == step) {
                return null;
            }

            xvalue = MeasurementValue.ZIDJERVELD_X_NORMALIZED.getValue(step);
            biased_steps[i][0] = xvalue;

            yvalue = MeasurementValue.ZIDJERVELD_Y_NORMALIZED.getValue(step);
            biased_steps[i][1] = yvalue;

            zvalue = MeasurementValue.ZIDJERVELD_Z_NORMALIZED.getValue(step);
            biased_steps[i][2] = zvalue;
        }

        double[] center_of_masses = new double[3];

        if (!containsOrigo(steps)) {
            for (int i = 0; i < 3; i++) {
                double sum = 0;
                for (int j = 0; j < biased_steps.length; j++) {
                    sum += biased_steps[j][i];
                }
                center_of_masses[i] = sum / biased_steps.length;
            }
        } else {
            for (int i = 0; i < 3; i++) {
                center_of_masses[i] = 0.0;
            }
        }

        double[][] repmat = new double[biased_steps.length][3];

        for (int i = 0; i < biased_steps.length; i++) {
            for (int j = 0; j < 3; j++) {
                repmat[i][j] = center_of_masses[j];
            }
        }

        Matrix biased = new Matrix(biased_steps);
        Matrix repmatrix = new Matrix(repmat);
        Matrix unbiased_steps = biased.minus(repmatrix);
        Matrix tensor = unbiased_steps.transpose().times(unbiased_steps);

        EigenvalueDecomposition e = new EigenvalueDecomposition(tensor);
        Matrix eigenMatrix = e.getV();

        double[] eigenvalues = e.getRealEigenvalues();
        int eigenvectorcolumn = 0;

        for (int i = 1; i < eigenvalues.length; i++) {
            if (eigenvalues[i] > eigenvalues[i - 1]) {
                eigenvectorcolumn = i;
            }
        }

        double[][] eigenVectors = eigenMatrix.getArrayCopy();

        double[] lmn = new double[6];  // includes l,m,n and eigenvalue variances

        // get values from the principal eigenvector
        lmn[0] = eigenVectors[0][eigenvectorcolumn]; // l (x)
        lmn[1] = eigenVectors[1][eigenvectorcolumn]; // m (y)
        lmn[2] = eigenVectors[2][eigenvectorcolumn]; // n (z)

        // get eigenvalue variances
        //TODO: ill-formed matrix can have negative eigenvalues, maybe do this:
        //for(int i=0; i<3; i++) {
        //    if(eigenvalues[i]<=0) { eigenvalues[i]=1.0; }
        //}

        //lmn[3] = Math.sqrt(eigenvalues[0]); // v3 variance
        //lmn[4] = Math.sqrt(eigenvalues[1]); // v2 variance
        lmn[3] = eigenvalues[0];
        lmn[4] = eigenvalues[1];
        lmn[5] = eigenvalues[2]; // v1 variance

        return lmn;
    }

    /**
     * Returns component declination.
     *
     * @param steps component steps
     * @return declination
     */
    public static double calculateDeclination(Vector steps) {

        if (null == steps || steps.size() <= 0) {
            return 0.0;
        }

        double[] lmn = calculate(steps);
        if (null == lmn) {
            return 0;
        }

        double l = lmn[0];
        double m = lmn[1];
        //double n = lmn[2];

        double declination = 0.0;

        if (l < 0) {
            declination = Math.toDegrees(Math.atan(m / l)) + 180;
        } else if (m > 0) {
            declination = Math.toDegrees(Math.atan(m / l));
        } else {
            declination = Math.toDegrees(Math.atan(m / l)) + 360;
        }

        return declination;
    }

    /**
     * Returns component inclination.
     *
     * @param steps component steps
     * @return inclination
     */
    public static double calculateInclination(Vector steps) {

        if (null == steps || steps.size() <= 0) {
            return 0.0;
        }

        double[] lmn = calculate(steps);
        if (null == lmn) {
            return 0;
        }

        double n = lmn[2];

        return Math.toDegrees(Math.asin(n));

        //Random RNG = new Random();
        //return RNG.nextGaussian() * 90.0;
    }

    /**
     * Returns angle of error (Maximum Angular Deviation) of component.
     *
     * @param steps component steps
     * @return angle of error
     */
    public static double calculateAoE(Vector steps) {

        if (null == steps || steps.size() <= 0) {
            return 0.0;
        }

        double[] lmn = calculate(steps);
        if (null == lmn) {
            return 0;
        }

        // test
        //System.out.println("v3: " + lmn[3] + " v2: " + lmn[4] + " v1: "+lmn[5]);

        double v3 = lmn[3];
        double v2 = lmn[4];
        double v1 = lmn[5];

        // return Math.toDegrees(Math.atan(Math.sqrt(Math.pow(v2, 2) + Math.pow(v3, 2)) / v1));
        return Math.toDegrees(Math.atan(Math.sqrt(v2 + v3)) / Math.sqrt(v1));
    }

    /**
     * Returns ancient pole latitude.
     *
     * @param steps   component steps
     * @param project the project
     * @return ancient pole latitude
     */
    public static double calculatePoleLat(Vector steps, Project project) {

        if (null == steps || steps.size() <= 0) {
            return 0.0;
        }

        double declination = calculateDeclination(steps);
        double inclination = calculateInclination(steps);
        double magnetic_colatitude, sinPlat, firstif, secondif;

        if (project.getProperty(Project.LATITUDE_PROPERTY).equals("") ||
                project.getProperty(Project.LATITUDE_PROPERTY) == null) {
            return 0.0;
        }

        double SITE_LATITUDE = Double.parseDouble(project.getProperty((Project.LATITUDE_PROPERTY)));

        /* try not to divide be zero */
        if (inclination == 0) {
            return 0.0;
        }

        double magnetic_colat = Math.toDegrees(Math.atan(2 / Math.tan(Math.toRadians(inclination))));

        if (magnetic_colat < 0)
            magnetic_colatitude = magnetic_colat + 180;
        else
            magnetic_colatitude = magnetic_colat;

        sinPlat = (Math.sin(Math.toRadians(SITE_LATITUDE)) * Math.cos(Math.toRadians(magnetic_colatitude))) + (Math.cos(Math.toRadians(SITE_LATITUDE)) * Math.sin(Math.toRadians(magnetic_colatitude)) * Math.cos(Math.toRadians(declination)));

        if (sinPlat < -0.99)
            firstif = -0.988;
        else
            firstif = sinPlat;

        if (firstif > 0.99)
            secondif = 0.988;
        else
            secondif = sinPlat;

        return Math.toDegrees(Math.atan(secondif / Math.sqrt(-secondif * secondif + 1)));
    }

    /**
     * Returns ancient pole longitude.
     *
     * @param steps   component steps
     * @param project the project
     * @return ancient pole longitude
     */
    public static double calculatePoleLong(Vector steps, Project project) {

        if (null == steps || steps.size() <= 0) {
            return 0.0;
        }

        double declination = calculateDeclination(steps);
        double inclination = calculateInclination(steps);

        if (project.getProperty(Project.LONGITUDE_PROPERTY).equals("") ||
                project.getProperty(Project.LONGITUDE_PROPERTY) == null) {
            return 0.0;
        }

        double SITE_LONGITUDE = Double.parseDouble(project.getProperty(Project.LONGITUDE_PROPERTY));

        if (project.getProperty(Project.LATITUDE_PROPERTY).equals("") ||
                project.getProperty(Project.LATITUDE_PROPERTY) == null) {
            return 0.0;
        }

        double SITE_LATITUDE = Double.parseDouble(project.getProperty(Project.LATITUDE_PROPERTY));
        double POLE_LATITUDE = calculatePoleLat(steps, project);
        double tmpvalue, tmpvalue2, beta, sinB, firstif, secondif, magnetic_colatitude;

        /* try not to divide be zero */
        if (inclination == 0 || POLE_LATITUDE == 0) {
            return 0.0;
        }

        double magnetic_colatitude_tmp = Math.toDegrees(Math.atan(2 / Math.tan(Math.toRadians(inclination))));

        if (magnetic_colatitude_tmp < 0)
            magnetic_colatitude = magnetic_colatitude_tmp + 180;
        else
            magnetic_colatitude = magnetic_colatitude_tmp;

        sinB = Math.sin(Math.toRadians(magnetic_colatitude)) * Math.sin(Math.toRadians(declination)) / Math.cos(Math.toRadians(POLE_LATITUDE));

        if (sinB > 0.99) {
            tmpvalue = 0.99;
        } else {
            tmpvalue = sinB;
        }

        if (tmpvalue < -0.99) {
            tmpvalue2 = -0.988;
        } else {
            tmpvalue2 = tmpvalue;
        }

        beta = Math.toDegrees(Math.atan(tmpvalue2 / Math.sqrt((-tmpvalue2 * tmpvalue2) + 1)));

        if (Math.cos(Math.toRadians(magnetic_colatitude)) < Math.sin(Math.toRadians(SITE_LATITUDE)) * Math.sin(Math.toRadians(POLE_LATITUDE)))
        {
            firstif = SITE_LONGITUDE + 180 - beta;
        } else {
            firstif = SITE_LONGITUDE + beta;
        }

        if (firstif > 360) {
            secondif = firstif - 360;
        } else {
            secondif = firstif;
        }

        if (secondif < 0) {
            return secondif + 360;
        } else {
            return secondif;
        }

    }

    /**
     * Returns the 3D starting point for drawing a line
     * of the magnetic component. The line starting point is
     * the farthest point into the same/positive direction
     * of the component vector.
     * <p/>
     * To get the point for plotting etc, use Session.getStartAndEndpoints
     * which adds some origo-step inclusion logic.
     *
     * @param steps
     * @return start point
     */
    public static Vector3d calculateStart(Vector steps) {
        return getLineProjectionEndpoint(steps, true);
    }

    /**
     * Returns the 3D end point for drawing a line
     * of the magnetic component. The line end point is
     * the farthest point into the opposite/negative direction
     * of the component vector.
     * <p/>
     * To get the point for plotting etc, use Session.getStartAndEndpoints
     * which adds some origo-step inclusion logic.
     *
     * @param steps
     * @return end point
     */
    public static Vector3d calculateEnd(Vector steps) {
        return getLineProjectionEndpoint(steps, false);
    }


    /**
     * Does some vector maths to project Zplot-normalised
     * measurement step values onto the 3D component vector,
     * then returns either start or end point.
     * <p/>
     * To get the point for plotting etc, use Session.getStartAndEndpoints
     * which adds some origo-step inclusion logic.
     *
     * @param osteps the steps, startNotEnd set to true if
     *               starting point should be returned
     * @return start or end point, null on error
     */
    private static Vector3d getLineProjectionEndpoint(Vector osteps, boolean startNotEnd) {

        //Vector steps = getStepsWithoutOrigo(osteps);
        Vector steps = osteps;

        MeasurementStep step;
        int totalsteps = steps.size();

        // get the center of mass CM
        Vector3d CM = calculateCenterOfMass(steps);
        if (null == CM) {
            return null;
        }

        // get the magnetic component direction from l,m,n values
        double[] tmp = calculate(steps);
        if (null == tmp) {
            return null;
        }
        Vector3d ComponentDirection = new Vector3d(tmp[0], tmp[1], tmp[2]);
        ComponentDirection.normalize();

        // get distances from CM along component
        Vector3d p = new Vector3d();            // vector of current step's point
        Vector3d tmpVect = new Vector3d();      // temporary vector for maths
        double d = 0.0;                         // distance of current point
        double extr = 0.0;                      // farthest distance found till now

        for (int i = 0; i < totalsteps; i++) {

            step = (MeasurementStep) steps.get(i);
            p.set(MeasurementValue.ZIDJERVELD_X_NORMALIZED.getValue(step),
                    MeasurementValue.ZIDJERVELD_Y_NORMALIZED.getValue(step),
                    MeasurementValue.ZIDJERVELD_Z_NORMALIZED.getValue(step)
            );

            // d = (CM - P) * Dir
            tmpVect.set(CM);
            tmpVect.sub(p);
            d = ComponentDirection.dot(tmpVect);

            // keep track of extreme
            if ((startNotEnd && d >= extr) ||
                    (!startNotEnd && d <= extr)) {
                extr = d;
            }
        }

        // return r = d_extr * Dir + CM
        Vector3d resultPoint = new Vector3d();
        resultPoint.scaleAdd(-extr, ComponentDirection, CM);
        return resultPoint;
    }


    public enum Axis { X, Y, Z };

    /**
     * Returns the 2D starting point for drawing a line
     * of the magnetic component. The line starting point is
     * the farthest point into the same/positive direction
     * of the component vector.
     *
     * TODO: very much incomplete and untested, though luckily also unused :)
     *
     * @param steps the measurement steps
     *              viewedFromAxis e.g. Axis.X, direction of view
     * @return start point
     */
    public static Vector2d calculateStart2D(Vector steps, Axis viewedFromAxis) {
        return getLineProjectionEndpoint2D(steps, true, viewedFromAxis);
    }

    /**
     * Returns the 2D end point for drawing a line
     * of the magnetic component. The line end point is
     * the farthest point into the opposite/negative direction
     * of the component vector.
     *
     * TODO: very much incomplete and untested, though luckily also unused :)
     *
     * @param steps the measurement steps
     *              viewedFromAxis e.g. Axis.X, direction of view
     * @return end point
     */
    public static Vector2d calculateEnd2D(Vector steps, Axis viewedFromAxis) {
        //if (containsOrigo(steps)) {   -- must be done in ZijderveldPlot plotting class!
        //    return new Vector2d(0.0, 0.0);
        //}
        return getLineProjectionEndpoint2D(steps, false, viewedFromAxis);
    }


    /**
     * Does some vector maths to project Zplot-normalised
     * measurement step values onto the component vector,
     * as viewed from the specified axis, and
     * then returns the 2D start or end point.
     *
     * TODO: very much incomplete and untested, though luckily also unused :)
     *
     * @param osteps the steps, startNotEnd set to true if
     *               starting point should be returned
     * @return start or end point, null on error
     */
    private static Vector2d getLineProjectionEndpoint2D(Vector osteps, boolean startNotEnd, Axis viewedFromAxis) {

        Vector steps = getStepsWithoutOrigo(osteps);

        MeasurementStep step;
        int totalsteps = steps.size();

        // based on the direction of view (i.e. along which axis), select the 2D point x1 and x2 coordinate
        // indices into the 3D point {x,y,z} (e.g., x1_idx=0,x2_idx=2 => 2D points = {x,z})
        MeasurementValue<Double> MVx1, MVx2;
        int x1_idx, x2_idx;
        double rotate;
        if (Axis.X == viewedFromAxis) {
            MVx1 = MeasurementValue.ZIDJERVELD_Y_NORMALIZED;
            x1_idx = 1;
            MVx2 = MeasurementValue.ZIDJERVELD_Z_NORMALIZED;
            x2_idx = 2;
            rotate = -90;
        } else if (Axis.Y == viewedFromAxis) {
            MVx1 = MeasurementValue.ZIDJERVELD_X_NORMALIZED;
            x1_idx = 0;
            MVx2 = MeasurementValue.ZIDJERVELD_Z_NORMALIZED;
            x2_idx = 2;
        } else {
            MVx1 = MeasurementValue.ZIDJERVELD_X_NORMALIZED;
            x1_idx = 0;
            MVx2 = MeasurementValue.ZIDJERVELD_Y_NORMALIZED;
            x2_idx = 1;
        }

        // calculate center of mass CM
        double x1sum, x2sum;
        x1sum = 0.0;
        x2sum = 0.0;
        if (!containsOrigo(steps)) {
            for (int i = 0; i < totalsteps; i++) {
                step = (MeasurementStep) steps.get(i);
                if (null == step) {
                    return null;
                }
                x1sum += MVx1.getValue(step);
                x2sum += MVx2.getValue(step);
            }
        }
        Vector2d CM = new Vector2d(x1sum / totalsteps, x2sum / totalsteps);

        // get the magnetic component direction from l,m,n values, "project" into 2D
        double[] tmp = calculate(steps);
        if (null == tmp) {
            return null;
        }
        Vector2d ComponentDir2D = new Vector2d(tmp[x1_idx], tmp[x2_idx]);
        ComponentDir2D.normalize();

        // get distances from CM along component
        Vector2d p = new Vector2d();            // vector of current step's point
        Vector2d tmpVect = new Vector2d();      // temporary vector for maths
        double d = 0.0;                         // distance of current point
        double extr = 0.0;                      // farthest distance found till now

        for (int i = 0; i < totalsteps; i++) {

            step = (MeasurementStep) steps.get(i);

            p.set(MVx1.getValue(step), MVx2.getValue(step));

            // d = (CM - P) * Dir
            tmpVect.set(CM);
            tmpVect.sub(p);
            d = ComponentDir2D.dot(tmpVect);

            // keep track of extreme
            if ((startNotEnd && d >= extr) ||
                    (!startNotEnd && d <= extr)) {
                extr = d;
            }
        }

        // return r = d_extr * Dir + CM
        Vector2d resultPoint = new Vector2d();
        resultPoint.scaleAdd(-extr, ComponentDir2D, CM);
        return resultPoint;
    }

    /**
     * Calculates the 3D start and end point of the
     * component, then returns the length of the
     * interconnecting line.
     *
     * TODO: it is still unclear whether this length has any physical meaning
     *
     * @param steps
     * @return line length
     */
    public static double calculateLength(Vector steps) {
        Vector3d start = calculateStart(steps);
        Vector3d end = calculateEnd(steps);
        if (null == start || null == end) {
            return 0.0;
        } else {
            start.sub(end);
            return start.length();
        }
    }


    /**
     * Returns true if the steps contain an origo step at
     * either the start or end.
     *
     * @param steps
     * @return true if start or end is origo
     */
    public static boolean containsOrigo(Vector steps) {
        if (steps == null || steps.size() <= 0) {
            return false;
        } else if (((MeasurementStep) steps.get(0)).getStepValue() == MeasurementStep.ORIGO_STEP ||
                ((MeasurementStep) steps.get(steps.size() - 1)).getStepValue() == MeasurementStep.ORIGO_STEP) {
            return true;
        }
        return false;
    }

    /**
     * Returns true if the steps contain an origo step at
     * either the start or end.
     *
     * @param steps MeasurementStep steps in a vector
     * @return vector with the input steps sans any origo steps
     */
    public static Vector getStepsWithoutOrigo(Vector steps) {
        if (steps == null) {
            return null;
        }
        Vector newSteps = (Vector) steps.clone();
        int i = 0;
        while (i < newSteps.size()) {
            if (((MeasurementStep) newSteps.get(i)).getStepValue() == MeasurementStep.ORIGO_STEP) {
                newSteps.remove(i);
                continue;
            } else {
                i++;
            }

        }
        return newSteps;
    }

}
