/*
 * MagneticComponents.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis;

import java.util.Vector;

/**
 * This class is used to store the steps of magnetic components and the antipole values
 * and to retrieve these steps and values.
 */

public class MagneticComponents {

    private Session session = null;
    private Vector<Vector<MeasurementStep>> steps = null;
    private Vector<int[]> stepNumbers = null;
    private Vector<Boolean> componentIsAntipole = null;

    public MagneticComponents(Session session) {
        this.session = session;
        steps = new Vector<Vector<MeasurementStep>>();
        stepNumbers = new Vector<int[]>();
        componentIsAntipole = new Vector<Boolean>();
    }

    /**
     * Adds a new collection of MeasurementSteps.
     *
     * @param steps A vector that contains the MeasurementSteps that should
     *              <p/>
     *              be considered as part of the added magnetic component.
     * @return true if successful
     */
    public synchronized boolean addComponent(Vector<MeasurementStep> steps) {
        boolean result;
        if (null == steps) {
            result = false;
        } else {
            this.steps.add(steps);
            result = true;
        }
        return result;
    }


    /**
     * Call to add the component's step indices to those as in the
     * <p/>
     * measurement sequence table
     */
    public synchronized boolean addStepNumbers(int[] stepnumbers) {
        boolean result;
        if (null == stepNumbers) {
            result = false;
        } else {
            stepNumbers.add(stepnumbers);
            result = true;
        }
        return result;
    }

    /**
     * Removes one component i.e. collection of MeasurementSteps. At the same time,
     * it also removes the associated step numbers.
     *
     * @param component Which magnetic component to remove (0 is the first).
     * @return true if successful
     */
    public synchronized boolean removeComponent(int component) {
        boolean result;
        if (null == steps || component < 0 || component >= steps.size()) {
            result = false;
        } else {
            steps.remove(component);
            if(component<stepNumbers.size()) {
                stepNumbers.remove(component);
            }
            result = true;
        }
        return result;
    }


    /**
     * Returns the collection of MeasurementSteps associated with
     * <p/>
     * the specified magnetic component.
     *
     * @param component Which magnetic component to return (0 is the first).
     * @return null if fails
     */
    public synchronized Vector<MeasurementStep> getSteps(int component) {
        Vector<MeasurementStep> vect;
        if (null == steps || component < 0 || component >= steps.size()) {
            vect = null;
        } else {
            vect = steps.get(component);
        }
        return vect;
    }

    /**
     * Returns an int[] of the component's step indices in the
     * <p/>
     * measurement sequence table
     */
    public synchronized int[] getStepNumbers(int component) {
        int[] stepnums;
        if (null == stepNumbers || component < 0 || component >= stepNumbers.size()) {
            stepnums = null;
        } else {
            stepnums = stepNumbers.get(component);
        }
        return stepnums;
    }


    /**
     * Returns the obvious.
     */
    public synchronized int getNumOfComponents() {
        if (null == steps) {
            return 0;
        } else {
            return steps.size();
        }
    }

    /**
     * Sets given component to be antipole or not.
     */
    public synchronized void setAntipole(int component, boolean isAntipole) {
        if (null == steps || component < 0 || component >= steps.size()) {
            return;
        }
        // Because steps, stepnumbers and antipole settings are kept in
        // different Vector's, need to adjust size to be equal or larger:
        if (componentIsAntipole.size() < steps.size()) {
            componentIsAntipole.setSize(steps.size());
        }
        componentIsAntipole.set(component, isAntipole);
    }

    /**
     * Returns if given component is antipole or not.
     */
    public synchronized Boolean getAntipole(int component) {
        if (null == steps || component < 0 || component >= steps.size()) {
            return false;
        }
        // Because steps, stepnumbers and antipole settings are kept in
        // different Vector's, need to adjust size to be equal or larger:
        if (componentIsAntipole.size() < steps.size()) {
            componentIsAntipole.setSize(steps.size());
        }
        if (null == componentIsAntipole.get(component)) {
            componentIsAntipole.set(component, false);
            return false;
        } else {
            return componentIsAntipole.get(component);
        }
    }

}
