/*
 * MeasurementResult.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

/**
 * A set of X, Y and Z values measured by the magnetometer. The raw XYZ values will be rotated in 3D space by using a
 * transformation matrix. The project will set and update the transformation whenever its parameters are changed.
 * <p/>
 * All units are mA/m.
 *
 * @author Esko Luontola
 */
public class MeasurementResult {

    /**
     * The type of this result.
     */
    private final Type type;

    /**
     * The rotation that the sample holder was in when this result was measured. The value is in range 0..360 degrees.
     */
    private final int rotation;

    /**
     * The unmodified measurements recieved from the squid. Will not change after it has been once set.
     */
    private final Vector3d rawVector = new Vector3d();

    /**
     * The measurements in sample coordinates. Has the rotation, noise and holder fixes applied to itself.
     */
    private final Vector3d sampleVector = new Vector3d();

    /**
     * The measurements in geographic coordinates. Equals the sample coordinates with the transformation matrix
     * applied.
     */
    private final Vector3d geographicVector = new Vector3d();

    /**
     * Creates a new measurement result. All units are mA/m.
     * <p/>
     * The sample and geographic coordinates are NOT set when a MeasurementResult is created.
     *
     * @param type     the type (background or rotation) of this result.
     * @param rotation the rotation of the sample holder in degrees (0..360).
     * @param x        the measured X coordinate value.
     * @param y        the measured Y coordinate value.
     * @param z        the measured Z coordinate value.
     * @throws NullPointerException     if type is null.
     * @throws IllegalArgumentException if the type is NOISE or HOLDER, but rotation is non-zero.
     */
    public MeasurementResult(Type type, int rotation, double x, double y, double z) {
        if (type == null) {
            throw new NullPointerException();
        }
        if ((type == Type.NOISE || type == Type.HOLDER) && rotation != 0) {
            throw new IllegalArgumentException("type = " + type + ", rotation = " + rotation);
        }
        this.type = type;
        this.rotation = rotation % 360;
        rawVector.set(x, y, z);
        applyFixes(null);
        setTransform(null);
    }

    /**
     * Creates a measurement result from the specified element.
     * <p/>
     * The sample and geographic coordinates are NOT set when a MeasurementResult is created.
     *
     * @param element the element from which this result will be created.
     * @throws NullPointerException     if element is null.
     * @throws IllegalArgumentException if the element was not in the right format.
     */
    public MeasurementResult(Element element) {
        if (element == null) {
            throw new NullPointerException();
        }

        // verify tag name
        if (!element.getTagName().equals("result")) {
            throw new IllegalArgumentException("Invalid tag name: " + element.getTagName());
        }

        // get type
        String s = element.getAttribute("type");
        try {
            this.type = Type.valueOf(s);
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("Invalid type: " + s, e);
        }

        // get rotation
        s = element.getAttribute("rotation");
        try {
            int rotation = Integer.parseInt(s) % 360;
            while (rotation < 0) {
                rotation += 360;
            }
            this.rotation = rotation;
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException("Invalid rotation: " + s, e);
        }

        // verify type and rotation
        if ((type == Type.NOISE || type == Type.HOLDER) && rotation != 0) {
            throw new IllegalArgumentException("type = " + type + ", rotation = " + rotation);
        }

        // get x, y, z
        try {
            rawVector.set(Double.parseDouble(element.getAttribute("x")),
                    Double.parseDouble(element.getAttribute("y")),
                    Double.parseDouble(element.getAttribute("z")));
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException("Invalid x, y or z: " + e.getMessage(), e);
        }

        // initialize sampleVector and geographicVector
        applyFixes(null);
        setTransform(null);
    }

    /**
     * Exports this result to a DOM element.
     *
     * @param document the document that will contain this element.
     */
    public Element getElement(Document document) {
        Element element = document.createElement("result");

        element.setAttribute("type", type.name());
        element.setAttribute("rotation", Integer.toString(rotation));
        element.setAttribute("x", Double.toString(rawVector.x));
        element.setAttribute("y", Double.toString(rawVector.y));
        element.setAttribute("z", Double.toString(rawVector.z));

        return element;
    }

    /**
     * Applies the holder, noise and rotation fixes and saves the results as the sample vector. Resets the geographic
     * vector to a copy of the sample vector. This method must be called before setTransform().
     *
     * @param step the measurement step that includes the holder and noise calibration values. If null, the holder and
     *             noise fixes are not applied.
     */
    protected void applyFixes(MeasurementStep step) {
        Vector3d holder, noise;

        sampleVector.set(rawVector);

        /* Use the old method for all but multi-rotation gyro AF */

        if ( step != null ) {
            holder = step.getHolder();  // single holder vector
            noise = step.getNoise();    // average of all noise vectors, noise always measured at 0deg rotation
        }

        if ( step == null
            || (step != null
                  // !(multirot-gyro-af)
                && !(step.getProject().getType()==Project.Type.GAF && step.getMeasurementsCount()>5)
                )
            )  {

            // non-translated noise, average value of backgrounds
            if ( step != null && step.getProject().getType() != Project.Type.GAF) {
                // apply the noise fix (for GAF this is done later below)
                noise = step.getNoise();
                sampleVector.sub(noise);
            }

            // apply rotation fix
            sampleVector.set(toTranslatedVector(sampleVector, rotation));

            if (step != null) {
                // apply the +/-Z fix
                if (step.getProject() != null && step.getProject().getOrientation() == Project.Orientation.MINUS_Z) {
                    sampleVector.set(sampleVector.x, -sampleVector.y, -sampleVector.z);
                }
                // apply the non-translated sample holder fix
                holder = step.getHolder();     // will be zero, if this project is the holder calibration project
                sampleVector.sub(holder);
            }

        } else {

            /* Use a new method for multi-rotation gyro AF */

            // really nasty manual matrix calcs ahead...
            // Note: this ONLY works for subrotations of 90deg multiples (i.e. 0, 90, 180, 270) - however, these
            // are apparently the only rotations used currently, so it should not be a problem
            Double[] dBgKx_beg = step.getDoubleNoiseVectorN(1);
            Double[] dBgKy_beg = step.getDoubleNoiseVectorN(2);
            Double[] dBgKz_beg = step.getDoubleNoiseVectorN(3);
            Double[] dBgKx_end = step.getDoubleNoiseVectorN(4);
            Double[] dBgKy_end = step.getDoubleNoiseVectorN(5);
            Double[] dBgKz_end = step.getDoubleNoiseVectorN(6);

            double xtrue, ytrue, ztrue;

            // Correct with x, y noise at four angles, still using 'unfixed'/untranslated raw sample vector
            // The holder is subtracted after the 'fix'
            switch (rotation) {
                case 0:
                    xtrue = sampleVector.x - 0.5*(dBgKx_beg[0]+dBgKx_end[0]); //checked
                    ytrue = sampleVector.y - 0.5*(dBgKx_beg[1]+dBgKx_end[1]); //checked
                    break;
                case 90:
                    xtrue = sampleVector.x - 0.5*(dBgKx_beg[1]+dBgKx_end[1]); //checked
                    ytrue = sampleVector.y - 0.5*(dBgKy_beg[0]+dBgKy_end[0]); //checked
                    break;
                case 180:
                    xtrue = sampleVector.x - 0.5*(dBgKx_beg[0]+dBgKx_end[0]); //checked
                    ytrue = sampleVector.y - 0.5*(dBgKx_beg[1]+dBgKx_end[1]); //checked
                    break;
                case 270:
                    xtrue = sampleVector.x - 0.5*(dBgKx_beg[1]+dBgKx_end[1]); //checked
                    ytrue = sampleVector.y - 0.5*(dBgKy_beg[0]+dBgKy_end[0]); //checked
                    break;
                default:
                    assert false;
                    throw new IllegalStateException("rotation = " + rotation);
            }
            // z has always the same noise correction
            ztrue = sampleVector.z - 0.5*(dBgKz_beg[2] + dBgKz_end[2]);

            // apply all components
            sampleVector.set(xtrue, ytrue, ztrue);

            // apply rotation fix
            sampleVector.set(toTranslatedVector(sampleVector, rotation));

            if (step != null) {
                // apply the +/-Z fix
                if (step.getProject() != null && step.getProject().getOrientation() == Project.Orientation.MINUS_Z) {
                    sampleVector.set(sampleVector.x, -sampleVector.y, -sampleVector.z);
                }
                // apply the non-translated sample holder fix
                holder = step.getHolder();     // will be zero, if this project is the holder calibration project
                sampleVector.sub(holder);
            }

        }

        // reset geographic vector
        setTransform(null);
    }


    /**
     * Applies rotational translation to the input vector, i.e. preserves z but
     * corrects the sign of x and y and also swaps them out if necessary.
     *
     * @param sampleVector as the input vector, rotat as the raw data rotation in degrees
     */
    protected Vector3d toTranslatedVector(Vector3d sampleVector, int rotat) throws IllegalStateException {

        // apply rotation fix
        //          @x sensor  @y sensonr  @z sensor
        //   0deg   a0    (x)  a90   (y)   at(z)
        //  90deg   a90   (y)  a180 (-x)   at(z)
        // 180deg   a180 (-x)  a270 (-y)   at(z)
        // 270deg   a270 (-y)  a0    (x)   at(z)
        if (rotat % 90 == 0) {
            // accurate and fast algorithm for trivial angles
            switch (rotat) {
            case 0:
                // NO NEED TO ROTATE
                break;
            case 90:
                sampleVector.set(sampleVector.y, -sampleVector.x, sampleVector.z);
                break;
            case 180:
                sampleVector.set(-sampleVector.x, -sampleVector.y, sampleVector.z);
                break;
            case 270:
                sampleVector.set(-sampleVector.y, sampleVector.x, sampleVector.z);
                break;
            default:
                assert false;
                throw new IllegalStateException("rotation = " + rotation);
            }
        } else {
            // rotate all non-trivial angles by using a matrix
            Matrix3d rotate = new Matrix3d();
            rotate.rotZ(Math.toRadians(-rotation));
            rotate.transform(sampleVector);
        }

        return sampleVector;
    }

    /**
     * Applies a transformation matrix to the sample vector and saves the results as the geographic vector. This method
     * must be called after applyFixes().
     *
     * @param transform the matrix to be applied. If null, will assume identity matrix.
     */
    protected void setTransform(Matrix3d transform) {
        if (transform != null) {
            transform.transform(sampleVector, geographicVector);
        } else {
            geographicVector.set(sampleVector);
        }
    }

    /**
     * Returns the type of this result.
     */
    public Type getType() {
        return type;
    }

    /**
     * Returns the rotation of this result. The value is in range 0..360 degrees.
     */
    public int getRotation() {
        return rotation;
    }

    /**
     * Returns the noise fixed, rotated and transformed X coordinate of this result. The value is in geographic
     * coordinates.
     *
     * @throws IllegalStateException if this result's type is not SAMPLE, in which case it should make no sense to call
     *                               this method.
     */
    public double getGeographicX() {
        if (type != Type.SAMPLE) {
            throw new IllegalStateException();
        }
        return geographicVector.x;
    }

    /**
     * Returns the noise fixed, rotated and transformed Y coordinate of this result. The value is in geographic
     * coordinates.
     *
     * @throws IllegalStateException if this result's type is not SAMPLE, in which case it should make no sense to call
     *                               this method.
     */
    public double getGeographicY() {
        if (type != Type.SAMPLE) {
            throw new IllegalStateException();
        }
        return geographicVector.y;
    }

    /**
     * Returns the noise fixed, rotated and transformed Z coordinate of this result. The value is in geographic
     * coordinates.
     *
     * @throws IllegalStateException if this result's type is not SAMPLE, in which case it should make no sense to call
     *                               this method.
     */
    public double getGeographicZ() {
        if (type != Type.SAMPLE) {
            throw new IllegalStateException();
        }
        return geographicVector.z;
    }

    /**
     * Returns a pointer to the geographic vector. WARNING! No modification to the returned object should be made. They
     * should be done on a copy of the object than the object itself.
     */
    protected Vector3d getGeographicVector() {
        return geographicVector;
    }

    /**
     * Returns the noise fixed and rotated X coordinate of this result. The value is in sample coordinates.
     */
    public double getSampleX() {
        return sampleVector.x;
    }

    /**
     * Returns the noise fixed and rotated Y coordinate of this result. The value is in sample coordinates.
     */
    public double getSampleY() {
        return sampleVector.y;
    }

    /**
     * Returns the noise fixed and rotated Z coordinate of this result. The value is in sample coordinates.
     */
    public double getSampleZ() {
        return sampleVector.z;
    }

    /**
     * Returns a pointer to the sample vector. WARNING! No modification to the returned object should be made. They
     * should be done on a copy of the object than the object itself.
     */
    protected Vector3d getSampleVector() {
        return sampleVector;
    }

    /**
     * Returns the unmodified X coordinate of this result. The value is in magnetometer coordinates.
     */
    public double getRawX() {
        return rawVector.x;
    }

    /**
     * Returns the unmodified Y coordinate of this result. The value is in magnetometer coordinates.
     */
    public double getRawY() {
        return rawVector.y;
    }

    /**
     * Returns the unmodified Z coordinate of this result. The value is in magnetometer coordinates.
     */
    public double getRawZ() {
        return rawVector.z;
    }

    /**
     * Returns a pointer to the raw vector. WARNING! No modification to the returned object should be made. They should
     * be done on a copy of the object than the object itself.
     */
    protected Vector3d getRawVector() {
        return rawVector;
    }

    public enum Type {
        SAMPLE, HOLDER, NOISE
    }
}