/*
 * MeasurementValue.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis;

import static mavis.MeasurementResult.Type.*;
import static java.lang.Math.atan2;
import static java.lang.Math.PI;
import static java.lang.Math.atan;
import static java.lang.Math.pow;
import static java.lang.Math.sqrt;
import static java.lang.Math.sin;
import static java.lang.Math.cos;
import java.lang.Math;

/**
 * Algorithms for calculating values from the measurements. A MeasurementValue object will be passed to the getValue()
 * method of a project to retrieve the desired value.
 *
 * @author Esko Luontola
 */
public abstract class MeasurementValue <T> {

    /**
     * Calculates the average of all X components in geographic coordinates.
     */
    public static final MeasurementValue<Double> GEOGRAPHIC_X =
            new MeasurementValue<Double>("X'", "Am\u00B2", "Average X (geographic coordinates)") {
                public Double getValue(MeasurementStep step) {
                    if ( MeasurementStep.ORIGO_STEP==step.getStepValue() ) {
                        return 0.0;
                    }
                    double sum = 0.0;
                    int count = 0;
                    for (int i = 0; i < step.getResults(); i++) {
                        MeasurementResult r = step.getResult(i);
                        if (r.getType() != SAMPLE) {
                            continue;
                        }
                        sum += r.getGeographicX();
                        count++;
                    }
                    if (count > 0) {
                        return new Double(sum / count);
                    } else {
                        return null;
                    }
                }
            };

    /**
     * Calculates the average of all Y components in geographic coordinates.
     */
    public static final MeasurementValue<Double> GEOGRAPHIC_Y =
            new MeasurementValue<Double>("Y'", "Am\u00B2", "Average Y (geographic coordinates)") {
                public Double getValue(MeasurementStep step) {
                    if ( MeasurementStep.ORIGO_STEP==step.getStepValue() ) {
                        return 0.0;
                    }
                    double sum = 0.0;
                    int count = 0;
                    for (int i = 0; i < step.getResults(); i++) {
                        MeasurementResult r = step.getResult(i);
                        if (r.getType() != SAMPLE) {
                            continue;
                        }
                        sum += r.getGeographicY();
                        count++;
                    }
                    if (count > 0) {
                        return new Double(sum / count);
                    } else {
                        return null;
                    }
                }
            };

    /**
     * Calculates the average of all Z components in geographic coordinates.
     */
    public static final MeasurementValue<Double> GEOGRAPHIC_Z =
            new MeasurementValue<Double>("Z'", "Am\u00B2", "Average Z (geographic coordinates)") {
                public Double getValue(MeasurementStep step) {
                    if ( MeasurementStep.ORIGO_STEP==step.getStepValue() ) {
                        return 0.0;
                    }
                    double sum = 0.0;
                    int count = 0;
                    for (int i = 0; i < step.getResults(); i++) {
                        MeasurementResult r = step.getResult(i);
                        if (r.getType() != SAMPLE) {
                            continue;
                        }
                        sum += r.getGeographicZ();
                        count++;
                    }
                    if (count > 0) {
                        return new Double(sum / count);
                    } else {
                        return null;
                    }
                }
            };

    /**
     * Calculates the average of all X components in sample coordinates.
     */
    public static final MeasurementValue<Double> SAMPLE_X =
            new MeasurementValue<Double>("X", "Am\u00B2", "Average X (sample coordinates)") {
                public Double getValue(MeasurementStep step) {
                    if ( MeasurementStep.ORIGO_STEP==step.getStepValue() ) {
                        return 0.0;
                    }
                    double sum = 0.0;
                    int count = 0;
                    for (int i = 0; i < step.getResults(); i++) {
                        MeasurementResult r = step.getResult(i);
                        if (r.getType() != SAMPLE) {
                            continue;
                        }
                        sum += r.getSampleX();
                        count++;
                    }
                    if (count > 0) {
                        return new Double(sum / count);
                    } else {
                        return null;
                    }
                }
            };

    /**
     * Calculates the average of all Y components in sample coordinates.
     */
    public static final MeasurementValue<Double> SAMPLE_Y =
            new MeasurementValue<Double>("Y", "Am\u00B2", "Average Y (sample coordinates)") {
                public Double getValue(MeasurementStep step) {
                    if ( MeasurementStep.ORIGO_STEP==step.getStepValue() ) {
                        return 0.0;
                    }
                    double sum = 0.0;
                    int count = 0;
                    for (int i = 0; i < step.getResults(); i++) {
                        MeasurementResult r = step.getResult(i);
                        if (r.getType() != SAMPLE) {
                            continue;
                        }
                        sum += r.getSampleY();
                        count++;
                    }
                    if (count > 0) {
                        return new Double(sum / count);
                    } else {
                        return null;
                    }
                }
            };

    /**
     * Calculates the average of all Z components in sample coordinates.
     */
    public static final MeasurementValue<Double> SAMPLE_Z =
            new MeasurementValue<Double>("Z", "Am\u00B2", "Average Z (sample coordinates)") {
                public Double getValue(MeasurementStep step) {
                    if ( MeasurementStep.ORIGO_STEP==step.getStepValue() ) {
                        return 0.0;
                    }
                    double sum = 0.0;
                    int count = 0;
                    for (int i = 0; i < step.getResults(); i++) {
                        MeasurementResult r = step.getResult(i);
                        if (r.getType() != SAMPLE) {
                            continue;
                        }
                        sum += r.getSampleZ();
                        count++;
                    }
                    if (count > 0) {
                        return new Double(sum / count);
                    } else {
                        return null;
                    }
                }
            };

    /**
     * Calculates the normalized average of all X components in geographic coordinates.
     */
    public static final MeasurementValue<Double> GEOGRAPHIC_X_NORMALIZED =
            new NormalizedValue("X'", "Normalized average X (geographic coordinates)") {
                protected Double getValue0(MeasurementStep step) {
                    return GEOGRAPHIC_X.getValue(step);
                }
            };

    /**
     * Calculates the normalized average of all Y components in geographic coordinates.
     */
    public static final MeasurementValue<Double> GEOGRAPHIC_Y_NORMALIZED =
            new NormalizedValue("Y'", "Normalized average Y (geographic coordinates)") {
                protected Double getValue0(MeasurementStep step) {
                    return GEOGRAPHIC_Y.getValue(step);
                }
            };

    /**
     * Calculates the normalized average of all Z components in geographic coordinates.
     */
    public static final MeasurementValue<Double> GEOGRAPHIC_Z_NORMALIZED =
            new NormalizedValue("Z'", "Normalized average Z (geographic coordinates)") {
                protected Double getValue0(MeasurementStep step) {
                    return GEOGRAPHIC_Z.getValue(step);
                }
            };

    /**
     * Calculates the normalized average of all X components in sample coordinates.
     */
    public static final MeasurementValue<Double> SAMPLE_X_NORMALIZED =
            new NormalizedValue("X", "Normalized average X (sample coordinates)") {
                protected Double getValue0(MeasurementStep step) {
                    return SAMPLE_X.getValue(step);
                }
            };

    /**
     * Calculates the normalized average of all Y components in sample coordinates.
     */
    public static final MeasurementValue<Double> SAMPLE_Y_NORMALIZED =
            new NormalizedValue("Y", "Normalized average Y (sample coordinates)") {
                protected Double getValue0(MeasurementStep step) {
                    return SAMPLE_Y.getValue(step);
                }
            };

    /**
     * Calculates the normalized average of all Z components in sample coordinates.
     */
    public static final MeasurementValue<Double> SAMPLE_Z_NORMALIZED =
            new NormalizedValue("Z", "Normalized average Z (sample coordinates)") {
                protected Double getValue0(MeasurementStep step) {
                    return SAMPLE_Z.getValue(step);
                }
            };

    /**
     * Calculates the declination from the component averages in geographic coordinates.
     */
    public static final MeasurementValue<Double> DECLINATION =
            new MeasurementValue<Double>("D", "\u00b0", "Geographic declination") {
                public Double getValue(MeasurementStep step) {
                    Double x = GEOGRAPHIC_X.getValue(step);
                    Double y = GEOGRAPHIC_Y.getValue(step);
                    if (x == null || y == null || step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return null;
                    } else {
                        double d = Math.atan2(y, x);
                        if (d < 0.0) {
                            d += PI * 2;
                        }
                        return Math.toDegrees(d);
                    }
                }
            };

    /**
     * Calculates the inclination from the component averages in geographic coordinates.
     */
    public static final MeasurementValue<Double> INCLINATION =
            new MeasurementValue<Double>("I", "\u00b0", "Geographic inclination") {
                public Double getValue(MeasurementStep step) {
                    Double x = GEOGRAPHIC_X.getValue(step);
                    Double y = GEOGRAPHIC_Y.getValue(step);
                    Double z = GEOGRAPHIC_Z.getValue(step);
                    if (x == null || y == null || z == null || (x == 0.0 && y == 0.0)) {
                        return null;
                    } else {
                        if (x == 0.0) {
                            x = 0.000000000001;
                        }
                        if (y == 0.0) {
                            y = 0.000000000001;
                        }
                        double d = atan(z / sqrt(pow(x, 2.0) + pow(y, 2.0)));
                        return Math.toDegrees(d);
                    }
                }
            };

    /**
     * Calculates the length of the vector from the component averages.
     */
    public static final MeasurementValue<Double> MOMENT =
            new MeasurementValue<Double>("M", "Am\u00B2", "Magnetic moment") {
                public Double getValue(MeasurementStep step) {
                    Double x = SAMPLE_X.getValue(step);
                    Double y = SAMPLE_Y.getValue(step);
                    Double z = SAMPLE_Z.getValue(step);
                    if (x == null || y == null || z == null) {
                        return null;
                    } else {
                        return sqrt(Math.pow(x, 2) + Math.pow(y, 2) + Math.pow(z, 2));
                    }
                }
            };

    /**
     * Calculates the magnetic intensity (or remanence) from the moment and the sample's volume or mass (depending on
     * the selected normalization).
     */
    public static final MeasurementValue<Double> MAGNETIZATION =
            new NormalizedValue("J", "Magnetic intensity") { // J=M/volume or J=M/mass
                protected Double getValue0(MeasurementStep step) {
                    return MOMENT.getValue(step);
                }
            };

    /**
     * Calculates the magnetic intensity (or remanence) relative to the first measurement's magnetic intensity.
     */
    public static final MeasurementValue<Double> RELATIVE_MAGNETIZATION =
            new MeasurementValue<Double>("J/Jo", "", "Relative magnetic intensity") {
                public Double getValue(MeasurementStep step) {
                    Project project = step.getProject();
                    if (project == null || step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return null;
                    }
                    Double j = MAGNETIZATION.getValue(step);
                    Double j0 = MAGNETIZATION.getValue(project.getStep(0));     // there is at least one step in the project
                    if (j == null || j0 == null || j0 == 0.0) {
                        return null;
                    } else {
                        return j.doubleValue() / j0.doubleValue();
                    }
                }
            };

    /**
     * Calculates the angular standard deviation (Theta 63) from the measurement result set.
     * Formulas are based on information given in fisher_GTK.xls spreadsheet.
     */
    public static final MeasurementValue<Double> THETA63 =
            new MeasurementValue<Double>("\u03b863", "\u00b0", "Angular standard deviation") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    if (step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return 0.1;
                    }

                    int count = 0;

                    double declination = 0.0;
                    double inclination = 0.0;

                    // sums of calculations
                    double cosDcosI = 0.0;
                    double sinDcosI = 0.0;
                    double sinI = 0.0;

                    for (int i = 0; i < step.getResults(); i++) {
                        MeasurementResult r = step.getResult(i);
                        if (r.getType() != SAMPLE) {
                            continue;
                        }

                        // get sample coordinates
                        double x = r.getSampleX();
                        double y = r.getSampleY();
                        double z = r.getSampleZ();

                        // calculate declination from sample coordinates
                        declination = atan2(x, y);

                        if (declination < 0.0) {
                        	declination += PI * 2;
                        }

                        declination = Math.toDegrees(declination);
                        
                        // calculate inclination from sample coordinates
                        inclination = Math.toDegrees(atan(z / sqrt(pow(x, 2) + pow(y, 2))));

                        // more mathematical magic to produce correct theta63 value
                        cosDcosI += cos(Math.toRadians(declination))*cos(Math.toRadians(inclination));
                        sinDcosI += sin(Math.toRadians(declination))*cos(Math.toRadians(inclination));
                        sinI += sin(Math.toRadians(inclination));

                        count++;
                    }

                    // default value 1.0 for measurements if measured only in one position
                    if (count == 1) {
                    	return 1.0;
                    }
                    else if (count < 1) {
                        return null;
                    }

                    double N = count;
                    double R = sqrt((cosDcosI*cosDcosI) + (sinDcosI*sinDcosI) + (sinI*sinI));
                    double k = (N - 1) / (N - R);

                    return 81.0 / sqrt(k);

                }
            };

    /**
     *  X coordinate for Zidjerveld plot and PCA
     */
    public static final MeasurementValue<Double> ZIDJERVELD_X =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    if (step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return 0.0;
                    }

                    double xzidj;
                    Double intensity = MAGNETIZATION.getValue(step);  // MAGNETIZATION OR MOMENT ???
                    Double inclination = INCLINATION.getValue(step);
                    Double declination = DECLINATION.getValue(step);
                    if (intensity == null || inclination == null || declination == null) {
                        return null;
                    }

                    xzidj = intensity * Math.cos(Math.toRadians(inclination)) * Math.cos(Math.toRadians(declination));

                    return xzidj;
                }
            };

    /**
     *  Y coordinate for Zidjerveld plot and PCA
     */
    public static final MeasurementValue<Double> ZIDJERVELD_Y =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    if (step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return 0.0;
                    }

                    double yzidj;
                    Double intensity = MAGNETIZATION.getValue(step);
                    Double inclination = INCLINATION.getValue(step);
                    Double declination = DECLINATION.getValue(step);
                    if (intensity == null || inclination == null || declination == null) {
                        return null;
                    }

                    yzidj = intensity * Math.cos(Math.toRadians(inclination)) * Math.sin(Math.toRadians(declination));

                    return yzidj;
                }
            };

    /**
     *  Z coordinate for Zidjerveld plot and PCA
     */
    public static final MeasurementValue<Double> ZIDJERVELD_Z =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    if (step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return 0.0;
                    }

                    double zzidj;
                    Double intensity = MAGNETIZATION.getValue(step);
                    Double inclination = INCLINATION.getValue(step);
                    if (intensity == null || inclination == null) {
                        return null;
                    }

                    zzidj = intensity * Math.sin(Math.toRadians(inclination));

                    return zzidj;
                }
            };

    /**
     *  Normalized X coordinate for Zidjerveld plot and PCA
     */
    public static final MeasurementValue<Double> ZIDJERVELD_X_NORMALIZED =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    if (step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return 0.0;
                    }

                    double xzidj;
                    Double relative_intensity = RELATIVE_MAGNETIZATION.getValue(step);
                    Double inclination = INCLINATION.getValue(step);
                    Double declination = DECLINATION.getValue(step);
                    if (relative_intensity == null || inclination == null || declination == null) {
                        return null;
                    }

                    xzidj = relative_intensity * Math.cos(Math.toRadians(inclination)) * Math.cos(Math.toRadians(declination));

                    return xzidj;
                }
            };

    /**
     *  Normalized Y coordinate for Zidjerveld plot and PCA
     */
    public static final MeasurementValue<Double> ZIDJERVELD_Y_NORMALIZED =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    if (step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return 0.0;
                    }

                    double yzidj;
                    Double relative_intensity = RELATIVE_MAGNETIZATION.getValue(step);
                    Double inclination = INCLINATION.getValue(step);
                    Double declination = DECLINATION.getValue(step);
                    if (relative_intensity == null || inclination == null || declination == null) {
                        return null;
                    }

                    yzidj = relative_intensity * Math.cos(Math.toRadians(inclination)) * Math.sin(Math.toRadians(declination));

                    return yzidj;
                }
            };

    /**
     *  Normalized Z coordinate for Zidjerveld plot and PCA
     */
    public static final MeasurementValue<Double> ZIDJERVELD_Z_NORMALIZED =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    if (step.getStepValue()==MeasurementStep.ORIGO_STEP) {
                        return 0.0;
                    }

                    double zzidj;
                    Double relative_intensity = RELATIVE_MAGNETIZATION.getValue(step);
                    Double inclination = INCLINATION.getValue(step);
                    if (relative_intensity == null || inclination == null) {
                        return null;
                    }

                    zzidj = relative_intensity * Math.sin(Math.toRadians(inclination));

                    return zzidj;
                }
            };

    /**
     * Calculates the difference in normalized X in geographic coordinates between two steps.
     * Used by DIFFERENCE_DECLINATION and DIFFERENCE_INCLINATION.
     */
    public static final MeasurementValue<Double> DIFFERENCE_GEOGRAPHIC_X =
            new MeasurementValue<Double>("", "", "") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double x = GEOGRAPHIC_X_NORMALIZED.getValue(step);
                    Double prevX = GEOGRAPHIC_X_NORMALIZED.getValue(prev);
                    if (x == null || prevX == null) {
                        return null;
                    }
                    return prevX - x;
                }
            };

    /**
     * Calculates the difference in normalized Y in geographic coordinates between two steps.
     * Used by DIFFERENCE_DECLINATION and DIFFERENCE_INCLINATION.
     */
    public static final MeasurementValue<Double> DIFFERENCE_GEOGRAPHIC_Y =
            new MeasurementValue<Double>("", "", "") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double y = GEOGRAPHIC_Y_NORMALIZED.getValue(step);
                    Double prevY = GEOGRAPHIC_Y_NORMALIZED.getValue(prev);
                    if (y == null || prevY == null) {
                        return null;
                    }
                    return prevY - y;
                }
            };

    /**
     * Calculates the difference in normalized Z in geographic coordinates between two steps.
     * Used by DIFFERENCE_INCLINATION.
     */
    public static final MeasurementValue<Double> DIFFERENCE_GEOGRAPHIC_Z =
            new MeasurementValue<Double>("", "", "") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double z = GEOGRAPHIC_Z_NORMALIZED.getValue(step);
                    Double prevZ = GEOGRAPHIC_Z_NORMALIZED.getValue(prev);
                    if (z == null || prevZ == null) {
                        return null;
                    }
                    return prevZ - z;
                }
            };

    /**
     * Calculates the difference in normalized X in sample coordinates between two steps.
     * Used by DIFFERENCE_MAGNETIZATION.
     */
    public static final MeasurementValue<Double> DIFFERENCE_SAMPLE_X =
            new MeasurementValue<Double>("", "", "") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double x = SAMPLE_X_NORMALIZED.getValue(step);
                    Double prevX = SAMPLE_X_NORMALIZED.getValue(prev);
                    if (x == null || prevX == null) {
                        return null;
                    }
                    return prevX - x;
                }
            };

    /**
     * Calculates the difference in normalized Y in sample coordinates between two steps.
     * Used by DIFFERENCE_MAGNETIZATION.
     */
    public static final MeasurementValue<Double> DIFFERENCE_SAMPLE_Y =
            new MeasurementValue<Double>("", "", "") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double y = SAMPLE_Y_NORMALIZED.getValue(step);
                    Double prevY = SAMPLE_Y_NORMALIZED.getValue(prev);
                    if (y == null || prevY == null) {
                        return null;
                    }
                    return prevY - y;
                }
            };

    /**
     * Calculates the difference in normalized Z in sample coordinates between two steps.
     * Used by DIFFERENCE_MAGNETIZATION.
     */
    public static final MeasurementValue<Double> DIFFERENCE_SAMPLE_Z =
            new MeasurementValue<Double>("", "", "") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double z = SAMPLE_Z_NORMALIZED.getValue(step);
                    Double prevZ = SAMPLE_Z_NORMALIZED.getValue(prev);
                    if (z == null || prevZ == null) {
                        return null;
                    }
                    return prevZ - z;
                }
            };

    /**
     * Calculates the difference in magnetization between two steps
     */
    public static final MeasurementValue<Double> DIFFERENCE_MAGNETIZATION =
            new MeasurementValue<Double>("\u0394J", "", "difference in magnetization") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double x = DIFFERENCE_SAMPLE_X.getValue(step);
                    Double y = DIFFERENCE_SAMPLE_Y.getValue(step);
                    Double z = DIFFERENCE_SAMPLE_Z.getValue(step);
                    if (x == null || y == null || z == null) {
                        return null;
                    } else {
                        return sqrt(Math.pow(x, 2) + Math.pow(y, 2) + Math.pow(z, 2));
                    }
                }

                // hack: return the correct unit
                @Override public String getUnit(Project project) {
                    if (project == null) {
                        return "mA/m";
                    } else if (project.getNormalization() == Project.Normalization.VOLUME) {
                        return "mA/m";
                    } else if (project.getNormalization() == Project.Normalization.MASS) {
                        return "mAm\u00B2/kg";
                    } else {
                        assert false;
                        return "";
                    }
                }
            };

    /**
     * Calculates the difference in declination between two steps.
     */
    public static final MeasurementValue<Double> DIFFERENCE_DECLINATION =
            new MeasurementValue<Double>("\u0394D", "\u00b0", "difference in declination") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double x = DIFFERENCE_GEOGRAPHIC_X.getValue(step);
                    Double y = DIFFERENCE_GEOGRAPHIC_Y.getValue(step);
                    if (x == null || y == null) {
                        return null;
                    } else {
                        double d = Math.atan2(y, x);
                        if (d < 0.0) {
                            d += PI * 2;
                        }
                        return Math.toDegrees(d);
                    }
                }
            };

    /**
     * Calculates the difference in inclination between two steps.
     */
    public static final MeasurementValue<Double> DIFFERENCE_INCLINATION =
            new MeasurementValue<Double>("\u0394I", "\u00b0", "difference in inclination") {
                public Double getValue(MeasurementStep step) {
                    MeasurementStep prev = getPrev(step);
                    if (prev == null) {
                        return null;
                    }
                    Double x = DIFFERENCE_GEOGRAPHIC_X.getValue(step);
                    Double y = DIFFERENCE_GEOGRAPHIC_Y.getValue(step);
                    Double z = DIFFERENCE_GEOGRAPHIC_Z.getValue(step);
                    if (x == null || y == null || z == null) {
                        return null;
                    } else {
                        if (x == 0.0) {
                            x = 0.000000000001;
                        }
                        if (y == 0.0) {
                            y = 0.000000000001;
                        }
                        double d = atan(z / sqrt(pow(x, 2.0) + pow(y, 2.0)));
                        return Math.toDegrees(d);
                    }
                }
            };

    /**
     * TODO: enter description
     */
    public static final MeasurementValue<Double> SIGNAL_TO_NOISE =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }

                    // compute the average length of the vectors
                    int count = 0;
                    double lengthSum = 0.0;
                    for (MeasurementResult r : step) {
                        lengthSum += r.getSampleVector().length();
                        count++;
                    }
                    if (count < 2) {
                        return null;
                    }
                    double lengthAvg = lengthSum / count;

                    // compute the standard deviation
                    double squareSum = 0.0;
                    for (MeasurementResult r : step) {
                        double d = r.getSampleVector().length() - lengthAvg;
                        squareSum += (d * d);
                    }
                    double stdev = Math.sqrt(squareSum / (count - 1));

                    return lengthAvg / stdev;
                }
            };

    /**
     * TODO: enter description
     */
    public static final MeasurementValue<Double> SIGNAL_TO_DRIFT =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    Double signal = MOMENT.getValue(step);
                    if (signal == null) {
                        return null;
                    }
                    double d = step.getNoise().length();
                    if (d == 0.0) {
                        return null;
                    }
                    return signal.doubleValue() / d;
                }
            };

    /**
     * TODO: enter description
     */
    public static final MeasurementValue<Double> SIGNAL_TO_HOLDER =
            new MeasurementValue<Double>("caption", "unit", "description") {
                public Double getValue(MeasurementStep step) {
                    if (step == null) {
                        return null;
                    }
                    Double signal = MOMENT.getValue(step);
                    if (signal == null) {
                        return null;
                    }
                    double d = step.getHolder().length();
                    if (d == 0.0) {
                        return null;
                    }
                    return signal.doubleValue() / d;
                }
            };


    /*    Begin instance declaration    */

    /**
     * A short name for the value.
     */
    private final String caption;

    /**
     * The unit of the value.
     */
    private final String unit;

    /**
     * A long description of the value.
     */
    private final String description;

    /**
     * Creates a new measurement value.
     *
     * @param caption     a short name for the value.
     * @param unit        the unit of the value.
     * @param description a long description of the value.
     * @throws NullPointerException if any of the arguments is null.
     */
    public MeasurementValue(String caption, String unit, String description) {
        if (caption == null || unit == null || description == null) {
            throw new NullPointerException();
        }
        this.caption = caption;
        this.unit = unit;
        this.description = description;
    }

    /**
     * Calculates a specific value from a measurement step.
     *
     * @param step the step from which the value will be calculated.
     * @return the calculated value, or null if it was not possible to calculate it.
     * @throws NullPointerException if step is null.
     */
    public abstract T getValue(MeasurementStep step);

    /**
     * Returns a short name for the value.
     *
     * @param project the currently active project, or null if no project is active. Used for returning a different text
     *                depending on the project.
     */
    public String getCaption(Project project) {
        return caption;
    }

    /**
     * Returns the unit of the value.
     *
     * @param project the currently active project, or null if no project is active. Used for returning a different text
     *                depending on the project.
     */
    public String getUnit(Project project) {
        return unit;
    }

    /**
     * Returns a long description of the value.
     *
     * @param project the currently active project, or null if no project is active. Used for returning a different text
     *                depending on the project.
     */
    public String getDescription(Project project) {
        return description;
    }

    /**
     * Returns the previous step in the measurement sequence
     * @param step the step whose
     * @return the step before the step given as a parameter.
     */
    private static MeasurementStep getPrev(MeasurementStep step) {
        if (step == null) {
            return null;
        }
        Project project = step.getProject();
        if (project == null) {
            return null;
        }
        int i = project.getStepIndex(step);
        if (i < 1) {
            return null;
        }
        return project.getStep(i - 1);
    }
}
