/*
 * NormalizedValue.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis;

import static mavis.MeasurementResult.Type.*;
import static java.lang.Math.atan2;
import static java.lang.Math.PI;
import static java.lang.Math.atan;
import static java.lang.Math.pow;
import static java.lang.Math.sqrt;
import static java.lang.Math.sin;
import static java.lang.Math.cos;
import java.lang.Math;

/**
 * Specializes the MeasurementValue class with automatic value normalization.
 *
 * @author Esko Luontola
 */
abstract class NormalizedValue extends MeasurementValue<Double> {

    public NormalizedValue(String caption, String description) {
        super(caption, "", description);
    }

    /**
     * Calculates a specific normalized value from a measurement step.
     *
     * @param step the step from which the value will be calculated.
     * @return the calculated value, or null if it was not possible to calculate it. The unit is mA/m or Am^2/kg
     *         depenging on the normalization that was used.
     * @throws NullPointerException if step is null.
     */
    public final Double getValue(MeasurementStep step) {
        return normalize(step, getValue0(step));
    }

    /**
     * Calculates a specific NOT NORMALIZED value from a measurement step.
     *
     * @param step the step from which the value will be calculated.
     * @return the calculated value, or null if it was not possible to calculate it. The unit is Am^2.
     * @throws NullPointerException if step is null.
     */
    protected abstract Double getValue0(MeasurementStep step);

    /**
     * Returns mA/m or Am^2/kg depending on the normalization that the project uses.
     */
    @Override public String getUnit(Project project) {
        if (project == null) {
            return "mA/m";
        } else if (project.getNormalization() == Project.Normalization.VOLUME) {
            return "mA/m";
        } else if (project.getNormalization() == Project.Normalization.MASS) {
            return "mAm\u00B2/kg";
        } else {
            assert false;
            return "";
        }
    }

    /**
     * Normalizes a value by mass or volume, as specified in the project or step.
     *
     * @param step  the step whose value is being normalized.
     * @param value the value to be normalised. The unit is Am^2.
     * @return the normalized value if the mass/volume was specified, otherwise
     *         the original un-normalized value, or null if errors occur. The unit is mAm^2 divided by m^3
     *         or kg depending on the used normalization.
     */
    private static Double normalize(MeasurementStep step, Double value) {
        Project project = step.getProject();
        if (project == null) {
            return null;
        }

        double normalizer;
        if (project.getNormalization() == Project.Normalization.VOLUME) {
            normalizer = step.getVolume();
            if (normalizer < 0.0) {
                normalizer = project.getVolume();
            }
            normalizer = normalizer / 1000000.0;    // convert cm^3 to m^3

        } else if (project.getNormalization() == Project.Normalization.MASS) {
            normalizer = step.getMass();
            if (normalizer < 0.0) {
                normalizer = project.getMass();
            }
            normalizer = normalizer / 1000.0;       // convert g to kg

        } else {
            assert false;
            return null;
        }
        if (normalizer <= 0.0) {
            return value;
        }

        if (value == null) {
            return null;
        } else {
            return (value * 1000.0) / normalizer;   // convert Am^2 to mAm^2
        }
    }
}