/*
 * Session.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis;

import mavis.gui.MagneticComponentTableModel;
import mavis.gui.StyledWrapper;

import javax.swing.event.EventListenerList;
import javax.swing.*;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector2d;
import java.util.Vector;
import java.util.Locale;
import java.util.Date;
import java.util.Random;
import java.io.*;
import java.text.NumberFormat;
import java.text.DateFormat;
import java.text.DecimalFormat;

/**
 * This class functions as a common interface to all magnetic
 * component related calculations and manipulations. It also
 * allows exporting a CSV file with component data.
 *
 */

// Notes:
//   - component index: 0 is the first

public class Session {

    private Project project = null;

    // measurementsteps associated to each component are kept in here:
    MagneticComponents components = null;

    // keep track of wheter components have already been saved to CSV or not
    boolean saved;


    /**
     * Listeners for this project.
     */
    private final EventListenerList listenerList = new EventListenerList();


    /**
     * Initialize a new Session instance.
     */
    public Session(Project proj) {
        project = proj;
        components = new MagneticComponents(this);
        saved = true;
    }

    /**
     * Adds a SessionListener to the session.
     *
     * @param listener the listener to be added.
     */
    public synchronized void addSessionListener(SessionListener listener) {
        listenerList.add(SessionListener.class, listener);
    }

    /**
     * Removes a SessionListener from the session.
     *
     * @param listener the listener to be removed
     */
    public synchronized void removeSessionListener(SessionListener listener) {
        listenerList.remove(SessionListener.class, listener);
    }

    /**
     * Notifies all listeners that have registered for SessionEvents.
     *
     * @param type type of the event.
     */
    protected synchronized void fireSessionEvent(SessionEvent.Type type) {
        final SessionEvent event = new SessionEvent(this, type);
        final SessionListener[] listeners = listenerList.getListeners(SessionListener.class);
        //jfw: invokeLater()=stupid since it doesn't create a thread so func is blocking in any case
        //SwingUtilities.invokeLater(new Runnable() {
        //    public void run() {
        for (SessionListener l : listeners) {
            try {
                l.sessionUpdated(event);
            } catch (Throwable t) {
                t.printStackTrace();
            }
        }
        //    }
        //});
    }

    /**
     * Deletes component and fires event to notify registered listeners.
     */
    public boolean deleteComponent(int toBeDeleted) {
        if (null == components) {
            return false;
        } else if (components.getNumOfComponents() < toBeDeleted) {
            return false;
        }
        if (!components.removeComponent(toBeDeleted)) {
            return false;
        }
        fireSessionEvent(SessionEvent.Type.COMPONENT_REMOVE);
        saved = false;
        return true;
    }

    /**
     * Adds a new component that consists of associates measurement steps
     * and their indices in the measurement sequence (these are just for
     * information). The component is placed at the end of the list of
     * current components.
     */
    public void addComponent(Vector<MeasurementStep> associatedSteps, int[] stepIndices) {
        if (null == components) {
            components = new MagneticComponents(this);
        }
        /* for Matlab data import:
		System.out.println("-------------");
        for (MeasurementStep s : associatedSteps) {
            System.out.print("" + MeasurementValue.ZIDJERVELD_X_NORMALIZED.getValue(s) + "\t");
            System.out.print("" + MeasurementValue.ZIDJERVELD_Y_NORMALIZED.getValue(s) + "\t");
            System.out.println("" + MeasurementValue.ZIDJERVELD_Z_NORMALIZED.getValue(s) + "\t");
        }
        System.out.println("-------------");
		*/
        components.addComponent(associatedSteps);
        components.addStepNumbers(stepIndices);

        // Calculated component data will be fetched by listeners
        // via calls to this session's getXXX() wrapper methods
        fireSessionEvent(SessionEvent.Type.COMPONENT_ADD_NEW);
        saved = false;
    }

    /**
     * Returns the number of magnetic components that the user has added.
     */
    public synchronized int getNumOfComponents() {
        if (null == components) {
            return 0;
        } else {
            return components.getNumOfComponents();
        }
    }

    /**
     * Returns maximum intensity of component.
     */
    public synchronized double getMaxIntensity(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            return ComponentCalculation.calculateMaxIntensity(components.getSteps(component));
        }
    }

    /**
     * Returns the minimum intensity of component.
     */
    public synchronized double getMinIntensity(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            return ComponentCalculation.calculateMinIntensity(components.getSteps(component));
        }
    }

    /**
     * Returns the average intensity of component.
     */
    public synchronized double getAvgIntensity(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            return ComponentCalculation.calculateAvgIntensity(components.getSteps(component));
        }
    }

    /**
     * Returns the inclination of component.
     */
    public synchronized double getInclination(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            double incl = ComponentCalculation.calculateInclination(components.getSteps(component));
            if (components.getAntipole(component)) {
                // this is an ugly "correction" for PCA results, invert the
                // component vector direction:
                incl *= -1.0;
            }
            return incl;
        }
    }

    /**
     * Returns the declination of component.
     */
    public synchronized double getDeclination(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            double decl = ComponentCalculation.calculateDeclination(components.getSteps(component));
            if (components.getAntipole(component)) {
                // this is an ugly "correction" for PCA results, invert the
                // component vector direction:
                decl += 180.0;
                if (decl>360.0) {
                    decl -= 360.0;
                }
            }
            return decl;
        }
    }

    /**
     * Returns angle of error (maximum angular deviation) of component.
     */
    public synchronized double getAoE(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            return ComponentCalculation.calculateAoE(components.getSteps(component));
        }
    }

    /**
     * Returns the ancient pole position (Latitude).
     */
    public synchronized double getPoleLat(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            return ComponentCalculation.calculatePoleLat(components.getSteps(component), project);
        }
    }

    /**
     * Returns ancient pole position (Longitude).
     */
    public synchronized double getPoleLong(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            return ComponentCalculation.calculatePoleLong(components.getSteps(component), project);
        }
    }

    /**
     * Returns the coords of the starting point (3d).
     */
    public synchronized Vector3d getStart(int component) {
        if (null == components || component < 0) {
            return null;
        } else if (components.getNumOfComponents() < component) {
            return null;
        } else {
            return ComponentCalculation.calculateStart(components.getSteps(component));
        }
    }

    /**
     * Sets a component to be antipole.
     */
    public synchronized void setAntipole(boolean isAntipole, int component) {
        components.setAntipole(component, isAntipole);
    }

    /**
     * Returns true if component is set to be an antipole.
     */
    public synchronized boolean getAntipole(int component) {
        return components.getAntipole(component);
    }

    /**
     * Returns the coords of the end point (3d).
     */
    public synchronized Vector3d getEnd(int component) {
        if (null == components || component < 0) {
            return null;
        } else if (components.getNumOfComponents() < component) {
            return null;
        } else {
            return ComponentCalculation.calculateEnd(components.getSteps(component));
        }
    }


    /**
     * Returns the coords of the starting and ending points,
     * corrected for the case of the origo being artificially
     * forced to be one of the end points of the magnetic
     * vector component.
     *
     * @return Vector3d[2] with the first entry being the start
     *         and the second entry the end point of the 3D component
     *         line.
     */
    public synchronized Vector3d[] getStartAndEndpoints(int component) {
        if (null == components || component < 0) {
            return null;
        } else if (components.getNumOfComponents() < component) {
            return null;
        } else {
            Vector<MeasurementStep> steps = components.getSteps(component);
            Vector3d[] result = new Vector3d[2];
            Vector3d start = ComponentCalculation.calculateStart(steps);
            Vector3d end = ComponentCalculation.calculateEnd(steps);

            if (ComponentCalculation.containsOrigo(steps)) {
                // Force the component to go to origo.

                // Implementation 1: fulfills specification that center-of-mass
                //                   should be ignored - shift component vector to
                //                   end at origo
                /*
                if(start.length()<=end.length()) {
                    // start closest to origo: shift start into origo, and end closer to origo
                    end.sub(start);
                    start.set(0, 0, 0);
                } else {
                    // end closest to origo: shift end into origo, and start closer to origo
                    start.sub(end);
                    end.set(0, 0, 0);
                }
                */
                // Implementation 2: fulfills requirement that component should end
                //                   at origo, looks nicer, but gives incorrect
                //                   component length - anchor one point into origo
                /*
                if(start.length()<=end.length()) {
                    start.set(0, 0, 0);
                } else {
                    end.set(0, 0, 0);
                }
                */
                // Implementation 3: fulfills specification that center-of-mass
                //                   should be ignored, but gives incorrect component
                //                   length - shift component vector to end at origo,
                //                   and extend the start point to orthogonally intersect
                //                   the earlier start point.

                Vector3d direction = null, vtemp = null, start_old = null;
                if (start.length() <= end.length()) {
                    // swap so that 'end' is the one that will end in origo
                    vtemp = start;
                    start = end;
                    end = vtemp;
                } else {
                    // 'end' already closest to origo, don't need to swap
                    // start=start;
                    // end=end;
                }
                start_old = new Vector3d(start);
                start.sub(end);                  // new start = shifted old start
                end.set(0, 0, 0);                // new end = origo
                direction = new Vector3d(start); // dir = (start-end)/|start-end|
                direction.normalize();
                start.set(direction);
                start.scale(direction.dot(start_old));  // extend from origo to previous start
            }
            result[0] = start;
            result[1] = end;
            return result;
        }
    }

    /**
     * Returns the coords of the starting point (2d), for component
     * line fitting done in 2D i.e. as seen from specified axis.
     */
    public synchronized Vector2d getStart2D(int component, ComponentCalculation.Axis axis) {
        if (null == components || null == axis || component < 0) {
            return null;
        } else if (components.getNumOfComponents() < component) {
            return null;
        } else {
            return ComponentCalculation.calculateStart2D(components.getSteps(component), axis);
        }
    }

    /**
     * Returns the coords of the end point (2d), for component
     * line fitting done in 2D i.e. as seen from specified axis.
     */
    public synchronized Vector2d getEnd2D(int component, ComponentCalculation.Axis axis) {
        if (null == components || null == axis || component < 0) {
            return null;
        } else if (components.getNumOfComponents() < component) {
            return null;
        } else {
            return ComponentCalculation.calculateEnd2D(components.getSteps(component), axis);
        }
    }

    /**
     * Returns the length.
     */
    public synchronized double getLength(int component) {
        if (null == components || component < 0) {
            return 0;
        } else if (components.getNumOfComponents() < component) {
            return 0;
        } else {
            return ComponentCalculation.calculateLength(components.getSteps(component));
        }
    }

    /**
     * Returns the indexes of steps that component includes.
     */
    public synchronized int[] getStepNumbers(int component) {
        if (null == components) {
            return null;
        } else if (components.getNumOfComponents() < component) {
            return null;
        } else {
            return (components.getStepNumbers(component));
        }
    }

    /**
     * Returns the indexes of steps that component includes,
     * format is like "1-8,10,15,18-20"
     */
    public synchronized String getStepNumbersString(int component) {
        int[] steps = getStepNumbers(component);
        if (null == steps || steps.length < 1) {
            return "N/A";
        }
        return "" + (steps[0] + 1) + "-" + (steps[steps.length - 1] + 1);
    }

    /**
     * Returns the indices of the excluded steps in the component's step range.
     *
     * @param component The magnetic component's index in the array.
     * @return The excluded steps in an int array.
     */
    public synchronized int[] getExcluded(int component) {
        if (components == null || components.getNumOfComponents() < component) {
            return null;
        }
        int[] steps = components.getStepNumbers(component);
        int len = steps.length;
        int rangeLen = steps[len - 1] - steps[0] + 1;
        // check if no steps are excluded
        if (rangeLen - len == 0) {
            return null;
        }
        int[] excluded = new int[rangeLen - len];
        int prev = steps[0];
        for (int i = 1, j = 0; i < len; i++) {
            int step = steps[i];
            if (step != prev + 1) {
                for (int s = prev + 1; s < step; s++) {
                    excluded[j++] = s;
                }
            }
            prev = step;
        }
        return excluded;
    }

    /**
     * Returns the indexes of steps that are in the step range, but  not included in the component,
     * format is like "2-8,10,15,18-20"
     *
     * @param component The magnetic component's index in the array.
     * @return The String listing the excluded steps.
     */
    public synchronized String getExcludedString(int component) {
        int[] steps = getExcluded(component);
        if (null == steps || steps.length < 1) {
            return "N/A";
        }
        String out = "";
        int stepsLen = steps.length;
        int icurr = 0;
        int start_val = steps[icurr];

        out += (start_val + 1);
        icurr ++;
        for (int istart = 0; istart < stepsLen && icurr < stepsLen; icurr++) {
            // steps e.g. {1 3} => 1,3  {1 2 3 5} => 1-3,5
            if (steps[icurr] != (start_val + (icurr - istart))) {
                if (icurr == (istart - 1) || istart == (icurr - 1)) {
                    out += "," + (steps[icurr] + 1);
                    start_val = steps[icurr];
                    istart = icurr;
                } else {
                    out += "-" + (steps[icurr - 1] + 1);
                    out += "," + (steps[icurr] + 1);
                    start_val = steps[icurr];
                    istart = icurr;
                }
            } else {
                // still going in 1's increments
                // check if last step
                if (icurr == (stepsLen - 1)) {
                    out += "-" + (steps[icurr] + 1);
                }
            }
        }
        return out;

    }

    /**
     * Returns the Project associated with this Session.
     *
     * @return The current Project
     */
    public synchronized Project getProject() {
        return project;
    }

    /**
     * Tells whether the components in the session have been saved.
     *
     * @return true, if the session has been saved, false otherwise.
     */
    public synchronized boolean isSaved() {
        return saved;
    }

    /**
     * Sets the save status.
     *
     * @param saved true or false.
     */
    public synchronized void setSaved(boolean saved) {
        this.saved = saved;
    }

    /**
     * Writes the magnetic components list to a file in CSV format.
     *
     * @param file the file to write/append to.
     * @return true if new component data was successfully added
     *         to the file.
     */
    public synchronized boolean saveToTable(File file) {
        PrintStream out = null;
        if (null == project || null == file) {
            return false;
        }
        if (null == components || components.getNumOfComponents()<=0) {
            return false;
        }
        try {

            /* Desired format:
            Sample	Start	End	D	I	err	Plat	Plon	Lat	Lon	Rocktype	Density	Q-value	NRM	Susc	Comment
            AD1-2a	100	-9999	282.8	29.5	1.9	19.2	166.1	33.8	249.5	Diabase	2938	1.51	1575.9	26200	coarse	grained	Sill
            AD1-2a	150	-9999	282.6	30.2	2.4	19.2	166.5	33.8	249.5	Diabase	2938	1.51	1575.9	26200	coarse	grained	Sill
            AD2-1a	150	-9999	263	41.9	6.5	7.8	183.5	33.8	249.5	Diabase	2915	0.66	676	25720
            AD2-2a	600	-1600	283.4	47.7	11.5	25.9	178.1	33.8	249.5	diabase	NAN.5	3.1	357.32	2893	road	A270	to

            Later changes: 28/04 - End column isn't negative, err is MAD, excluded steps into own column
                                   own addition: Antipole column

            Type of csv separator: unspecified - assuming ';' (cf Settings.java)
            Type of decimal separator: unspecified - assuming fixed '.'
            */

            boolean bAppending = file.exists();
            FileOutputStream fileos = new FileOutputStream(file, bAppending);
            out = new PrintStream(fileos, true, "ISO-8859-1");

            // CSV separator currently not selectable via GUI,
            // but manual edit of the config file works ;-)
            String SPR = Settings.getCSVfileSeparator();

            double d = 0.0;
            String str = "";

            // fixed EN/US locale formatters for numbers
            NumberFormat nfTwoDecimal = NumberFormat.getInstance(Locale.US);
            if (nfTwoDecimal instanceof DecimalFormat) {
                ((DecimalFormat) nfTwoDecimal).setDecimalSeparatorAlwaysShown(false);
            }
            nfTwoDecimal.setGroupingUsed(false);

            // TODO: number of integer digits does not restrict num of decimal digits yet

            NumberFormat nfAngle = (NumberFormat)nfTwoDecimal.clone();
            nfAngle.setMinimumIntegerDigits(1);         // e.g. "282.6", "263"
            nfAngle.setMaximumIntegerDigits(4);         // (3 and 1 spare)
            nfAngle.setMinimumFractionDigits(0);
            nfAngle.setMaximumFractionDigits(1);

            nfTwoDecimal.setMaximumFractionDigits(2);   // e.g. "357.32", "1575.9" or "498"
            nfTwoDecimal.setMinimumFractionDigits(0);

            NumberFormat nfOneTwoDecimal = (NumberFormat)nfTwoDecimal.clone();
            nfOneTwoDecimal.setMinimumFractionDigits(1);

            // Write data headers, if this is a new file
            if(!bAppending) {
                String headings = "Sample\tStart\tEnd\tD\tI\tMAD\tPlat\tPlon\tLat\tLon\tRocktype\tDensity\tQ-value\tNRM\tSusc\tExcl\tAP\tComment";
                headings = headings.replace('\t', SPR.charAt(0));
                out.println(headings);
            }

            // Project data
            for (int c = 0; c < components.getNumOfComponents(); c++) {

                out.print(project.getName() + SPR);    // "Name"
                double val = 0.0;

                int[] step_indices = components.getStepNumbers(c);
                try {                                  // "Start"
                    val = project.getStep(step_indices[0]).getStepValue();
                    out.print("" + nfTwoDecimal.format(val) + SPR);
                } catch (Exception e) {
                    out.print(SPR);
                }

                try {                                  // "End"
                    // TODO: mavis_table.xls indicates end points/values should be negative. silly. check.
                    val = project.getStep(step_indices[step_indices.length - 1]).getStepValue();
                    out.print("" + nfTwoDecimal.format(val) + SPR);
                } catch (IndexOutOfBoundsException e) {
                    out.print(SPR);
                }

                val = getDeclination(c);
                out.print("" + nfAngle.format(val) + SPR);  // "D"

                val = getInclination(c);
                out.print("" + nfAngle.format(val) + SPR);  // "I"

                val = getAoE(c);
                out.print("" + nfAngle.format(val) + SPR);  // "MAD"

                val = getPoleLat(c);
                out.print("" + nfAngle.format(val) + SPR);  // "Plat"

                val = getPoleLong(c);
                out.print("" + nfAngle.format(val) + SPR);  // "Plon"

                // possible TODO: format Lat, Lon strings with nfAngle
                str = project.getProperty(Project.LATITUDE_PROPERTY, "");
                out.print(str + SPR);                   // "Lat"

                str = project.getProperty(Project.LONGITUDE_PROPERTY, "");
                out.print(str + SPR);                   // "Lon"

                str = project.getProperty(Project.ROCK_TYPE_PROPERTY, "");
                out.print(str + SPR);                   // "Rocktype"

                                                        // "Density"
                if (project.getMass() >= 0 && project.getVolume() > 0) {
                    d = project.getMass() / project.getVolume() * 1000.0;
                    out.print(nfTwoDecimal.format(d) + SPR);
                } else {
                    out.print("NAN.5" + SPR);
                }

                if (project.getStep(0) != null) {       // "Q-value"
                    // first step should be with none-demagnetization so its Intensity is NRM
                    Double mag = project.getValue(0, MeasurementValue.MAGNETIZATION);
                    if (mag != null && project.getSusceptibility() != null && project.getSusceptibility() > 0) {
                        double q = 25.13 * mag / project.getSusceptibility();
                        out.print(nfOneTwoDecimal.format(q) + SPR);
                    } else {
                        out.print("NAN.5" + SPR);
                    }
                } else {
                    out.print("N/A" + SPR);
                }

                val = this.getAvgIntensity(c);
                out.print(nfTwoDecimal.format(val) + SPR); // "NRM"

                if (null != project.getSusceptibility()) { // "Susc"
                    // possible TODO: format as integer, no decimals
                    out.print(project.getSusceptibility() + SPR);
                }

                out.print("" + getExcludedString(c) + SPR); // "Excl"

                out.print("" + components.getAntipole(c) + SPR); // "AP"

                str = project.getProperty(Project.COMMENT_PROPERTY, "");
                str = str.replaceAll("\\s", " ");
                str = str.replaceAll(" ", SPR);
                out.print(str + SPR);                      // "Comment"

                out.println("");
            }

            this.saved = true;

            // exporting finished
            return true;

        } catch (FileNotFoundException e) {
            JOptionPane.showMessageDialog(null,
                    "Could not open the file\r\n" + file + "\r\nfor writing. " +
                            "Check permissions and that no other applications are using the file.",
                    "Error", JOptionPane.ERROR_MESSAGE);
        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            if (out != null) {
                out.close();
            }
        }
        return false;
    }


}
