/*
 * MeasurementSequenceTableModel.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis.gui;

import mavis.*;

import javax.swing.border.Border;
import javax.swing.*;
import java.text.ParseException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.awt.*;

/**
 * Represents a column in the magnetic component table. Calculates the values of that column.
 */
public enum ComponentColumn {

    /**
     * Showing ordinal number of the magnetic component, starting from number 1.
     */
    COUNT("#") {
        {
            toolTipText = "Number of the magnetic component";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = headerWrapper;
            if (session != null) {
                wrapper.value = new Integer(rowIndex + 1);
            }
            return wrapper;
        }
    },

    /**
     * Showing the steps included in the magnetic component.
     */
    STEPS("Steps"){

        {
            toolTipText = "Steps included in component";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                wrapper.value = session.getStepNumbersString(rowIndex);
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the intensity/temperature range of the magnetic component.
     */
    RANGE("Range"){

        @Override public String getToolTipText(Project project) {
            final String AF_TOOL_TIP = "Intensity range of component";
            final String TEMP_TOOL_TIP = "Temperature range of component";
            if (project == null) {
                return AF_TOOL_TIP;
            }
            String s;
            switch (project.getType()) {
                case AF:
            case GAF:
            case CALIBRATION:
                s = AF_TOOL_TIP;
                break;
            case THELLIER:
            case THERMAL:
                s = TEMP_TOOL_TIP;
                break;
            default:
                assert false;
                s = null;
                break;
            }
            return s;
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                double min = session.getMinIntensity(rowIndex);
                double max = session.getMaxIntensity(rowIndex);
                String returnValue = min + "-" + max;
                wrapper.value = returnValue;
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the declination of the magnetic component.
     */
    D("D (\u00b0)"){
        {
            toolTipText = "Declination of component";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                wrapper.value = numberFormat.format(session.getDeclination(rowIndex));
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the inclination of the magnetic component.
     */
    I("I (\u00b0)"){
        {
            toolTipText = "Inclination of component";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                wrapper.value = numberFormat.format(session.getInclination(rowIndex));
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the maximum angular deviation of the magnetic component.
     */
    ANGLE("MAD") {
        {
            toolTipText = "Maximum angular deviation of component";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                wrapper.value = numberFormat.format(session.getAoE(rowIndex));
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the Median intensity of the magnetic component.
     */
    DJ("dj") {
        {
            toolTipText = "Median intensity";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                wrapper.value = numberFormat.format(session.getAvgIntensity(rowIndex));
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the Ancient pole position (Longitude) of the magnetic component.
     */
    PLON("Plon") {
        {
            toolTipText = "Ancient pole position (Longitude)";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                wrapper.value = numberFormat.format(session.getPoleLong(rowIndex));
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the Ancient pole position (Latitude) of the magnetic component.
     */
    PLAT("Plat") {
        {
            toolTipText = "Ancient pole position (Latitude)";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                wrapper.value = numberFormat.format(session.getPoleLat(rowIndex));
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the excluded points in the range of the magnetic component.
     */
    EXCLUDED("Excluded") {
        {
            toolTipText = "Excluded points in the range of component";
        }

        @Override public StyledWrapper getValue(int rowIndex, Session session) {
            StyledWrapper wrapper = defaultWrapper;
            if (session != null && session.getNumOfComponents() > rowIndex) {
                wrapper.value = session.getExcludedString(rowIndex);
            } else {
                wrapper.value = null; // right-pointing double angle quotation mark
            }
            return wrapper;
        }
    },

    /**
     * Showing the antipole of the magnetic component.
     */
    ANTIPOLE("AP") {
         {
            toolTipText = "Antipole";
         }

        @Override public Boolean getValue(int rowIndex, Session session) {
             if (session != null && session.getNumOfComponents() > rowIndex) {
                 return session.getAntipole(rowIndex);
             } else {
                 return false;
             }
        }

        @Override public void setValue(Object value, int rowIndex, Session session) {
            if (session != null && rowIndex>=0 && rowIndex<session.getNumOfComponents()
                && value instanceof Boolean) {
                session.setAntipole((Boolean)value, rowIndex);
            }
        }

        @Override public boolean isCellEditable(int rowIndex) {
            return true;
        }

        @Override public Class getColumnClass() {
            return Boolean.class;
        }

    };

    /**
     * Returns all the columns supported by the program. The returned values are in the order that they should be shown
     * in the measurement sequence table.
     */
    public static ComponentColumn[] getAllColumns() {
        return new ComponentColumn[]{
            COUNT,
            STEPS,
            RANGE,
            D,
            I,
            ANGLE,
            DJ,
            PLON,
            PLAT,
            ANTIPOLE,
            EXCLUDED,
        };
    }

    /** Style for the default cell renderer */
    private static final StyledWrapper defaultWrapper = Settings.getDefaultWrapperInstance();
    /** Style for the cell header renderer */
    private static final StyledWrapper headerWrapper = new StyledWrapper();

    static {
        // align all fields to right
        defaultWrapper.horizontalAlignment = SwingConstants.TRAILING;
        headerWrapper.horizontalAlignment = SwingConstants.TRAILING;

        // styles for the COUNT column
        Color headerBackground = new Color(0xE1E1E1);
        Border headerBorder = BorderFactory.createEmptyBorder(1, 2, 1, 2);
        headerWrapper.opaque = true;
        headerWrapper.background = headerBackground;
        headerWrapper.selectedBackground = headerBackground;
        headerWrapper.focusBackground = headerBackground;
        headerWrapper.selectedFocusBackground = headerBackground;
        headerWrapper.border = headerBorder;
        headerWrapper.selectedBorder = headerBorder;
        headerWrapper.focusBorder = headerBorder;
        headerWrapper.selectedFocusBorder = headerBorder;
    }

    /* Begin class ComponentColumn */

    /**
     * Name of the column.
     */
    protected String columnName;

    /**
     * The tooltip for the column.
     */
    protected String toolTipText;

    /**
     * The value for the column.
     */
    protected String value;

    /**
     * The number format for the values in this column.
     */
    protected NumberFormat numberFormat;

    private ComponentColumn(String columnName) {
        this.columnName = columnName;
        this.toolTipText = null;
        this.value = null;
        this.numberFormat = NumberFormat.getNumberInstance();
        this.numberFormat.setGroupingUsed(false);
        this.numberFormat.setMaximumFractionDigits(3);
    }

    /**
     * Wraps the specified object to a styled renderer's wrapper according to the state of the measurement step.
     *
     * @param value    the object to be wrapped.
     * @param rowIndex the index of the row. Can be greater than the number of measurement steps.
     * @param session  the project whose value to get. Can be null.
     * @return the wrapped object.
     */
    public StyledWrapper wrap(Object value, int rowIndex, Session session) {

        // choose the style according to the state of the step
        StyledWrapper wrapper;
        wrapper = defaultWrapper;
        wrapper.border = null;
        wrapper.selectedBorder = null;
        wrapper.focusBorder = null;
        wrapper.selectedFocusBorder = null;

        // wrap the cell's value and return it
        wrapper.value = value;
        return wrapper;
    }

    /**
     * Returns the value for this column's specified row. The default implementation is to use the algoritm of a
     * MeasurementValue object. If no MeasurementValue has been provided, will return an empty string. Subclasses can
     * override the default behaviour.
     *
     * @param rowIndex the index of the row. Can be greater than the number of measurement steps.
     * @param session  the project whose value to get. Can be null.
     * @return the wrapped value (class e.g. Boolean, StyledWrapper) that should be shown in that cell.
     */
    public Object getValue(int rowIndex, Session session) {
        return wrap(null, rowIndex, session);
    }

    public void setValue(Object value, int rowIndex, Session session) {
        // @Override this in enum's, if required
    }


    /**
     * Returns the name of this column. The name will be shown in the header of the table. The default implementation
     * returns always the columnName property. Subclasses can override the default behaviour.
     *
     * the open project or null if no project is active.
     */
    public String getColumnName() {
            return columnName;
    }

    /**
     * Returns the tooltip text for this column. The default implementation returns the description of the provided
     * MeasurementValue or null if none is provided. Subclasses can override the default behaviour.
     *
     * the open project or null if no project is active.
     * @return the tooltip text or null if none is set.
     */
    public String getToolTipText(Project project) {
            return toolTipText;
    }

    /**
     * Returns the class of this column regardless of the row. The default implementation is to return
     * StyledTableCellRenderer.Wrapper.class. Subclasses can override the default behaviour.
     */
    public Class<?> getColumnClass() {
        return StyledWrapper.class;
    }

    /**
     * Returns the number format used for rendering the numbers in this column.
     */
    public NumberFormat getNumberFormat() {
        return numberFormat;
    }

    /**
     * Sets the number format used for rendering the numbers in this column.
     *
     * @throws NullPointerException if numberFormat is null.
     */
    public void setNumberFormat(NumberFormat numberFormat) {
        if (numberFormat == null) {
            throw new NullPointerException();
        }
        this.numberFormat = numberFormat;
    }

    /**
     * Returns true if the cell can be edited. Default: false
     * Override if necessary.
     *
     * @param rowIndex    the row being queried
     * @return false
     */
    public boolean isCellEditable(int rowIndex) {
        return false;
    }

}
