/*
 * DataPrintPanel.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis.gui;

import com.intellij.uiDesigner.core.GridLayoutManager;
import com.intellij.uiDesigner.core.GridConstraints;
import com.intellij.uiDesigner.core.Spacer;

import javax.swing.*;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.text.NumberFormat;
import java.text.DateFormat;
import java.util.*;

import mavis.Project;
import mavis.Settings;
import mavis.MeasurementValue;
import mavis.Session;
import mavis.util.ComponentPrinter;

/**
 * Creates layout from measurement values to be printed.
 * PrintedPanel is preview of print and there are controls to print or cancel.
 *
 * Derived from PrintPanel written by
 * @author Aki Korpua
 *
 * Modified for mavis by
 * @author Juho Julkunen
 */
public class DataPrintPanel extends JPanel {

    private JDialog creator;
    private Project project;
    private Session session;

    private JPanel contentPane;

    private JPanel controlPanel;
    private JButton cancel;
    private JButton print;

    private JScrollPane scrollPane;
    private JPanel printedPanel;

    private JTable sequenceTable;
    private JTable differenceTable;
    private JTable componentTable;

    private TableModel sequenceTableModel;
    private TableModel differenceTableModel;
    private TableModel componentTableModel;

    private JLabel comments3;
    private JLabel comments2;
    private JLabel comments;
    private JLabel specimen;
    private JLabel area;
    private JLabel rocktype;
    private JLabel density;
    private JLabel volume;
    private JLabel mass;
    private JLabel sampleType;
    private JLabel strike;
    private JLabel qValue;
    private JLabel susceptibility;
    private JLabel longitude;
    private JLabel dip;
    private JLabel latitude;
    private JLabel operator;


    /**
    * Creates an instance of GraphPrintPanel.
    *
    * @param creator Calling Dialog
    * @param sess    The current session
    * @param proj    The currently active project
    * @param printDirectly if true print without preview, if false display preview window
    */
    public DataPrintPanel(JDialog creator, Session sess, Project proj, boolean printDirectly) {
        if (proj == null) {
            throw new NullPointerException();
        }
        this.creator = creator;
        this.session = sess;
        this.project = proj;

        NumberFormat nf = NumberFormat.getInstance();
        nf.setGroupingUsed(false);

        /* Sample Information */
        area.setText(project.getProperty(Project.AREA_PROPERTY, "") + ", " + project.getProperty(Project.SITE_PROPERTY, ""));
        area.setFont(area.getFont().deriveFont(Font.BOLD));
        String commentRow = project.getProperty(Project.COMMENT_PROPERTY, "N/A");
        commentRow = commentRow.trim();
        String [] commentRows = { "", "", "" };
        commentRows[0] = commentRow;
        int result = 51;
        if(commentRows[0].length() > 50) {
            while (result > 50) {
                result = commentRows[0].lastIndexOf(" ");
                if(result != -1) commentRows[0] = commentRows[0].substring(0, result);
                else {
                    commentRows[0] = commentRows[0].substring(0, 50);
                    result = 50;
                }
            }
            comments.setText(commentRows[0]);
            commentRows[1] = commentRow.substring(result, commentRow.length());
            if(commentRows[1].length() > 50) {
                result = 51;
                while (result > 50) {
                    result = commentRows[1].lastIndexOf(" ");
                    if(result != -1) commentRows[1] = commentRows[1].substring(0, result);
                    else {
                        commentRows[1] = commentRows[1].substring(0, 50);
                        result = 50;
                    }
                }
                comments2.setText(commentRows[1].trim());
                commentRows[2] = commentRow.substring(commentRows[0].length() + commentRows[1].length(), commentRow.length());
                if(commentRows[2].length() > 50) commentRows[2] = commentRows[2].substring(0, 50);
                comments3.setText(commentRows[2].trim());
            }
            else {
                comments2.setText(commentRows[1].trim());
                comments3.setVisible(false);
            }
        }
        else {
            comments.setText(commentRows[0]);
            comments2.setVisible(false);
            comments3.setVisible(false);
        }
        sampleType.setText(String.valueOf(project.getSampleType()));
        specimen.setText(project.getName() + " (" + project.getType().toString() + ") ");
        specimen.setFont(specimen.getFont().deriveFont(Font.BOLD));
        rocktype.setText(project.getProperty(Project.ROCK_TYPE_PROPERTY, "N/A"));
        operator.setText(project.getProperty(Project.OPERATOR_PROPERTY, "") + " / " +
                project.getProperty(Project.DATE_PROPERTY, DateFormat.getDateInstance().format(new Date())));
        nf.setMaximumFractionDigits(1);
        nf.setMinimumFractionDigits(1);
        try {
            double d = Double.parseDouble(project.getProperty(Project.LATITUDE_PROPERTY, "0.0"));
            latitude.setText(nf.format(d));
        } catch (NumberFormatException e) {
            latitude.setText("");
        }
        try {
            double d = Double.parseDouble(project.getProperty(Project.LONGITUDE_PROPERTY, "0.0"));
            longitude.setText(nf.format(d));
        } catch (NumberFormatException e) {
            longitude.setText("");
        }
        strike.setText(nf.format(project.getStrike()));
        dip.setText(nf.format(project.getDip()));
        nf.setMinimumFractionDigits(3);
        nf.setMaximumFractionDigits(3);
        mass.setText(project.getMass() > 0 ? nf.format(project.getMass()) + " g" : "");
        volume.setText(project.getVolume() > 0 ? nf.format(project.getVolume()) + " cm\u00B3" : "");
        nf.setMinimumFractionDigits(0);
        nf.setMaximumFractionDigits(0);
        susceptibility.setText(
                project.
                        getSusceptibility() != null ? nf.format(project.getSusceptibility()) + "E-6 SI" : "");

        /* calculate Density and Q */

        if (project.getStep(0) != null) {
            // first step should be with no demagnetization so its Intensity is NRM
            Double mag = project.getValue(0, MeasurementValue.MAGNETIZATION);
            Double susc = project.getSusceptibility();
            if (mag != null && susc != null && susc > 0) {
                nf.setMinimumFractionDigits(2);
                nf.setMaximumFractionDigits(2);
                double q = 25.13 * mag / susc;
                qValue.setText(nf.format(q));
            } else {
                qValue.setText("");
            }
        }
        if (project.getMass() > 0 && project.getVolume() > 0) {
            nf.setMinimumFractionDigits(0);
            nf.setMaximumFractionDigits(0);
            double d = (project.getMass() / project.getVolume()) * 1000;
            density.setText(nf.format(d) + " kg/m\u00B3");
        } else {
            density.setText("");
        }

        /* Sequence Table */
        sequenceTableModel = new PrintSequenceTableModel(project);
        /* Font reduced so values would better fit the table. More robust solution would be preferrable. */
        sequenceTable.setFont(sequenceTable.getFont().deriveFont(sequenceTable.getFont().getSize()-1.0f));
        sequenceTable.setModel(sequenceTableModel);
        sequenceTable.getTableHeader().setReorderingAllowed(false);
        sequenceTable.getTableHeader().setResizingAllowed(false);
        sequenceTable.setDefaultRenderer(StyledWrapper.class, new StyledTableCellRenderer());
        sequenceTable.setBorder(BorderFactory.createMatteBorder(sequenceTable.getIntercellSpacing().height,
                sequenceTable.getIntercellSpacing().width, 0, 0, sequenceTable.getGridColor()));
        updateColumns(sequenceTable);

        /* Difference vector table */
        differenceTableModel = new PrintDifferenceTableModel(project);
        differenceTable.setFont(differenceTable.getFont().deriveFont(differenceTable.getFont().getSize()-1.0f));
        differenceTable.setModel(differenceTableModel);
        differenceTable.getTableHeader().setReorderingAllowed(false);
        differenceTable.getTableHeader().setResizingAllowed(false);
        differenceTable.setDefaultRenderer(StyledWrapper.class, new StyledTableCellRenderer());
        differenceTable.setBorder(BorderFactory.createMatteBorder(differenceTable.getIntercellSpacing().height,
                differenceTable.getIntercellSpacing().width, 0, 0, differenceTable.getGridColor()));
        updateColumns(differenceTable);

        /* Component Table */
        componentTableModel = new PrintComponentTableModel(session);
        componentTable.setFont(componentTable.getFont().deriveFont(componentTable.getFont().getSize()-1.0f));
        componentTable.setModel(componentTableModel);
        componentTable.getTableHeader().setReorderingAllowed(false);
        componentTable.getTableHeader().setResizingAllowed(false);
        componentTable.setDefaultRenderer(StyledWrapper.class, new StyledTableCellRenderer());
        componentTable.setBorder(BorderFactory.createMatteBorder(componentTable.getIntercellSpacing().height,
                componentTable.getIntercellSpacing().width, 0, 0, componentTable.getGridColor()));
        updateColumns(componentTable);

        /* Layout */
        setLayout(new BorderLayout());
        add(contentPane, BorderLayout.CENTER);
        setMinimumSize(new Dimension(500, 700)); //magic numbers, see ComponentPrinter.java

        /* White background */
        setOpaque(printedPanel, false);
        printedPanel.getParent().setBackground(Color.WHITE);

        /* Listeners */
        print.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                closeDialog();
                ComponentPrinter.printComponent(getPrintedDocument(), DataPrintPanel.this.project.getName());
            }
        });
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                closeDialog();
            }
        });

        // HACK: for some reason printing without preview does not work
        if (printDirectly) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    print.doClick();
                }
            });
        }

    }


    /**
     * Recursively sets the opaque value of the specified JComponent and its subcomponents.
     */
    private static void setOpaque(JComponent container, boolean opaque) {
        java.util.Queue<Component> components = new LinkedList<Component>();
        components.add(container);
        Component component; // = null;
            while ((component = components.poll()) != null) {
                if (component instanceof JComponent) {
                    ((JComponent) component).setOpaque(opaque);
                }
                if (component instanceof Container) {
                    for (Component c : ((Container) component).getComponents()) {
                        components.add(c);
                    }
                }
            }
    }

    /**
     * Gets container which is meant to be printed
     *
     * @return JPanel
     */
    public JPanel getPrintedDocument() {
        return printedPanel;
    }

    /**
     * Closes this window
     */
    private void closeDialog() {
        if (creator != null) {
            creator.setVisible(false);
        }
    }

    /**
     * Resize the table's columns to fit the content.
     */
    private void updateColumns(JTable tm) {
        TableColumnModel columnModel = tm.getColumnModel();
        for (int col = 0; col < columnModel.getColumnCount(); col++) {
            // find out the column's preferred width using the actual cell contents
            int width = 20;
            Component comp;
            for (int row = 0; row < tm.getRowCount(); row++) {
                comp = tm.getCellRenderer(row, col).getTableCellRendererComponent(tm,
                        tm.getValueAt(row, col), false, false, row, col);
                width = Math.max(width, comp.getPreferredSize().width);
            }
            columnModel.getColumn(col).setPreferredWidth(width);
        }
    }

    {
// GUI initializer generated by IntelliJ IDEA GUI Designer
// >>> IMPORTANT!! <<<
// DO NOT EDIT OR ADD ANY CODE HERE!
        $$$setupUI$$$();
    }

    /**
     * Method generated by IntelliJ IDEA GUI Designer
     * >>> IMPORTANT!! <<<
     * DO NOT edit this method OR call it in your code!
     */
    private void $$$setupUI$$$() {
        contentPane = new JPanel();
        contentPane.setLayout(new GridLayoutManager(2, 1, new Insets(11, 11, 11, 11), -1, -1));
        final JPanel panel1 = new JPanel();
        panel1.setLayout(new GridLayoutManager(1, 1, new Insets(0, 0, 0, 0), -1, -1));
        contentPane.add(panel1, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), null));
        scrollPane = new JScrollPane();
        scrollPane.setHorizontalScrollBarPolicy(31);
        scrollPane.setVerticalScrollBarPolicy(22);
        panel1.add(scrollPane, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, null, null, null));
        printedPanel = new JPanel();
        printedPanel.setLayout(new GridLayoutManager(6, 1, new Insets(0, 0, 0, 0), -1, 16));
        scrollPane.setViewportView(printedPanel);
        sequenceTable = new JTable();
        sequenceTable.setEnabled(false);
        sequenceTable.setPreferredScrollableViewportSize(new Dimension(-1, -1));
        printedPanel.add(sequenceTable, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_WANT_GROW, null, null, null));
        final JPanel panel2 = new JPanel();
        panel2.setLayout(new GridLayoutManager(9, 9, new Insets(0, 0, 0, 0), 4, 4));
        printedPanel.add(panel2, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        final JLabel label1 = new JLabel();
        label1.setText("Operator / Date");
        panel2.add(label1, new GridConstraints(2, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        operator = new JLabel();
        operator.setText("N/A");
        panel2.add(operator, new GridConstraints(2, 2, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        latitude = new JLabel();
        latitude.setText("N/A");
        panel2.add(latitude, new GridConstraints(2, 5, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label2 = new JLabel();
        label2.setText("Dip");
        panel2.add(label2, new GridConstraints(4, 6, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        dip = new JLabel();
        dip.setText("N/A");
        panel2.add(dip, new GridConstraints(4, 8, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label3 = new JLabel();
        label3.setText(":");
        panel2.add(label3, new GridConstraints(2, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label4 = new JLabel();
        label4.setText(":");
        panel2.add(label4, new GridConstraints(2, 4, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label5 = new JLabel();
        label5.setText(":");
        panel2.add(label5, new GridConstraints(4, 7, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label6 = new JLabel();
        label6.setText("Comments");
        panel2.add(label6, new GridConstraints(6, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label7 = new JLabel();
        label7.setText(":");
        panel2.add(label7, new GridConstraints(6, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label8 = new JLabel();
        label8.setText(":");
        panel2.add(label8, new GridConstraints(3, 4, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        longitude = new JLabel();
        longitude.setText("N/A");
        panel2.add(longitude, new GridConstraints(3, 5, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label9 = new JLabel();
        label9.setText(":");
        panel2.add(label9, new GridConstraints(4, 4, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        susceptibility = new JLabel();
        susceptibility.setText("N/A");
        panel2.add(susceptibility, new GridConstraints(4, 5, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        qValue = new JLabel();
        qValue.setText("N/A");
        panel2.add(qValue, new GridConstraints(5, 5, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label10 = new JLabel();
        label10.setText(":");
        panel2.add(label10, new GridConstraints(5, 4, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label11 = new JLabel();
        label11.setText("Strike");
        panel2.add(label11, new GridConstraints(3, 6, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label12 = new JLabel();
        label12.setText(":");
        panel2.add(label12, new GridConstraints(3, 7, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        strike = new JLabel();
        strike.setText("N/A");
        panel2.add(strike, new GridConstraints(3, 8, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label13 = new JLabel();
        label13.setText("Sample type");
        panel2.add(label13, new GridConstraints(2, 6, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label14 = new JLabel();
        label14.setText(":");
        panel2.add(label14, new GridConstraints(2, 7, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        sampleType = new JLabel();
        sampleType.setText("N/A");
        panel2.add(sampleType, new GridConstraints(2, 8, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label15 = new JLabel();
        label15.setText("Latitude");
        panel2.add(label15, new GridConstraints(2, 3, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label16 = new JLabel();
        label16.setText("Longitude");
        panel2.add(label16, new GridConstraints(3, 3, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label17 = new JLabel();
        label17.setText("Susceptibility");
        panel2.add(label17, new GridConstraints(4, 3, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label18 = new JLabel();
        label18.setText("Q");
        panel2.add(label18, new GridConstraints(5, 3, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label19 = new JLabel();
        label19.setText("Mass ");
        panel2.add(label19, new GridConstraints(3, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label20 = new JLabel();
        label20.setText(":");
        panel2.add(label20, new GridConstraints(3, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        mass = new JLabel();
        mass.setText("N/A");
        panel2.add(mass, new GridConstraints(3, 2, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label21 = new JLabel();
        label21.setText("Volume");
        panel2.add(label21, new GridConstraints(4, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label22 = new JLabel();
        label22.setText(":");
        panel2.add(label22, new GridConstraints(4, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        volume = new JLabel();
        volume.setText("N/A");
        panel2.add(volume, new GridConstraints(4, 2, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label23 = new JLabel();
        label23.setText("Density");
        panel2.add(label23, new GridConstraints(5, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label24 = new JLabel();
        label24.setText(":");
        panel2.add(label24, new GridConstraints(5, 1, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        density = new JLabel();
        density.setText("N/A");
        panel2.add(density, new GridConstraints(5, 2, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label25 = new JLabel();
        label25.setText("Rock type");
        panel2.add(label25, new GridConstraints(5, 6, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label26 = new JLabel();
        label26.setText(":");
        panel2.add(label26, new GridConstraints(5, 7, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        rocktype = new JLabel();
        rocktype.setText("N/A");
        panel2.add(rocktype, new GridConstraints(5, 8, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        area = new JLabel();
        area.setText("N/A");
        panel2.add(area, new GridConstraints(0, 3, 1, 4, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        specimen = new JLabel();
        specimen.setText("N/A");
        panel2.add(specimen, new GridConstraints(1, 3, 1, 4, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label27 = new JLabel();
        label27.setText("Specimen : ");
        panel2.add(label27, new GridConstraints(1, 1, 1, 2, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label28 = new JLabel();
        label28.setHorizontalAlignment(10);
        label28.setText("Area, Site : ");
        panel2.add(label28, new GridConstraints(0, 1, 1, 2, GridConstraints.ANCHOR_EAST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        comments = new JLabel();
        comments.setText("N/A");
        panel2.add(comments, new GridConstraints(6, 2, 1, 7, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        comments2 = new JLabel();
        comments2.setText("...");
        panel2.add(comments2, new GridConstraints(7, 2, 1, 7, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        comments3 = new JLabel();
        comments3.setText("...");
        panel2.add(comments3, new GridConstraints(8, 2, 1, 7, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        differenceTable = new JTable();
        differenceTable.setEnabled(false);
        differenceTable.setPreferredScrollableViewportSize(new Dimension(-1, -1));
        printedPanel.add(differenceTable, new GridConstraints(3, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_WANT_GROW, null, null, null));
        componentTable = new JTable();
        componentTable.setEnabled(false);
        componentTable.setPreferredScrollableViewportSize(new Dimension(-1, -1));
        printedPanel.add(componentTable, new GridConstraints(5, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_WANT_GROW, null, null, null));
        final JLabel label29 = new JLabel();
        label29.setText("Difference Vectors");
        printedPanel.add(label29, new GridConstraints(2, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final JLabel label30 = new JLabel();
        label30.setText("Calculated Components");
        printedPanel.add(label30, new GridConstraints(4, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        controlPanel = new JPanel();
        controlPanel.setLayout(new GridLayoutManager(1, 3, new Insets(0, 0, 4, 0), -1, -1));
        contentPane.add(controlPanel, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        print = new JButton();
        print.setText("Print");
        controlPanel.add(print, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        cancel = new JButton();
        cancel.setText("Cancel");
        controlPanel.add(cancel, new GridConstraints(0, 2, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final Spacer spacer1 = new Spacer();
        controlPanel.add(spacer1, new GridConstraints(0, 1, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_WANT_GROW, 1, null, null, null));
    }


    /**
       * Shows the the data of a project in printable version. Uses the contents of the MeasurementSequenceTableModel.
       *
       * @author Esko Luontola
       */
    private class PrintSequenceTableModel extends AbstractTableModel {

        private StyledWrapper wrapper = new StyledWrapper();

        private MeasurementSequenceTableModel model;

        public PrintSequenceTableModel(Project project) {
            model = new MeasurementSequenceTableModel();
             model.changeNumberFormat(Settings.getPrintoutNumberFormat());
             model.setProject(project);
             wrapper.horizontalAlignment = SwingConstants.TRAILING;
        }

        public int getRowCount() {
            if (model.getProject().isSequenceEditEnabled()) {
                return model.getRowCount();         // should never be editable
            } else {
                return model.getRowCount();     // the first row is for the header
            }                                   // do not show origin row
        }

        public int getColumnCount() {
            return model.getColumnCount();
        }

        @Override public Class<?> getColumnClass(int columnIndex) {
            return StyledWrapper.class;
        }

        public Object getValueAt(int rowIndex, int columnIndex) {
            if (rowIndex == 0) {
                wrapper.value = model.getColumnName(columnIndex);
                return wrapper;
            } else {
                Object obj = model.getValueAt(rowIndex - 1, columnIndex);
                if (obj instanceof StyledWrapper) {
                    StyledWrapper wrapper = (StyledWrapper) obj;
                    obj = wrapper.value;
                }
                wrapper.value = obj;
                return wrapper;
            }
        }
    }

    /**
     * Shows the details of the magnetic components calculated from sample data
     * in printable version. Uses the contents of the MagneticComponentTableModel.
     *
     * @author Juho Julkunen (derived from PrintSequenceTableModel written by Esko Luontola).
     */
    private class PrintComponentTableModel extends AbstractTableModel {

        private StyledWrapper wrapper = new StyledWrapper();

        private MagneticComponentTableModel model;

        public PrintComponentTableModel(Session sess) {
            model = new MagneticComponentTableModel();
            model.changeNumberFormat(Settings.getPrintoutNumberFormat());
            model.setSession(sess);
            wrapper.horizontalAlignment = SwingConstants.TRAILING;
        }

        public int getRowCount() {
            return model.getRowCount() + 1;
        }

        public int getColumnCount() {
            return model.getColumnCount();
        }

        @Override public Class<?> getColumnClass(int columnIndex) {
            return StyledWrapper.class;
        }

        public Object getValueAt(int rowIndex, int columnIndex) {
            if (rowIndex == 0) {
                wrapper.value = model.getColumnName(columnIndex);
                return wrapper;
            } else {
                Object obj = model.getValueAt(rowIndex - 1, columnIndex);
                if (obj instanceof StyledWrapper) {
                    StyledWrapper wrapper = (StyledWrapper) obj;
                    obj = wrapper.value;
                }
                wrapper.value = obj;
                return wrapper;
            }
        }
    }

    /**
     * Shows the difference vectors calculated from sample data in printable version.
     * Uses the contents of the DifferenceVectorTableModel.
     *
     * @author Juho Julkunen (derived from PrintSequenceTableModel written by Esko Luontola).
     */
    private class PrintDifferenceTableModel extends AbstractTableModel {

        private StyledWrapper wrapper = new StyledWrapper();

        private DifferenceVectorTableModel model;

        public PrintDifferenceTableModel(Project p) {
            model = new DifferenceVectorTableModel();
            model.changeNumberFormat(Settings.getPrintoutNumberFormat());
            model.setProject(p);
            wrapper.horizontalAlignment = SwingConstants.TRAILING;

        }

        public int getRowCount() {
            return model.getRowCount();
        }

        public int getColumnCount() {
            return model.getColumnCount();
        }

        @Override public Class<?> getColumnClass(int columnIndex) {
            return StyledWrapper.class;
        }

        public Object getValueAt(int rowIndex, int columnIndex) {
            if (rowIndex == 0) {
                wrapper.value = model.getColumnName(columnIndex);
                return wrapper;
            } else {
                Object obj = model.getValueAt(rowIndex - 1, columnIndex);
                if (obj instanceof StyledWrapper) {
                    StyledWrapper wrapper = (StyledWrapper) obj;
                    obj = wrapper.value;
                }
                wrapper.value = obj;
                return wrapper;
            }
        }
    }

}
