/*
 * ProjectExplorerPanel.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis.gui;

import mavis.*;
import mavis.util.LastExecutor;

import javax.swing.*;
import javax.swing.text.BadLocationException;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.FileFilter;
import java.util.Arrays;

import com.intellij.uiDesigner.core.GridLayoutManager;
import com.intellij.uiDesigner.core.GridConstraints;
import com.intellij.uiDesigner.core.Spacer;

/**
 * Creates a history/autocomplete field (browserField) for choosing the save directory, a button for saving the
 * component table, a button for printing and checkboxes to choose which components will be printed. 
 * Derived from ProjectExplorer.java
 *
 * @author Janne Laukkarinen (originally by Samuli Kaipiainen)
 */
public class FileActionsPanel extends ProjectComponent {

    /**
     * Holds all the components
     */
    private JPanel contentPane;

    /**
     * Text field for writing directory to change to. Autocomplete results appear to Combo Box' popup window, scheduled
     * by LastExecutor. Directory history appears to the same popup window when the down-arrow right to text field is
     * clicked.
     */
    private JComboBox browserField;
    private FittedComboBoxRenderer browserFieldRenderer;
    private JTextField browserFieldEditor; // WARNING: look-and-feel-dependant code
    private ComponentFlasher browserFieldFlasher;

    /**
     * Tells whether current popup menu is autocomplete list (and not directory history).
     */
    boolean browserFieldPopupIsAutocomplete = false;

    /**
     * Tells whether the next-to-be-shown popup menu will be autocomplete list (and not directory history).
     */
    // private boolean browserFieldNextPopupIsAutocomplete = false;

    /**
     * Tells whether browserField's popup menu list is being updated, and we don't want those ActionEvents.
     */
    boolean browserFieldUpdatingPopup = false;

    /**
     * Button for choosing the save directory
     */
    private JButton browseButton;

    /**
     * Buttons for saving and printing
     */
    private JButton saveToTableButton;
    private JButton printButton;
    private JButton printPreviewButton;

    /**
     * LastExecutor for scheduling autocomplete results to separate thread (disk access and displaying).
     */
    private LastExecutor autocompleteExecutor = new LastExecutor(100, true);

    /**
     * Currently open directory.
     */
    private File directory = null;

    /**
     * Currently open CSV-file.
     */
    private File file = null;

    /**
     * Checkboxes that tell what information will be printed
     */
    private JCheckBox sampleDataBox;
    private JCheckBox graphsBox;

    /**
     * The extension for component table files.
     */
    private static final String FILE_TYPE = "csv";

    /**
     * Creates all components, sets directory as the last open directory or opened project's directory, initializes
     * files with files from that directory.
     */
    public FileActionsPanel() {

        // combo box / text field
//        browserField = new JComboBox(getDirectoryHistory());
        browserField.setBackground(Color.WHITE);
        browserFieldEditor = (JTextField) browserField.getEditor().getEditorComponent();
        browserFieldFlasher = new ComponentFlasher(browserFieldEditor);
        browserFieldEditor.setFocusTraversalKeysEnabled(false); // disable tab-exiting from browserField
        // custom renderer for browserField's items so that long path names in the popup menu are made shorter
        browserFieldRenderer = new FittedComboBoxRenderer(browserFieldEditor);
        browserField.setRenderer(browserFieldRenderer);

        // get the saved states for the checkboxes
        sampleDataBox.setSelected(Settings.getSampleDataCheckBox());
        graphsBox.setSelected(Settings.getGraphsCheckBox());
        printButton.setEnabled(sampleDataBox.isSelected() || graphsBox.isSelected());
        printPreviewButton.setEnabled(sampleDataBox.isSelected() || graphsBox.isSelected());

        this.setLayout(new BorderLayout());
        this.add(contentPane, "Center");
        contentPane.setBorder(BorderFactory.createEmptyBorder(0, 4, 8, 4));

        // set the directory to the previous selected directory
        setDirectory(Settings.getLastTableDirectory());
        saveToTableButton.setEnabled(false);
        browseButton.setEnabled(false);
        // FileActionsPanel events

        /**
         * Event D: On browseButton click - open a FileChooser dialog for choosing new directory,
         * set it to directory, update files listing, update explorerTable and browserField.
         */
        browseButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JFileChooser fc = new JFileChooser(directory);
                fc.setDialogTitle("Change directory");
                fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
                fc.setFileFilter(new GenericFileFilter(FILE_TYPE.toUpperCase() + " File", FILE_TYPE));
                fc.setSelectedFile(new File(directory, getProject().getName()));
                fc.setApproveButtonText("Save");
                fc.setApproveButtonToolTipText("Save the specified file");
                if (fc.showOpenDialog(browseButton) == JFileChooser.APPROVE_OPTION) {
                    File file = fc.getSelectedFile();
                    if (!file.getName().toLowerCase().endsWith("." + FILE_TYPE)) {
                        file = new File(file.getAbsolutePath() + "." + FILE_TYPE);
                    }
                    getSession().saveToTable(file);
                    //setDirectory(file.getParentFile());
                    setFile(file);
                    Settings.updateTableDirectoryHistory(file.getParentFile());
                    setBrowserFieldPopup(getDirectoryHistory(), false);
                }
            }
        });

        /**
         * Event C: On browserField popup window click - set clicked line as directory, update files
         * listing, update explorerTable and browserField.
         */
        browserField.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // check if event is caused by popup menu update
                if (browserFieldUpdatingPopup) {
                    return;
                }

                // JComboBox event types:
                // Edited -- pressed enter in the text field or selected an item by pressing enter
                // Changed -- cycled item list with up/down, selected item with mouse or pressed enter in text field
                // System.out.println(e.getActionCommand() + ": " + browserField.getSelectedItem());

                // we only want changed-events, not duplicate edited-events -- actually, we want those too,
                // so that multiple tries to change to an invalid directory with enter will flash text field
                // if (!e.getActionCommand().equals("comboBoxChanged")) return;

                // TODO: cycling through popup menu list with up/down keys changes directory;
                // it shouldn't, but can't recognize those changed-events from mouse clicks

                // item is File if selected from list, String if written to text field
                Object item = browserField.getEditor().getItem();
                File dir = item instanceof File ? (File) item : new File((String) item).getAbsoluteFile();

                // try to set directory, flash browserField red if error
                if (!setFile(dir)) {
                    browserFieldFlasher.flash();
                }
            }
        });

        /**
         * Event B: On browserField down-arrow-click - show directory history in browserField's popup window.
         */
        browserField.addPopupMenuListener(new PopupMenuListener() {
            public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
                // set popup menu as directory history when closing popup, unless it's already dir history
                if (browserFieldPopupIsAutocomplete) {
                    browserFieldPopupIsAutocomplete = false;
                    Object item = browserField.getSelectedItem();
                    setBrowserFieldPopup(getDirectoryHistory(), false);
                    browserField.setSelectedItem(item);
                }
            }

            public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
            }

            public void popupMenuCanceled(PopupMenuEvent e) {
            }
        });

        /**
         * Event A: On browserField change - send autocomplete-results-finder with browserField's text
         * to LastExecutor via autocompleteExecutor.execute(Runnable), which schedules disk access and
         * displaying autocomplete results in browserField's popup window.
         */
        browserFieldEditor.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                if (e.getKeyCode() == KeyEvent.VK_TAB) {
                    // change to down- or up-arrow event
                    if ((e.getModifiers() & KeyEvent.SHIFT_MASK) != 0) {
                        // this deprecation is required for shift-tab to work
                        e.setModifiers(e.getModifiers() ^ KeyEvent.SHIFT_MASK);
                        e.setKeyCode(KeyEvent.VK_UP);
                    } else {
                        e.setKeyCode(KeyEvent.VK_DOWN);
                    }
                }
            }

            public void keyTyped(KeyEvent e) {
                if (e.getKeyChar() == KeyEvent.VK_ESCAPE) {
                    browserField.setSelectedItem(directory);
                    // browserField.getEditor().selectAll();
                    return;
                } else if (e.getKeyChar() == KeyEvent.VK_ENTER) {
                    return;
                } else if (e.getKeyChar() == KeyEvent.VK_TAB) return;

                if (e.getModifiers() == KeyEvent.CTRL_MASK && e.getKeyChar() == KeyEvent.VK_DELETE) {
                    // delete one directory name at a time
                    int pos = browserFieldEditor.getCaretPosition();
                    String text = browserFieldEditor.getText();
                    String textA = text.substring(0, pos);
                    String textB = text.substring(pos);
                    textA = textA.substring(0, Math.max(0, textA.lastIndexOf(System.getProperty("file.separator"))));
                    browserFieldEditor.setText(textA + textB);
                    browserFieldEditor.setCaretPosition(textA.length());

                } else if ((e.getModifiers() & KeyEvent.ALT_MASK) != 0 || (e.getModifiers() & KeyEvent.CTRL_MASK) != 0) {
                    // avoid the popup menu from showing, when the Project Explorer tab is hidden with ALT+P
                    browserField.hidePopup();
                    return;
                }

                autocompleteExecutor.execute(new Runnable() {
                    public void run() {
                        doAutoComplete();
                    }
                });
            }
        });

        // When the saveToTableButton is pressed, get the directory and the file name for the component table and save
        saveToTableButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (getProject() == null) {
                    return;
                }
                String fileName;
                try {
                    fileName = browserFieldEditor.getDocument().getText(0, browserFieldEditor.getDocument().getLength());
                } catch (BadLocationException ex) {
                    fileName = "";
                }
                if (fileName.contains("?") || fileName.contains("*") || fileName.contains(">") || fileName.contains("<")) {
                    System.out.println("Cannot write to file " + fileName);
                    return;
                }
                File file;
                if (fileName.equals("")) {
                    String dir = directory.getAbsolutePath();
                    // file name for the table
                    String name = getProject().getName() + "." + FILE_TYPE;
                    file = new File(dir, name);
                } else {
                    file = new File(fileName);
                    if (file.isDirectory()) {
                        String name = getProject().getName() + "." + FILE_TYPE;
                        file = new File(file.getAbsolutePath(), name);
                    } else {
                        if (!file.getName().toLowerCase().endsWith("." + FILE_TYPE)) {
                             file = new File(file.getAbsolutePath() + "." + FILE_TYPE);
                        }
                    }
                }
                // Save the component table
                getSession().saveToTable(file);
                setFile(file);
                // Update the table directory history
                Settings.updateTableDirectoryHistory(file.getParentFile());
                setBrowserFieldPopup(getDirectoryHistory(), false);
            }
        });

        // When the printButton is clicked, print the sample data and/or the graphs
        printButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if ( sampleDataBox.isSelected() )
                  SettingsDialog.showDataPrintPreview(getParentFrame(), "Print Preview", getSession(), getProject(), true);
                if ( graphsBox.isSelected() )
                  SettingsDialog.showGraphPrintPreview(getParentFrame(), "Print Preview", getSession(), getProject(), true);
            }
        });

        // Display print preview window.
        printPreviewButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if ( sampleDataBox.isSelected() )
                  SettingsDialog.showDataPrintPreview(getParentFrame(), "Print Preview", getSession(), getProject(), false);
                if ( graphsBox.isSelected() )
                  SettingsDialog.showGraphPrintPreview(getParentFrame(), "Print Preview", getSession(), getProject(), false);
            }
        });

        // The states of the checkboxes are saved when they are changed
        sampleDataBox.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Settings.setSampleDataCheckBox(sampleDataBox.isSelected());
                boolean enabled = sampleDataBox.isSelected() || graphsBox.isSelected();
                printButton.setEnabled(enabled);
                printPreviewButton.setEnabled(enabled);
            }
        });

        graphsBox.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Settings.setGraphsCheckBox(graphsBox.isSelected());
                boolean enabled = sampleDataBox.isSelected() || graphsBox.isSelected();
                printButton.setEnabled(enabled);
                printPreviewButton.setEnabled(enabled);
            }
        });
    }

    /**
     * Call super.setProject(project), highlight selected project, or unhighlight unselected project.
     *
     * @param project project opened, or null to open no project.
     */
    public void setProject(Project project) {
        super.setProject(project);

        // update directory history, as it might have changed
        browserField.hidePopup();
        setBrowserFieldPopup(getDirectoryHistory(), false);
    }

    /**
     * Attempts to change to the given directory. Updates browserField with new directory.
     *
     * @param directory directory to change to.
     * @return true if succesful, false otherwise.
     */
    private boolean setDirectory(File directory) {
        if (directory == null || !directory.isDirectory()) {
            return false;
        }
        this.directory = directory;
        // update browserField with new directory
        browserField.setSelectedItem(this.directory);
        return true;
    }

    /**
     * Attempts to change to the given file. If the file is a directory, change to that directory. Updates browserField
     * with the new file.
     * @param file The file to change to.
     * @return true if successful, false otherwise.
     */
    private boolean setFile(File file) {
        if (file == null) {
            return false;
        }
        if (file.isDirectory()) {
            setDirectory(file);
            return true;
        }
        this.file = file;
        browserField.setSelectedItem(this.file);
        return true;
    }

    /**
     * Reads current directory history from Settings.
     *
     * @return current directory history. Should never return null.
     */
    private File[] getDirectoryHistory() {
        return Settings.getTableDirectoryHistory();
    }

    /**
     * Reads matching directories from given directory name's parent.
     *
     * @param dirmatch beginning of directory to which match the directories in its parent directory...
     * @return matching directories.
     */
    private File[] getAutocompleteFiles(String dirmatch) {
        File dirfile = new File(dirmatch);
        if (!dirfile.isAbsolute()) {
            return File.listRoots();
        }
        File dir = dirfile.isDirectory() ? dirfile : dirfile.getParentFile();

        // protect against no-parent-null and invalid-dir-list-null
        if (dir == null) {
            dir = directory;
        } else if (!dir.isDirectory()) {
            return new File[0];
        }

        final String match = dirfile.isDirectory() ? "" : dirfile.getName();

        File[] files = dir.listFiles(new FileFilter() {
            public boolean accept(File file) {
                return (file.isDirectory() && file.getName().toLowerCase().startsWith(match.toLowerCase()));
            }
        });

        // listFiles() does not guarantee any order for the returned files, so they must be sorted first
        Arrays.sort(files);
        return files;
    }

    /**
     * Updates autocomplete popup-menu.
     */
    private void doAutoComplete() {
        final File[] files = getAutocompleteFiles(browserField.getEditor().getItem().toString());

        // gui updating must be done from event-dispatching thread
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                // when the popup is hidden before showing, it will be automatically resized
                // (disable change-to-dirhistory-on-popup-hide for this)
                browserFieldPopupIsAutocomplete = false;
                browserField.hidePopup();

                browserFieldPopupIsAutocomplete = true;
                setBrowserFieldPopup(files, true);
                if (files.length > 0) {
                    browserField.showPopup();
                }
            }
        });
    }

    /**
     * Sets browserField popup-menu-list as given files; also clears any selection.
     *
     * @param files      list of files to set the list to.
     * @param uniformFit true if all the file paths should be cut short from the same directory, false otherwise.
     */
    private void setBrowserFieldPopup(File[] files, boolean uniformFit) {
        // purkkaillaan -- some hardcore bubblegum stitching (the whole method)
        browserFieldUpdatingPopup = true;

        String browserFieldEditorText = browserFieldEditor.getText();
        int browserFieldEditorCursorPosition = browserFieldEditor.getCaretPosition();

        browserField.removeAllItems();
        for (File file : files) {
            browserField.addItem(file);
        }

        // fit the contents of the menu list
        if (uniformFit) {
            int maxFitValue = 0;
            for (int i = 0; i < browserField.getItemCount(); i++) {
                Object item = browserField.getItemAt(i);
                int fitValue = browserFieldRenderer.fitValue(item);
                maxFitValue = Math.max(maxFitValue, fitValue);
            }
            browserFieldRenderer.setFitLimit(maxFitValue);
        } else {
            browserFieldRenderer.setFitLimit(-1);
        }

        browserField.setSelectedIndex(-1);
        browserFieldEditor.setText(browserFieldEditorText);
        browserFieldEditor.setCaretPosition(browserFieldEditorCursorPosition);

        browserFieldUpdatingPopup = false;
    }

    /**
     * Will be invoked whenever a session event happens.
     * @param event the event that happened.
     */
    @Override public void sessionUpdated(SessionEvent event) {
        if (event != null) {
            Session s = event.getSession();
            SessionEvent.Type type = event.getType();
            if (SessionEvent.Type.COMPONENT_ADD_NEW == type) {
                saveToTableButton.setEnabled(true);
                browseButton.setEnabled(true);
            } else if (SessionEvent.Type.DATA_CHANGE == type) {
                saveToTableButton.setEnabled(true);
                browseButton.setEnabled(true);
            } else if (SessionEvent.Type.COMPONENT_REMOVE == type) {
                saveToTableButton.setEnabled(s.getNumOfComponents() > 0);
                browseButton.setEnabled(s.getNumOfComponents() > 0);
            }
        }
    }

    /**
     * Sets browserField's cursor to text field's (right) end.
     *
     * @deprecated not needed anymore; cursor seems to be there anyway?
     */
    private void setBrowserFieldCursorToEnd() {
        browserFieldEditor.setCaretPosition(browserFieldEditor.getDocument().getLength());
    }

    {
// GUI initializer generated by IntelliJ IDEA GUI Designer
// >>> IMPORTANT!! <<<
// DO NOT EDIT OR ADD ANY CODE HERE!
        $$$setupUI$$$();
    }

    /**
     * Method generated by IntelliJ IDEA GUI Designer
     * >>> IMPORTANT!! <<<
     * DO NOT edit this method OR call it in your code!
     */
    private void $$$setupUI$$$() {
        contentPane = new JPanel();
        contentPane.setLayout(new GridLayoutManager(3, 6, new Insets(0, 0, 0, 0), -1, -1));
        saveToTableButton = new JButton();
        saveToTableButton.setText("Save to Table");
        contentPane.add(saveToTableButton, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        printButton = new JButton();
        printButton.setText("Print");
        contentPane.add(printButton, new GridConstraints(2, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final Spacer spacer1 = new Spacer();
        contentPane.add(spacer1, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_VERTICAL, 1, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(-1, 4), null));
        browseButton = new JButton();
        browseButton.setText("Save As...");
        contentPane.add(browseButton, new GridConstraints(0, 5, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        browserField = new JComboBox();
        browserField.setEditable(true);
        contentPane.add(browserField, new GridConstraints(0, 3, 1, 2, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, new Dimension(160, -1), new Dimension(160, -1), new Dimension(160, -1)));
        final JLabel label1 = new JLabel();
        label1.setText("Print:");
        contentPane.add(label1, new GridConstraints(2, 3, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        sampleDataBox = new JCheckBox();
        sampleDataBox.setText("Sample Data");
        contentPane.add(sampleDataBox, new GridConstraints(2, 4, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        graphsBox = new JCheckBox();
        graphsBox.setText("Graphs");
        contentPane.add(graphsBox, new GridConstraints(2, 5, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        printPreviewButton = new JButton();
        printPreviewButton.setText("Preview");
        contentPane.add(printPreviewButton, new GridConstraints(2, 1, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final Spacer spacer2 = new Spacer();
        contentPane.add(spacer2, new GridConstraints(2, 2, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_WANT_GROW, 1, null, null, null));
        final Spacer spacer3 = new Spacer();
        contentPane.add(spacer3, new GridConstraints(0, 2, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_WANT_GROW, 1, null, null, null));
    }
}
