/*
 * GraphPrintPanel.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis.gui;

import com.intellij.uiDesigner.core.GridLayoutManager;
import com.intellij.uiDesigner.core.GridConstraints;
import com.intellij.uiDesigner.core.Spacer;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.*;

import mavis.Project;
import mavis.Session;
import mavis.util.ComponentPrinter;


/**
 * Creates layout from Plots to be printed. PrintedPanel is preview of print and
 * there are controls to print or cancel.
 *
 * Derived from PrintPanel written by
 * @author Aki Korpua
 *
 * Modified for mavis by
 * @author Juho Julkunen
 */
public class GraphPrintPanel extends JPanel {

    private JDialog creator;
    private Project project;
    private Session session;

    private JPanel contentPane;
    private JPanel controlPane;

    private JButton cancel;
    private JButton print;

    private JScrollPane scrollPane;
    private JPanel printedPanel;

    private JLabel diffLabel;
    private JLabel zLabel;
    private JLabel stereoLabel;
    private JLabel intensityLabel;

    private JPanel difference;
    private JPanel stereo;
    private JPanel zijderveld;
    private JPanel intensity;

    private AbstractPlot iPlot;
    private AbstractPlot sPlot;
    private ZijderveldPlot zPlot;
    private AbstractPlot dPlot;

    private Vector<AbstractPlot> plots = new Vector<AbstractPlot>();

    /**
     * Creates an instance of GraphPrintPanel.
     *
     * @param creator Calling Dialog
     * @param sess    The current session
     * @param project The currently active project
     * @param printDirectly if true print without preview, if false display preview window
     */
     public GraphPrintPanel(JDialog creator, Session sess, Project project, boolean printDirectly) {
        if (project == null) {
            throw new NullPointerException();
        }
        this.creator = creator;
        this.project = project;
        this.session = sess;

        /* Plots */
        iPlot = new IntensityPlot();
        sPlot = new StereoPlot();
        zPlot = new ZijderveldPlot();
        /* todo: Add DifferenceVectorPlot to printout. */
        //dPlot = new StereoPlot();

        plots.add(iPlot);
        plots.add(sPlot);
        plots.add(zPlot);
        //plots.add(dPlot);

        intensity.setLayout(new BorderLayout());
        stereo.setLayout(new BorderLayout());
        zijderveld.setLayout(new BorderLayout());
        difference.setLayout(new BorderLayout());

        intensity.add(iPlot, "Center");
        stereo.add(sPlot, "Center");
        zijderveld.add(zPlot, "Center");
        //difference.add(dPlot, "Center");

        intensity.setPreferredSize(new Dimension(200, 200));
        stereo.setPreferredSize(new Dimension(200, 200));
        zijderveld.setPreferredSize(new Dimension(200, 200));
        difference.setPreferredSize(new Dimension(200, 200));

        /* Add points from the current sample's measurement results to plots. */
        for (Plot plot : plots) {
            plot.reset();
            if (project != null) {
                for (int i = 0; i < project.getSteps(); i++) {
                    plot.add(project.getStep(i));
                }
            }
        }

        /* Add calculated components from current session to Zijderveld Plot. */
        zPlot.updateLine(session); 

        /* Layout */
        setLayout(new BorderLayout());
        add(contentPane, BorderLayout.CENTER);
        setMinimumSize(new Dimension(500, 700)); // Magic numbers, see ComponentPrinter.java

        /* White background */
        setOpaque(printedPanel, false);
        printedPanel.getParent().setBackground(Color.WHITE);

        /* Listeners */
        print.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                closeDialog();
                ComponentPrinter.printComponent(getPrintedDocument(), GraphPrintPanel.this.project.getName());
            }
        });
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                closeDialog();
            }
        });

        // HACK: for some reason printing without preview does not work
        if (printDirectly) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    print.doClick();
                }
            });
        }
    }

    /**
     * Recursively sets the opaque value of the specified JComponent and its subcomponents.
     */
    private static void setOpaque(JComponent container, boolean opaque) {
        java.util.Queue<Component> components = new LinkedList<Component>();
        components.add(container);
        Component component = null;
        while ((component = components.poll()) != null) {
            if (component instanceof JComponent) {
                ((JComponent) component).setOpaque(opaque);
            }
            if (component instanceof Container) {
                for (Component c : ((Container) component).getComponents()) {
                    components.add(c);
                }
            }
        }
    }

    /**
     * Gets container which is meant to be printed
     *
     * @return JPanel
     */
    public JPanel getPrintedDocument() {
        return printedPanel;
    }

    /**
     * Closes this window
     */
    private void closeDialog() {
        if (creator != null) {
            creator.setVisible(false);
        }
    }

    {
// GUI initializer generated by IntelliJ IDEA GUI Designer
// >>> IMPORTANT!! <<<
// DO NOT EDIT OR ADD ANY CODE HERE!
        $$$setupUI$$$();
    }

    /**
     * Method generated by IntelliJ IDEA GUI Designer
     * >>> IMPORTANT!! <<<
     * DO NOT edit this method OR call it in your code!
     */
    private void $$$setupUI$$$() {
        contentPane = new JPanel();
        contentPane.setLayout(new GridLayoutManager(3, 1, new Insets(11, 11, 11, 11), -1, -1));
        final JPanel panel1 = new JPanel();
        panel1.setLayout(new GridLayoutManager(1, 1, new Insets(0, 0, 0, 0), -1, -1));
        contentPane.add(panel1, new GridConstraints(2, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), null));
        scrollPane = new JScrollPane();
        scrollPane.setHorizontalScrollBarPolicy(31);
        scrollPane.setVerticalScrollBarPolicy(22);
        panel1.add(scrollPane, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, null, null, null));
        printedPanel = new JPanel();
        printedPanel.setLayout(new GridLayoutManager(4, 2, new Insets(0, 0, 0, 0), -1, 16));
        scrollPane.setViewportView(printedPanel);
        intensity = new JPanel();
        printedPanel.add(intensity, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        zijderveld = new JPanel();
        printedPanel.add(zijderveld, new GridConstraints(3, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        stereo = new JPanel();
        printedPanel.add(stereo, new GridConstraints(1, 1, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        difference = new JPanel();
        printedPanel.add(difference, new GridConstraints(3, 1, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, null, null, null));
        intensityLabel = new JLabel();
        intensityLabel.setHorizontalAlignment(0);
        intensityLabel.setHorizontalTextPosition(0);
        intensityLabel.setText("Intensity plot");
        printedPanel.add(intensityLabel, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_NORTH, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        stereoLabel = new JLabel();
        stereoLabel.setFocusTraversalPolicyProvider(true);
        stereoLabel.setHorizontalAlignment(0);
        stereoLabel.setHorizontalTextPosition(0);
        stereoLabel.setText("Stereoplot");
        printedPanel.add(stereoLabel, new GridConstraints(0, 1, 1, 1, GridConstraints.ANCHOR_NORTH, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        zLabel = new JLabel();
        zLabel.setFocusTraversalPolicyProvider(false);
        zLabel.setHorizontalAlignment(0);
        zLabel.setHorizontalTextPosition(0);
        zLabel.setText("Zijderveld Plot");
        printedPanel.add(zLabel, new GridConstraints(2, 0, 1, 1, GridConstraints.ANCHOR_NORTH, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        diffLabel = new JLabel();
        diffLabel.setHorizontalAlignment(0);
        diffLabel.setHorizontalTextPosition(0);
        diffLabel.setText("");
        printedPanel.add(diffLabel, new GridConstraints(2, 1, 1, 1, GridConstraints.ANCHOR_NORTH, GridConstraints.FILL_NONE, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        controlPane = new JPanel();
        controlPane.setLayout(new GridLayoutManager(1, 3, new Insets(0, 0, 4, 0), -1, -1));
        contentPane.add(controlPane, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        print = new JButton();
        print.setText("Print");
        controlPane.add(print, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        cancel = new JButton();
        cancel.setText("Cancel");
        controlPane.add(cancel, new GridConstraints(0, 2, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final Spacer spacer1 = new Spacer();
        controlPane.add(spacer1, new GridConstraints(0, 1, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_WANT_GROW, 1, null, null, null));
    }
}
