package mavis.gui;

import mavis.*;

import javax.swing.*;
import javax.swing.table.TableColumnModel;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;

import com.intellij.uiDesigner.core.GridConstraints;
import com.intellij.uiDesigner.core.GridLayoutManager;
import com.intellij.uiDesigner.core.Spacer;


/**
 * Shows the magnetic components of a project and provides controls for removing these components.
 * Derived from original MeasurementSequencePanel.java
 */
public class MagneticComponentPanel extends ProjectComponent {

    /* Measurement Sequence Table */
    private JTable magTable;
    private MagneticComponentTableModel magTableModel;
    private JButton removeButton;
    private JPanel controlsPane;

    /**
     * Assign session object to this instance.
     *
     * @param session The new session
     */
    @Override
    public void setSession(Session session) {
        super.setSession(session);
        magTableModel.setSession(session);
    }

    /**
     * Creates default MagneticComponentPanel.
     */
    public MagneticComponentPanel() {

        /* Components Table */
        magTableModel = new MagneticComponentTableModel();
        magTable = new JTable(magTableModel);
        magTable.getTableHeader().setReorderingAllowed(false);
        magTable.getTableHeader().setResizingAllowed(true);
        magTable.setColumnSelectionAllowed(false);
        magTable.setRowSelectionAllowed(true);
        StyledTableCellRenderer stcrDefault = new StyledTableCellRenderer();
        stcrDefault.setTooltipVisible(true);
        magTable.setDefaultRenderer(StyledWrapper.class, stcrDefault);
        magTable.setDefaultEditor(StyledWrapper.class, new StyledCellEditor(new JTextField()));
        magTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        magTableModel.setTable(magTable);
        /* Build Layout */
        JScrollPane scrollPane = new JScrollPane(magTable);
        scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        scrollPane.getViewport().setBackground(Color.WHITE);

        setLayout(new BorderLayout());
        add(scrollPane);

        controlsPane = new JPanel();
        controlsPane.setBorder(BorderFactory.createEmptyBorder(0, 4, 4, 4));
        controlsPane.setLayout(new GridLayoutManager(2, 2, new Insets(0, 0, 0, 0), 5, 0));
        removeButton = new JButton();
        removeButton.setText("Remove component");
        removeButton.setEnabled(false);
        controlsPane.add(removeButton, new GridConstraints(1, 1, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null));
        final Spacer spacer1 = new Spacer();
        controlsPane.add(spacer1, new GridConstraints(1, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_WANT_GROW, 1, null, null, null));
        final Spacer spacer2 = new Spacer();
        controlsPane.add(spacer2, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_VERTICAL, 1, GridConstraints.SIZEPOLICY_FIXED, null, new Dimension(-1, 4), null));
        add(controlsPane, "South");

        /* Event Listeners */

        removeButton.addActionListener(new ActionListener() {

            // "Remove" button was clicked, delete the selected component
            public void actionPerformed(ActionEvent e) {
                int row = magTable.getSelectedRow();
                magTableModel.getSession().deleteComponent(magTable.getSelectedRow());
                int lastRowIndex = magTableModel.getSession().getNumOfComponents() - 1;
                if (lastRowIndex >= 0) {
                    // select earlier row next, if any
                    if (row > 0) {
                        magTable.setRowSelectionInterval(row - 1, row - 1);
                    } else {
                        magTable.setRowSelectionInterval(0, 0);
                    }
                } else {
                    removeButton.setEnabled(false);
                }
            }
        });

        magTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e) {
                boolean enabled;
                // enable the "Remove" button only if a component row has been selected
                enabled = magTable.getSelectedRowCount() > 0;
                removeButton.setEnabled(enabled);
            }
        });

        // resize the table columns to be always the right size
        magTable.getColumnModel().addColumnModelListener(new TableColumnModelListener() {
            public void columnAdded(TableColumnModelEvent e) {
                updateColumns();
            }

            public void columnRemoved(TableColumnModelEvent e) {
                updateColumns();
            }

            public void columnMoved(TableColumnModelEvent e) {
                updateColumns();
            }

            public void columnMarginChanged(ChangeEvent e) {
                // DO NOTHING
            }

            public void columnSelectionChanged(ListSelectionEvent e) {
                // DO NOTHING
            }
        });


        magTable.getTableHeader().addMouseListener(new MouseListener() {
            public void mouseClicked() {

            }

            public void mouseReleased(MouseEvent e) {

            }

            public void mousePressed(MouseEvent e) {

            }

            public void mouseExited(MouseEvent e) {

            }

            public void mouseEntered(MouseEvent e) {

            }

            public void mouseClicked(MouseEvent e) {

            }
        });

        // show the column's description in the header as a tooltip
        magTable.getTableHeader().addMouseMotionListener(new MouseMotionAdapter() {

            private int currentColumn;

            @Override
            public void mouseMoved(MouseEvent e) {
                int column = magTable.getColumnModel().getColumnIndexAtX(e.getX());

                if (column != currentColumn) {
                    magTable.getTableHeader().setToolTipText(magTableModel.getColumnToolTip(column));
                    currentColumn = column;     // avoin unnecessary setToolTipText() calls
                }
            }
        });

        // repaint the table when the underlying data has changed
        magTableModel.addTableModelListener(new TableModelListener() {
            public void tableChanged(TableModelEvent e) {
                magTable.revalidate();
            }
        });


    }


    /**
     * Resize the table's columns to fit the content.
     * Cells are set to one and the same fixed width.
     */
    private void updateColumns() {
        TableColumnModel columnModel = magTable.getColumnModel();
        for (int col = 0; col < columnModel.getColumnCount(); col++) {
            if (columnModel.getColumn(col).getHeaderValue().equals(ComponentColumn.COUNT.getColumnName())) {
                // find out the column's preferred width using the actual cell contents
                int width = 20;
                Component comp;
                for (int row = 0; row < magTable.getRowCount(); row++) {
                    comp = magTable.getCellRenderer(row, col).getTableCellRendererComponent(magTable,
                            magTable.getValueAt(row, col), false, false, row, col);
                    width = Math.max(width, comp.getPreferredSize().width);
                }
                columnModel.getColumn(col).setPreferredWidth(width);
                columnModel.getColumn(col).setMaxWidth(width);
            }
            // Resize the ANTIPOLE column
            else if (columnModel.getColumn(col).getHeaderValue().equals(ComponentColumn.ANTIPOLE.getColumnName())) {
                int width = 20;
                columnModel.getColumn(col).setPreferredWidth(width);
                columnModel.getColumn(col).setMaxWidth(width);
            }
        }
    }

}
