/*
 * MeasurementSequenceTableModel.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis.gui;

import mavis.*;

import javax.swing.table.AbstractTableModel;
import javax.swing.*;
import java.util.ArrayList;
import java.util.List;
import java.text.DecimalFormat;

/**
 * This is basically an intermediate wrapper for the data in the actual JTable table.
 * It mainly takes care of number format presentation, and notifies listening clients
 * about data content changes.
 * It will also tell the JTable to autoselect the newest table entry, if
 * a new component was added (done externally from this class).
 * Derived from original MeasurementSequenceTableModel.java
 *
 * @author Jan Wagner (originally by Esko Luontola)
 */
public class MagneticComponentTableModel extends AbstractTableModel implements SessionListener, ProjectListener {

    // which columns are visible in the table (currently all are)
    private static final String VISIBLE_COLUMNS_PROPERTY = "columns";

    private JTable table = null;

    private Session session = null;

    private List<ComponentColumn> columns = new ArrayList<ComponentColumn>();

    /**
     * Creates a new MeasurementSequenceTableModel with no active project.
     */
    public MagneticComponentTableModel() {
        // initialize with no session
        setSession(null);
    }


    /**
     * 'Session' accessors.
     */

    public Session getSession() {
        return session;
    }

    public void setSession(Session sess) {
        // unregister old listener connections
        if (this.session != null) {
            Project project = session.getProject();
            if (project != null) {
                project.removeProjectListener(this);
            }
            session.removeSessionListener(this);
        }

        this.session = sess;

        // set all possible columns, in order of appearance
        columns.clear();
        ComponentColumn[] allColumns = ComponentColumn.getAllColumns();
        for (ComponentColumn column : allColumns) {
            columns.add(column);
            //showColumn(column, false);
        }
        fireTableStructureChanged();

        if (session != null) {
            session.addSessionListener(this);
            Project project = session.getProject();
            if (project != null) {
                project.addProjectListener(this);
            }
        }
    }

    /**
     * Method automatically called from outside on changes in Session
     */
    public void sessionUpdated(SessionEvent event) {
        if (event != null) {
            Session s = event.getSession();
            SessionEvent.Type type = event.getType();

            if (SessionEvent.Type.COMPONENT_ADD_NEW == type) {
                // currently, new components are always added to
                // the end of the components list
                int idx = s.getNumOfComponents();
                if (idx > 0) {
                    idx--;
                    fireTableRowsUpdated(idx, idx);
                    // autoselect the newest entry
                    // TODO: this shouldn't be in this class
                    if (table != null) {
                        table.requestFocusInWindow();
                        table.changeSelection(table.getRowCount() - 1, 0, false, false);
                    }
                }
            } else if (SessionEvent.Type.COMPONENT_REMOVE == type) {
                fireTableDataChanged();
            } else if (SessionEvent.Type.DATA_CHANGE == type) {
                fireTableDataChanged();
            }
        }
    }

    public void projectUpdated(ProjectEvent event) {
        if (event.getType() == ProjectEvent.Type.DATA_CHANGED) {
            //fireTableDataChanged();  cf FILE_SAVED
        } else if (event.getType() == ProjectEvent.Type.STATE_CHANGED) {
            //fireTableDataChanged();
        } else if (event.getType() == ProjectEvent.Type.FILE_SAVED) {
            // apparently some fields do not generate DATA_CHANGED,
            // so use FILE_SAVED instead
            fireTableDataChanged();
        }
    }

    /**
     * Returns the number of rows in the model. A <code>JTable</code> uses this method to determine how many rows it
     * should display.  This method should be quick, as it is called frequently during rendering.
     *
     * @return the number of rows in the model
     * @see #getColumnCount
     */
    public int getRowCount() {
        if (null == session) {
            return 0;
        } else {
            return session.getNumOfComponents();
        }
    }

    public void setTable(JTable table) {
        this.table = table;
    }

    /**
     * Returns the number of columns in the model. A <code>JTable</code> uses this method to determine how many columns
     * it should create and display by default.
     *
     * @return the number of columns in the model
     * @see #getRowCount
     */
    public int getColumnCount() {
        return Math.max(1, columns.size());
    }

    /**
     * Returns the value for the cell at <code>columnIndex</code> and <code>rowIndex</code>.
     *
     * @param rowIndex    the row whose value is to be queried
     * @param columnIndex the column whose value is to be queried
     * @return the value Object at the specified cell
     */
    public Object getValueAt(int rowIndex, int columnIndex) {
        Object o = null;
        try {
            o = columns.get(columnIndex).getValue(rowIndex, session);
        } catch (Throwable t) {
            t.printStackTrace();
        }
        return o; 
    }

    /**
     * Sets the value for the cell at <code>columnIndex</code> and <code>rowIndex</code>.
     *
     * @param value       the new data
     * @param rowIndex    the row whose value is to be queried
     * @param columnIndex the column whose value is to be queried
     */
    public void setValueAt(Object value, int rowIndex, int columnIndex) {
        try {
            columns.get(columnIndex).setValue(value, rowIndex, session);
            fireTableDataChanged();
            // TODO: clicking checkbox cells invokes setValueAt() but after this they are deselected - add fix?
        } catch (Throwable t) {
            t.printStackTrace();
        }
    }

    /**
     * Returns a name for the column. If <code>column</code> cannot be found, returns an empty string.
     *
     * @param column the column being queried.
     * @return a string containing the default name of column.
     */
    @Override public String getColumnName(int column) {
        if (columns.size() <= column) {
            return "< right-click to select columns >";
        }
        return columns.get(column).getColumnName();
    }

    /**
     * Returns the tooltip text for the specified column. Will be shown in the table header.
     */
    public String getColumnToolTip(int column) {
        if (column < columns.size() && session != null) {
            return columns.get(column).getToolTipText(session.getProject());
        }
        return null;
    }

    /**
     * Returns <code>Object.class</code> regardless of <code>columnIndex</code>.
     *
     * @param columnIndex the column being queried
     * @return the Object.class
     */
    @Override public Class<?> getColumnClass(int columnIndex) {
        if (columns.size() <= columnIndex) {
            return Object.class;
        }
        return columns.get(columnIndex).getColumnClass();
    }

    /**
     * Changes certain columns into the specified number format.
     * @param format to use
     */
    public void changeNumberFormat(String format) {
        String [] ColumnsToChange = {"DJ"};

        SequenceColumn[] allColumns = SequenceColumn.getAllColumns();
        for (int i = 0; i < columns.size(); i++) {
            for (String columnToChange : ColumnsToChange) {
                if(columns.get(i).name().equals(columnToChange))
                    columns.get(i).setNumberFormat(new DecimalFormat(format));
            }
        }
        fireTableDataChanged();
    }

    /**
     * Returns true if the cell can be edited.  See ComponentColumn.java
     *
     * @param rowIndex    the row being queried
     * @param columnIndex the column being queried
     * @return false
     */
    @Override public boolean isCellEditable(int rowIndex, int columnIndex) {
        return columns.get(columnIndex).isCellEditable(rowIndex);
    }

}