package mavis.gui;

import mavis.*;

import javax.vecmath.Vector3d;
import java.awt.*;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Point2D;
import java.util.Vector;
import java.text.DecimalFormat;


/**
 *  Draws ZijderveldPlot and accompanying component-vector according to steps given.
 *  Also handles zooming of the plots. 
 * 
 * @author Kimmo Holm
 * Date: 23-Apr-2006
 */
public class ZijderveldPlot extends AbstractPlot implements MouseListener, MouseMotionListener {

	/**
	 * Contains all the data that is shown in graphs.
	 */

	private static Vector<Integer> StepIndex = new Vector<Integer>();
	private static Vector<Vector<Integer>> StepIndexCache;
	private static int StepI = -1;

	private static Vector<Point2D> points = new Vector<Point2D>();
	private static Vector<Vector<Point2D>> pointsCache;
	private static int pointsIndex;

	private static Vector<Point2D> pointsNU = new Vector<Point2D>();
	private static Vector<Vector<Point2D>> pointsNUCache;
	private static int pointsNUIndex;

	private static Vector<Point2D> pointsReal = new Vector<Point2D>();
	private static Vector<Vector<Point2D>> pointsRealCache;
	private static int pointsRealIndex;

	private static Vector<Point2D> pointsRealNU = new Vector<Point2D>();
	private static Vector<Vector<Point2D>> pointsRealNUCache;
	private static int pointsRealNUIndex;

	private Vector<Vector3d> start = new Vector<Vector3d>();
	private Vector<Vector3d> end = new Vector<Vector3d>();

	private int value;
	private boolean check = false; // true = zplotNW, false = zplotNU
	private boolean both = false;

	// Coordinate-points when mouse-button is pressed and released
	private int xa;
	private int ya;
	private int xb;
	private int yb;

	private static boolean zoom = false;

	private int xFix; // placement of Origo 
	private int yFix; // ..Origo

	// Max values for X- and Y-coordinates
	private double xMax;
	private double xMin;
	private double yMax;
	private double yMin;

	// Drawable area
	private int yArea;
	private int xArea;

	// Ratio to fit all the steps in same area
	private double ratioX;
	private double ratioY;

	// 
	private int ps;

	// Mouse-position when dragged
	private int xDrag;
	private int yDrag;
	private boolean dragged = false;

	private int button;

	/**
	 * Will create a new ZijderveldPlot.
	 *
	 * @param x boolean value, true or false, to determinate which plot is in question; NW or NU.
	 */
	public ZijderveldPlot(boolean x) {
		this.check = x;
		addMouseListener(this);
		addMouseMotionListener(this);
		addPlotListener(this);

		pointsCache = new Vector<Vector<Point2D>>();
		pointsIndex = -1;
		pointsRealCache = new Vector<Vector<Point2D>>();
		pointsRealIndex = -1;

		pointsNUCache = new Vector<Vector<Point2D>>();
		pointsNUIndex = -1;
		pointsRealNUCache = new Vector<Vector<Point2D>>();
		pointsRealNUIndex = -1;
		StepIndexCache = new Vector<Vector<Integer>>();
	}

	/**
	 * Will create a ZijderveldPlot that will have both plots, NW and NU, in same graph.
	 *
	 */
	public ZijderveldPlot() {
		this.both = true;
	}
	
	/**
	 * Will redraw the graphs when called.
	 *
	 */
	public void updatePlots(){
		repaint();
	}

	/**
	 * When mouse is dragged across the surface, while button is pressed, will update X- and Y-values
	 * and repaint the plot to basically draw a rectangle between start- and current mouse position. 
	 *
	 * @param e MouseEvent
	 */
	public void mouseDragged(MouseEvent e){

		if(button == MouseEvent.BUTTON1){
			this.dragged = true;
			this.xDrag = e.getX();
			this.yDrag = e.getY();
			repaint();
		}
	}

	public void mouseMoved(MouseEvent e){}

	/**
	 * When mouse-button (left) is clicked, will update X- and Y-values.
	 *
	 * @param e MouseEvent
	 */
	public void mousePressed(MouseEvent e) {

		this.button = e.getButton();
		if(button == MouseEvent.BUTTON1){
			this.xa = e.getX();
			this.ya = e.getY();
		}
		if(e.getButton() == MouseEvent.BUTTON3 && zoom)
			this.clearCache();
	}

	/**
	 * When mouse-button has been released, will determinate which steps belong in the chosen area.
	 *
	 * @param e MouseEvent
	 */
	public void mouseReleased(MouseEvent e) {

		if(e.getButton() == MouseEvent.BUTTON1){
			this.dragged = false;

			this.xb = e.getX();
			this.yb = e.getY();

			// Determinating max- and min-values
			int yMaxb = ( this.ya > this.yb ) ? this.ya : this.yb;
			int yMinb = ( this.ya < this.yb ) ? this.ya : this.yb;

			int xMaxb = ( this.xa > this.xb ) ? this.xa : this.xb;
			int xMinb = ( this.xa < this.xb ) ? this.xa : this.xb;

			if(Math.abs(this.xa - this.xb) > 10 && Math.abs(this.ya - this.yb) > 10
					&& (pointsCache != null && pointsNUCache != null)){

				// Copying current vectors to cache-vectors.
				zoom = true;

				pointsCache.add(new Vector<Point2D>());
				pointsIndex++;

				pointsRealCache.add(new Vector<Point2D>());
				pointsRealIndex++;

				pointsNUCache.add(new Vector<Point2D>());
				pointsNUIndex++;

				pointsRealNUCache.add(new Vector<Point2D>());
				pointsRealNUIndex++;

				StepIndexCache.add(new Vector<Integer>());
				StepI++;

				for(Integer step : StepIndex)
					StepIndexCache.elementAt(StepI).addElement(step);
				StepIndex.clear();

				for (Point2D point : points)
					pointsCache.elementAt(pointsIndex).addElement(point);
				points.clear();

				for (Point2D point : pointsReal)
					pointsRealCache.elementAt(pointsRealIndex).addElement(point);
				pointsReal.clear();

				for (Point2D point : pointsNU)
					pointsNUCache.elementAt(pointsNUIndex).addElement(point);
				pointsNU.clear();

				for (Point2D point : pointsRealNU)
					pointsRealNUCache.elementAt(pointsRealNUIndex).addElement(point);
				pointsRealNU.clear();

				for(int i = 0; i < pointsCache.elementAt(pointsIndex).size(); i++){
					int x = 0;
					int y = 0;

					if(this.check){
						x = xFix + new Double(
								pointsCache.elementAt(pointsIndex).elementAt(i).getX() * this.xArea * this.ratioX
							).intValue() - (ps / 2);
						y = yFix - new Double(
								pointsCache.elementAt(pointsIndex).elementAt(i).getY() * this.yArea * this.ratioY
							).intValue() - (ps / 2);
					}
					if(!this.check){
						x = xFix + new Double(
								pointsNUCache.elementAt(pointsNUIndex).elementAt(i).getX() * this.xArea * this.ratioX
							).intValue() - (ps / 2);
						y = yFix - new Double(
								pointsNUCache.elementAt(pointsNUIndex).elementAt(i).getY() * this.yArea * this.ratioY
							).intValue() - (ps / 2);
					}

					// those coordinates (X,Y), that are within the rectangle drawn, will be added to real vector(s).
					if(!(x > xMaxb || x < xMinb ) && !(y > yMaxb || y < yMinb)){
						points.add(pointsCache.elementAt(pointsIndex).elementAt(i));
						pointsReal.add(pointsRealCache.elementAt(pointsRealIndex).elementAt(i));
						pointsNU.add(pointsNUCache.elementAt(pointsNUIndex).elementAt(i));
						pointsRealNU.add(pointsRealNUCache.elementAt(pointsRealNUIndex).elementAt(i));
						StepIndex.add(StepIndexCache.elementAt(StepI).elementAt(i));
					}
				}

				if(StepIndex.size() > 2){
					super.updatePlots();
				}
				else
					clearCache();
			}
		} else
			repaint();
	}

	/**
	 * Zoom out if plot has been zoomed in and right-button is pressed
	 *
	 * @param e MouseEvent
	 */

	public void mouseClicked(MouseEvent e) {}

	/**
	 * Will copy previous steps from cache to real vector and remove current cache-vector.
	 *
	 */
	public void clearCache() {

		points.clear();
		pointsReal.clear();
		pointsNU.clear();
		pointsRealNU.clear();
		StepIndex.clear();

		for (Point2D point : pointsCache.elementAt(pointsIndex))
			points.addElement(point);
		pointsCache.removeElementAt(pointsIndex);
		pointsIndex--;

		for (Point2D point : pointsRealCache.elementAt(pointsRealIndex))
			pointsReal.addElement(point);
		pointsRealCache.removeElementAt(pointsRealIndex);
		pointsRealIndex--;

		for (Point2D point : pointsNUCache.elementAt(pointsNUIndex))
			pointsNU.addElement(point);
		pointsNUCache.removeElementAt(pointsNUIndex);
		pointsNUIndex--;

		for (Point2D point : pointsRealNUCache.elementAt(pointsRealNUIndex))
			pointsRealNU.addElement(point);
		pointsRealNUCache.removeElementAt(pointsRealNUIndex);
		pointsRealNUIndex--;

		for(Integer step : StepIndexCache.elementAt(StepI))
			StepIndex.addElement(step);
		StepIndexCache.removeElementAt(StepI);
		StepI--;

		if(pointsIndex < 0)
			zoom = false;

		super.updatePlots();
	}

	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}

	/**
	 * Update component Vectors when a new component has been added.
	 *
	 * @param s Current session
	 */
	public void updateLine(Session s) {

		Vector3d[] endpoints;
		start.clear();
		end.clear();

		this.value = s.getNumOfComponents();

		for (int i = 0; i < value; ++i) {
			endpoints = s.getStartAndEndpoints(i);
			if (endpoints != null) {
				start.add(endpoints[0]);
				end.add(endpoints[1]);
			}
		}
		super.updatePlots();
	}

	/**
	 * Adds one measurement step to this graph and converts the data to x- and y- coordinates
	 * 
	 * @param step step to be added
	 */
	public void add(MeasurementStep step) {

		Double valueX;
		Double valueY;
		Double valueZ;

		valueX = MeasurementValue.ZIDJERVELD_X_NORMALIZED.getValue(step);
		valueY = MeasurementValue.ZIDJERVELD_Y_NORMALIZED.getValue(step);
		valueZ = MeasurementValue.ZIDJERVELD_Z_NORMALIZED.getValue(step);

		Double valueXb;
		Double valueYb;
		Double valueZb;

		valueXb = MeasurementValue.ZIDJERVELD_X.getValue(step);
		valueYb = MeasurementValue.ZIDJERVELD_Y.getValue(step);
		valueZb = MeasurementValue.ZIDJERVELD_Z.getValue(step);


		if (valueX != null && valueY != null && valueZ != null) {
			points.add(new Point2D.Double(valueX, -valueY));
			pointsReal.add(new Point2D.Double(valueXb, -valueYb));
			pointsNU.add(new Point2D.Double(valueX, -valueZ));
			pointsRealNU.add(new Point2D.Double(valueXb, -valueZb));
			StepIndex.addElement((int)step.getProject().getStepIndex(step));
		}
	}

	/**
	 * Resets the graph data and repaints its contents.
	 */
	public void reset() {
		points.clear();
		pointsReal.clear();
		pointsNU.clear();
		pointsRealNU.clear();
		StepIndex.clear();

		if(pointsCache != null && pointsNUCache != null){
			pointsCache.clear();
			pointsRealCache.clear();
			StepIndexCache.clear();

			zoom = false;
			StepI = -1;
			pointsIndex = -1;
			pointsRealIndex = -1;

			pointsNUCache.clear();
			pointsRealNUCache.clear();
			pointsNUIndex = -1;
			pointsRealNUIndex = -1;
		}

		start.clear();
		end.clear();

		super.updatePlots();
	}

	/**
	 * Returns the number of points in this graph.
	 *
	 * @return the number of points in this graph
	 */
	public int getNumMeasurements() {
		return points.size();
	}

	/**
	 * Draws the contents of the plot
	 *
	 * @param w  Width of the drawable area
	 * @param h  Height of the drawable area
	 * @param g2 Graphics context
	 */
	public void render(int w, int h, Graphics2D g2) {


		if(this.dragged){
			g2.drawLine(this.xa, this.ya, this.xDrag, this.ya);
			g2.drawLine(this.xa, this.ya, this.xa, this.yDrag);

			g2.drawLine(this.xDrag, this.ya, this.xDrag, this.yDrag);
			g2.drawLine(this.xa, this.yDrag, this.xDrag, this.yDrag);
		}

		// margin
		int m = 20;

		// y-axis padding from arrow top to max values
		int yPad = 20;

		// x-axis padding from arrow top to max values
		int xPad = 20;

		double yMaxReal = 0.0;
		double yMinReal = 0.0;
		double xMaxReal = 0.0;
		double xMinReal = 0.0;


		// minmax values of Y- and X-axis
		if(this.check && points.size() > 0){
			this.yMax = points.elementAt(0).getY();
			this.xMax = points.elementAt(0).getX();
			this.yMin = points.elementAt(0).getY();
			this.xMin = points.elementAt(0).getX();
			yMaxReal = pointsReal.elementAt(0).getY();
			xMaxReal = pointsReal.elementAt(0).getX();
			yMinReal = pointsReal.elementAt(0).getY();
			xMinReal = pointsReal.elementAt(0).getX();
		}
		if(!this.check && pointsNU.size() > 0){
			this.yMax = pointsNU.elementAt(0).getY();
			this.xMax = pointsNU.elementAt(0).getX();
			this.yMin = pointsNU.elementAt(0).getY();
			this.xMin = pointsNU.elementAt(0).getX();
			yMaxReal = pointsRealNU.elementAt(0).getY();
			xMaxReal = pointsRealNU.elementAt(0).getX();
			yMinReal = pointsRealNU.elementAt(0).getY();
			xMinReal = pointsRealNU.elementAt(0).getX();

		}

		for (Point2D point : points) {
			this.yMax = Math.max(this.yMax, point.getY());
			this.xMax = Math.max(this.xMax, point.getX());
			this.yMin = Math.min(this.yMin, point.getY());
			this.xMin = Math.min(this.xMin, point.getX());
		}

		for (Point2D point : pointsReal) {
			yMaxReal = Math.max(yMaxReal, point.getY());
			xMaxReal = Math.max(xMaxReal, point.getX());
			yMinReal = Math.min(yMinReal, point.getY());
			xMinReal = Math.min(xMinReal, point.getX());
		}
		
		for (Point2D point : pointsNU) {
			this.yMax = Math.max(this.yMax, point.getY());
			this.xMax = Math.max(this.xMax, point.getX());
			this.yMin = Math.min(this.yMin, point.getY());
			this.xMin = Math.min(this.xMin, point.getX());
		}

		for (Point2D point : pointsRealNU) {
			yMaxReal = Math.max(yMaxReal, point.getY());
			xMaxReal = Math.max(xMaxReal, point.getX());
			yMinReal = Math.min(yMinReal, point.getY());
			xMinReal = Math.min(xMinReal, point.getX());
		}
		
		// pixels on Y-area and X-area
		this.yArea = getSize().height - ((2 * m) + yPad);
		this.xArea = getSize().width - ((2 * m) + xPad);

		// Fitting all the steps in the available area
		this.ratioY = (1.0 / (0.0005 + this.yMax - this.yMin));
		this.ratioX = (1.0 / (0.0005 + this.xMax - this.xMin));
		
		this.xMax = this.xMax * this.xArea * this.ratioX;
		this.xMin = this.xMin * this.xArea * this.ratioX;

		this.yMax = this.yMax * this.yArea * this.ratioY;
		this.yMin = this.yMin * this.yArea * this.ratioY;

		// font for texts
		g2.setFont(new Font("Arial", Font.PLAIN, 8 + (Math.min(xArea, yArea) / 150)));
		
		// Origo placement
		// X-fix
		if (this.xMin < 0 || this.xMin > 50)
			this.xFix = (int) (m - this.xMin);
		else
			this.xFix = m;

		// Y-fix
		if (this.yMin < 0 || this.yMin > 50)
			this.yFix = (int) (h - m + this.yMin);
		else
			this.yFix = h - m;

		// draw Y-axis
		int xStart = this.xFix;
		int yStart = 15;
		int xEnd = this.xFix;
		int yEnd = getSize().height - 2;

		g2.drawLine(xStart, yStart, xEnd, yEnd);

		// draw X-axis
		int xStart_x = 1;
		int xEnd_x = getSize().width - 2;
		int yStart_x = this.yFix;
		int yEnd_x = this.yFix;

		g2.drawLine(xStart_x, yStart_x, xEnd_x, yEnd_x);
		
		// adjusting ticks
		int tickSize = 2;
		int tickFixX = this.xFix;
		int tickFixY = this.yFix;
		boolean arrowX = true;
		boolean arrowY = true;
		
		// Y
		if(this.xFix > (getSize().width - 5)){
			tickFixX = getSize().width - 5;
			tickSize = 2;
			arrowX = false;
		}
		if(this.xFix < 5){
			tickFixX = 5;
			tickSize = 2;
		}
		// X
		if(this.yFix > (getSize().height - 5)){
			tickFixY = getSize().height - 5;
			tickSize = 2;
		}

		if(this.yFix < 15){
			tickFixY = 15;
			tickSize = 2;
			arrowY = false;
		}

		// arrow width
		int aw = 3;
		
		// Y-arrow
		if(arrowY){
			g2.drawLine(xStart - aw, yStart + aw, xStart, yStart);
			g2.drawLine(xStart + aw, yStart + aw, xStart, yStart);
		}

		// X-arrow
		if(arrowX){
			g2.drawLine(xEnd_x - aw, this.yFix - aw, xEnd_x, this.yFix);
			g2.drawLine(xEnd_x - aw, this.yFix + aw, xEnd_x, this.yFix);
		}

		int adjY = 5;
		int adjX = 5;

		if (this.check && !this.both) {
			// Y - axis unit
			if(arrowY)
				g2.drawString("W", xStart + adjX, yStart + adjY);

			// X - axis unit
			if(arrowX)
				g2.drawString("N", xEnd_x - adjX, yEnd_x - adjY);
		}

		if (!this.check && !this.both) {
			// Y - axis unit
			if(arrowY)
				g2.drawString("U", xStart + adjX, yStart + adjY);

			// X - axis unit
			if(arrowX)
				g2.drawString("N", xEnd_x - adjX, yEnd_x - adjY);
		}

		if (this.both) {
			// Y - axis unit
			if(arrowY)
				g2.drawString("W,U", xStart + adjX, yStart + adjY);

			// X - axis unit
			if(arrowX)
				g2.drawString("N", xEnd_x - adjX, yEnd_x - adjY);
		}

		// ticks
		
		double tickGapValueX = 0.1 * (xMaxReal - xMinReal);
		double tickGapValueY = 0.1 * (yMaxReal - yMinReal);

		// Actual tick-lengths, adjusted based on X- and Y-axis ratios (calculated above)
		int tickGapX = (int)(0.1 * (this.xMax - this.xMin));
		int tickGapY = (int)(0.1 * (this.yMax - this.yMin));

		// avoiding loops
		if(tickGapX < 1)
			tickGapX = 1;
		if(tickGapY < 1)
			tickGapY = 1;

		// drawing X-axis ticks
		for(int i = tickFixX; i < xEnd_x; i += tickGapX)
			g2.drawLine(i, tickFixY+tickSize, i, tickFixY-tickSize);
		for(int i = tickFixX; i > xStart_x; i -= tickGapX)
			g2.drawLine(i, tickFixY+tickSize, i, tickFixY-tickSize);

		// drawing Y-axis ticks
		for(int i = tickFixY; i > yStart; i -= tickGapY)
			g2.drawLine(tickFixX-tickSize, i, tickFixX+tickSize, i);
		for(int i = tickFixY; i < yEnd; i += tickGapY)
			g2.drawLine(tickFixX-tickSize, i, tickFixX+tickSize, i);
		
		// Legend
		DecimalFormat intensity = new DecimalFormat("#0.0000 mA/m");
		String gapValueX = intensity.format(tickGapValueX);
		String gapValueY = intensity.format(tickGapValueY);
		
		g2.drawLine(5, 6, 35, 6);
		g2.drawLine(5, 4, 5, 8);
		g2.drawLine(35, 4, 35, 8);
		
		if(!this.both)
			g2.setPaint(Color.blue.darker());
		g2.drawString("X: "+gapValueX+" Y: "+gapValueY, 40, 10);

		// draw points
		this.ps = (Math.min(this.xArea, this.yArea) / 250) + 4; // points size

		if (this.check || this.both) {
			for (int i = 0; i < points.size(); i++) {
				int x = new Double((points.elementAt(i).getX() * this.xArea) * this.ratioX).intValue();
				int y = new Double((points.elementAt(i).getY() * this.yArea) * this.ratioY).intValue();

				if(!this.both)
					g2.setPaint(Color.black);
				g2.fillOval((this.xFix + x) - (this.ps / 2), (this.yFix - y) - (this.ps / 2), this.ps, this.ps);
				if(!this.both)
					g2.setPaint(Color.red.darker());
				int step = StepIndex.elementAt(i);
				int pos = 2;
				if(((this.yFix - y) - (this.ps / 2)) < 15)
					pos = - 15;
				g2.drawString("" + (step+1), (this.xFix + x) - (this.ps / 2) + 2, (this.yFix - y) - (this.ps / 2) - pos);
			}
		}

		if (!this.check || this.both) {
			for (int i = 0; i < pointsNU.size(); i++) {
				int x = new Double((pointsNU.elementAt(i).getX() * this.xArea) * this.ratioX).intValue();
				int y = new Double((pointsNU.elementAt(i).getY() * this.yArea) * this.ratioY).intValue();

				if(!this.both)
					g2.setPaint(Color.black);
				g2.drawOval((this.xFix + x) - (this.ps / 2), (this.yFix - y) - (this.ps / 2), this.ps, this.ps);

				if(!this.both)
					g2.setPaint(Color.red.darker());
				int step = StepIndex.elementAt(i);
				int pos = 2;
				if(((this.yFix - y) - (this.ps / 2)) < 15)
					pos = - 15;
				g2.drawString("" + (step+1), (this.xFix + x) - (this.ps / 2) + 2, (this.yFix - y) - (this.ps / 2) - pos);
			}
		}

		// Draw component-vector
		if (!start.isEmpty() && !end.isEmpty()) {
			if (this.check || this.both) {
				for (int i = 0; i < this.value; ++i) {

					xStart = this.xFix + new Double(
							start.elementAt(i).x * this.xArea * this.ratioX).intValue() - (this.ps / 2);
					xEnd = this.xFix + new Double(
							end.elementAt(i).x * this.xArea * this.ratioX).intValue() - (this.ps / 2);

					yStart = this.yFix - new Double(
							start.elementAt(i).y * (-1.0) * this.yArea * this.ratioY).intValue() - (this.ps / 2);
					yEnd = this.yFix - new Double(
							end.elementAt(i).y * (-1.0) * this.yArea * this.ratioY).intValue() - (this.ps / 2);

					if(!this.both)
						g2.setPaint(Color.blue.darker());
					g2.drawLine(xStart, yStart, xEnd, yEnd);
				}
			}
			if (!this.check || this.both) {
				for (int i = 0; i < this.value; ++i) {

					xStart = this.xFix + new Double(
							start.elementAt(i).x * this.xArea * this.ratioX).intValue() - (this.ps / 2);
					xEnd = this.xFix + new Double(
							end.elementAt(i).x * this.xArea * this.ratioX).intValue() - (this.ps / 2);

					yStart = this.yFix - new Double(
							start.elementAt(i).z * (-1.0) * this.yArea * this.ratioY).intValue() - (this.ps / 2);
					yEnd = this.yFix - new Double(
							end.elementAt(i).z * (-1.0) * this.yArea * this.ratioY).intValue() - (this.ps / 2);

					if(!this.both)
						g2.setPaint(Color.blue.darker());
					g2.drawLine(xStart, yStart, xEnd, yEnd);
				}
			}
		}

	}
}
