/*
 * ComponentPrinter.java
 *
 * Copyright (C) 2005 Project SQUID, http://www.cs.helsinki.fi/group/squid/
 *
 * This file is part of Mavis.
 *
 * Ikayaki is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ikayaki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ikayaki; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package mavis.util;

import javax.swing.*;
import java.awt.*;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;

/**
 * Offers methods to print Components (only for Graph/DataPrintPanel actually)
 *
 * @author Juho Julkunen (original version Aki Korpua)
 */
public class ComponentPrinter implements Printable {

    /**
     * Component to be printed
     */
    private Component componentToBePrinted;

    /**
     * Creates new printable "component"
     *
     * @param componentToBePrinted Component
     */
    public ComponentPrinter(Component componentToBePrinted) {
        this.componentToBePrinted = componentToBePrinted;
    }

    /**
     * Static printing command
     *
     * @param c Component to be printed (use Only DataPrintPanel/GraphPrintPanel)
     */
    public static void printComponent(Component c) {
        new ComponentPrinter(c).print(null);
    }

    /**
     * Static printing command
     *
     * @param c       Component to be printed (use Only Graph/DataPrintPanel)
     * @param jobName name for the printing job
     */
    public static void printComponent(Component c, String jobName) {
        new ComponentPrinter(c).print(jobName);
    }

    /**
     * Opens printer dialog and start printing job if we get printer
     */
    public void print(String jobName) {
        if (jobName == null) {
            jobName = "Java Printing";
        }
        PrinterJob printJob = PrinterJob.getPrinterJob();
        printJob.setPrintable(this);
        System.out.println(printJob.getJobName());
        printJob.setJobName(jobName);
        if (printJob.printDialog()) {
            try {
                printJob.print();
            } catch (PrinterException pe) {
                System.out.println("Error printing: " + pe);
            }
        }
    }

    /**
     * Absolutely chaotic printing mechanism.
     *
     * @param g          Graphics
     * @param pageFormat PageFormat
     * @param pageIndex  int
     * @return int
     */
    public int print(Graphics g, PageFormat pageFormat, int pageIndex) {
        Dimension dim = componentToBePrinted.getSize();
        if (dim.getHeight() < (pageIndex * 695)) {
            return (NO_SUCH_PAGE);
        }
        Graphics2D g2d = (Graphics2D) g;

            g2d.translate(pageFormat.getImageableX(),
                    pageFormat.getImageableY() - 695 * pageIndex);

        disableDoubleBuffering(componentToBePrinted);
        componentToBePrinted.paint(g2d);

        enableDoubleBuffering(componentToBePrinted);
        return (PAGE_EXISTS);
    }


    public static void disableDoubleBuffering(Component c) {
        RepaintManager currentManager = RepaintManager.currentManager(c);
        currentManager.setDoubleBufferingEnabled(false);
    }

    public static void enableDoubleBuffering(Component c) {
        RepaintManager currentManager = RepaintManager.currentManager(c);
        currentManager.setDoubleBufferingEnabled(true);
    }
}
